package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

import java.util.HashSet;

public class AlphabetsVisitor extends ExtendedDepthFirstAdapter {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the directives
    private DirectivesVisitor directives;

    // Visitor that has the information about the constants
    private ConstantsAndFunctionsVisitor constants;
    
    // Stack of process definitions visited during analysis of 
    // a single process
    private Stack visitedProcesses;
    
    // Channel that are written to
    private Set<String> writtenChannels;

    // Channel that are read from
    private Set<String> readChannels;
    
    // Channels found when analysing a channel set
    private Set<String> channelsInSet;

    // Hidden channels that is being constructed
    private Set<String> hiddenChannels;

    // Visible channels that is being constructed
    private Set<String> visibleChannels;
    
    // Substitution channels used in renaming 
    private Map <String,Set<String>> substitutionChannels;
    
    // Indicates if we are in a channel set
    private boolean isChannelSet;
    
    // Current process being analysed
    String currentAnalysedProcess = "";
    
    public AlphabetsVisitor(ProcessesInfoVisitor procInfo, DirectivesVisitor directives,
    		ConstantsAndFunctionsVisitor constants) {
        this.procsInfo = procInfo;
        this.directives = directives;
        this.constants = constants;
        this.writtenChannels = new HashSet<String>();
        this.readChannels = new HashSet<String>();
        this.hiddenChannels = new HashSet<String>();
        this.channelsInSet = new HashSet<String>();
        this.visibleChannels = new HashSet<String>();
        this.visitedProcesses = new Stack();
        this.isChannelSet = false;
        resetSubstitutionChannels();
    }
   
    public void setCurrentAnalysedProcess(String s){
    	this.currentAnalysedProcess = s;
    }
    public void resetChannels() { 
    	this.resetReferencedChannels(); 
    	this.resetHiddenChannels();
    	this.resetVisibleChannels();
    	this.resetChannelsInSet();
    	this.resetSubstitutionChannels();
    }
    
    public void setIsChannelSet(boolean b) { this.isChannelSet = b; }
    
    public Set<String> getReferencedChannels() {
    	Set<String> alphabet = new HashSet<String>();
    	
    	Iterator<String> out = this.writtenChannels.iterator();
    	while(out.hasNext()) {
    		alphabet.add(out.next());
    	}
    	Iterator<String> in = this.readChannels.iterator();
    	while(in.hasNext()) {
    		alphabet.add(in.next());
    	}
    	return alphabet; 
    }
    public Set<String> getWrittenChannels() { 
    	return this.writtenChannels; 
    }
    public Set<String> getReadChannels() { 
    	return this.readChannels; 
  	}
    public Set<String> getHiddenChannels() { 
    	return this.hiddenChannels; 
    }
    public Set<String> getVisibleChannels() { 
    	return this.visibleChannels; 
    }
    public Set<String> getChannelsInSet() { 
    	return this.channelsInSet; 
    }

    public Map <String,Set<String>> getSubstitutionChannels() {
		return substitutionChannels;
	}

	public void setSubstitutionChannels(Map <String,Set<String>> substitutionChannels) {
		this.substitutionChannels = substitutionChannels;
	}

	private void resetReferencedChannels() { 
    	this.writtenChannels = new HashSet<String>(); 
    	this.readChannels = new HashSet<String>(); 
    }
    private void resetSubstitutionChannels() { setSubstitutionChannels(new Hashtable<String, Set<String>>()); }
    private void resetHiddenChannels() { this.hiddenChannels = new HashSet<String>(); }
    private void resetVisibleChannels() { this.visibleChannels = new HashSet<String>(); }
    private void resetChannelsInSet() { this.channelsInSet = new HashSet<String>(); }
    
    public Set<String> analyseChannelSet(Node node) {
        this.resetChannels();
        this.setIsChannelSet(true);
        node.apply(this);
        Set<String> channels = this.getChannelsInSet();
        
        this.setIsChannelSet(false);
        this.resetChannels();

        return channels;
    }

    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        String name = definition.getName().toString().trim();
        this.setCurrentAnalysedProcess(name);

        ACspParameters parameters = (ACspParameters)definition.getParameters();
        PExpression expression = definition.getExpression();
        if (parameters == null) {
            parameters = new ACspParameters();
        }
        expression.apply(this);
    }
    
    public void caseACspEvent(ACspEvent node){
        String name = node.getChannelName().toString().trim();
    	if (this.isChannelSet){
        	this.channelsInSet.add(name);
    	}
    }
    
    public void caseACspAssert(ACspAssert node) {
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	String name = node.getName().getText().trim();
    	if (this.constants.containsConstant(name)) {
        	this.constants.getConstantsDefinitions(name).apply(this);
    	}
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
    	String fname = node.getName().getText().trim();
    	
    	if (this.isChannelSet){
    		
    		if (HandelC.getInstance().isEmbeddedFunction(fname)) {
        		PArgumentList arguments = ((ACspArguments)node.getArguments()).getArgumentList();

    			if (HandelC.getInstance().getEmbeddedFunctionCard(fname) == 2) {
        			// Getting first argument
        			PArgument argument1 = ((ACspLinkedArgumentList)arguments).getArgument();

        			// Getting second argument
        			PArgument argument2 = ((ACspBaseArgumentList)((ACspLinkedArgumentList)arguments).getNext()).getArgument();

        			argument1.apply(this);
        	        Set<String> leftChannels = this.getChannelsInSet();
        	        this.resetChannels();
        	        
        			argument2.apply(this);
        	        Set<String> rightChannels = this.getChannelsInSet();
        	        this.resetChannels();

            		if (fname.equals(HandelC.CSP_SET_UNION)) {
            			leftChannels.addAll(rightChannels);
            		} else if (fname.equals(HandelC.CSP_SET_DIFF)) {
            			leftChannels.removeAll(rightChannels);
            		} else if (fname.equals(HandelC.CSP_SET_INTER)) {
            			leftChannels.retainAll(rightChannels);
            		} else {
                		throw new UndefinedEmbeddedFunctionEvaluationException(node.toString());
            		}
            		this.channelsInSet = leftChannels;
    			}
    		} else {
        		throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_FUNCTION,this.currentAnalysedProcess);
    		}
    	}
    }

    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
    	String pname = node.getName().toString().trim();

        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.visitedProcesses.contains(pname)) {

        	String previousAnalysedProcess = this.currentAnalysedProcess;
        	this.setCurrentAnalysedProcess(pname);

        	this.visitedProcesses.add(pname);
        	PExpression proc = this.procsInfo.getProcDefinition(pname);

        	proc.apply(this);

            this.visitedProcesses.remove(pname);

        	this.setCurrentAnalysedProcess(previousAnalysedProcess);
        
        }
    }
    
    public void caseACspRenamingProcess(ACspRenamingProcess node) {
    	
    	Map currentSubstitutionChannels = getSubstitutionChannels();
    	resetSubstitutionChannels();
    	
    	PSubstitutionList substitutions = node.getSubstitutionList();
    	substitutions.apply(this);
    	
    	PProcess process = node.getProcess();
    	process.apply(this);
    	
    	for (String oldChannel : getSubstitutionChannels().keySet()) {
    		Set <String> list = getSubstitutionChannels().get(oldChannel);
			if (this.writtenChannels.contains(oldChannel)){
				this.writtenChannels.remove(oldChannel);
				this.writtenChannels.addAll(list);
			}
			if (this.readChannels.contains(oldChannel)) {
				this.readChannels.remove(oldChannel);
				this.readChannels.addAll(list);
			}
			if (this.visibleChannels.contains(oldChannel)) {
				this.visibleChannels.remove(oldChannel);
				this.visibleChannels.addAll(list);
			}
			if (this.hiddenChannels.contains(oldChannel)) {
				this.hiddenChannels.remove(oldChannel);
				this.hiddenChannels.addAll(list);
			}
		}
    	
    	setSubstitutionChannels(currentSubstitutionChannels);
    	
    }
    
    public void caseACspSubstitution(ACspSubstitution node) {
    	String oldE = node.getOldEvent().toString().trim();
        String newE = node.getNewEvent().toString().trim();
        
        if (getSubstitutionChannels().containsKey(oldE)){
        	Set<String> subs = getSubstitutionChannels().get(oldE);
        	subs.add(newE);
        	getSubstitutionChannels().put(oldE, subs);
        } else {
        	Set<String> subs = new HashSet<String>();
        	subs.add(newE);
        	getSubstitutionChannels().put(oldE, subs);
		}
    }

    public void caseACspHidingProcess(ACspHidingProcess node){
        Set visibleChannelsAbove = this.visibleChannels;
        Set hiddenChannelsAbove = this.hiddenChannels;
        this.resetVisibleChannels(); 
        this.resetHiddenChannels(); 

        Node hiddenChanExp = ((ACspHidingProcess)node).getHidenEvents();
        Set<String> hidden = this.analyseChannelSet(hiddenChanExp);
        this.hiddenChannels.addAll(hidden);

        Node process = ((ACspHidingProcess)node).getProcess();
        process.apply(this);

        this.visibleChannels.removeAll(hidden);
        this.visibleChannels.addAll(visibleChannelsAbove);
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        PHelp proc = node.getProcess();

    	String channel = ((ACspEvent)event).getChannelName().toString().trim();
        this.visibleChannels.add(channel);

        ListIterator commIter = comm.listIterator(0);

        if (!commIter.hasNext()) {
            if (this.directives.containsProcChannelInOut(this.currentAnalysedProcess)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentAnalysedProcess);
                if (channelInOut.containsKey(channel)){
                    String inOut = (String)channelInOut.get(channel);
                    if (inOut.equals(HandelC.A_READ)) {
                        this.readChannels.add(channel);
                    } else if (inOut.equals(HandelC.A_WRITE)) {
                        this.writtenChannels.add(channel);
                    }
                } else {
                    throw new UndefinedChannelInOutException(channel,this.currentAnalysedProcess);
                }
            } else {
            	// It is not a bus
            	// It will raise an exception anyway because this channel should be either input or output
            	// However, I will leave the restrictionsVisitor to through the right exception
            	if(!this.directives.getBuses().contains(channel)) {
            		//throw new UndefinedChannelInOutException(channel,this.currentAnalysedProcess);
            	}
            }
        } else {
            Object next = commIter.next();
            if (next instanceof ACspInputComunication) {
                this.readChannels.add(channel);
            } else if (next instanceof ACspOutputComunication) {
                this.writtenChannels.add(channel);
            }
        }

        proc.apply(this);
    }
}
