package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.type.CspCollectionType;
import lmf.formula.csp.type.CspSetType;
import lmf.formula.csp.type.CspType;
import lmf.formula.csp.type.CspUnknownType;
import lmf.formula.csp.analysis.*;
import mathCollection.HashMathSet;
import mathCollection.MathSet;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ChannelsVisitor extends ExtendedDepthFirstAdapter {

	private DirectivesVisitor directives;
	private DatatypesVisitor types;
	private ConstantsAndFunctionsVisitor constants;
	
    // Type of the channels
    // Name:String -> Types:PType Vector
    private Hashtable channelTypesMap;
    
    // Indicate which channels communicate values using projections c.*
    private List projChannels; 

    // Indicate which channels communicate values using either ! or ?
    // String Vector
    private List commChannels; 

    // Indicates to the visitor that we are checking the names of
    // the channels being declared
    private boolean checkingChannelNames;

    // Stores the channel names of the current channel definition
    private List channelNames;
    
    // Indicates to the visitor that we are checking the type of
    // the channels being declared
    private boolean checkingChannelTypes;

    // String containing the type cardinality
    private String typeCard;
    
    // String containing the type handel-c equivalent
    private String typeHandelC;
    
    // Stores the channel types of the current channel definition
    private List channelTypes;
    
    // Stores the list of channels (just their name - c)
    private List<String> channels;
    
    // Stores the input buses
    private MathSet inputBuses;
    
    // Stores the output buses
    private MathSet outputBuses;

    // Stores the projection buses (these are actually not allowed)
    private MathSet projectionBuses;
    
    // Stores the channels belonging to a process - AddIvan
    private Hashtable <String,List<String>> processChannels;
    
    private String currentProcess;
    
    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Current Channel Definition
    private ACspChannelDefinition currentChannelDefinition;
    
    public ChannelsVisitor(DirectivesVisitor directives, DatatypesVisitor types, ConstantsAndFunctionsVisitor constants) {
    	this.directives = directives;
    	this.types = types;
    	this.constants = constants;

    	this.processChannels = new Hashtable<String, List<String>>(); // AddIvan
    	this.channelTypesMap = new Hashtable();
        this.commChannels = new Vector();
        this.projChannels = new Vector();
        this.typeCard = "";
        this.typeHandelC = "";
        
        this.inputBuses = new HashMathSet();
        this.outputBuses = new HashMathSet();
        this.projectionBuses = new HashMathSet();
    }
    
    public List<String> getChannels(){
    	if (this.channels == null) {
    		this.channels = new Vector<String>();
            Iterator iterator = this.channelTypesMap.keySet().iterator();
            while(iterator.hasNext()) {
            	String channel = (String)iterator.next();
            	// Buses are not channels
            	if (!this.directives.getBuses().contains(channel)) {
            		this.channels.add(channel);
            	}
            }
    	}
        return this.channels;
    }

    public List<String> getChannelNames(){
    	List<String> channels = this.getChannels();
    	List names = new Vector();
    	for(int i=0; i<channels.size(); i++) {
    		names.add(HandelC.getInstance().getChannelName(channels.get(i)));
    	}
        return names;
    }
    
    public List getProjChannels(){
        return this.projChannels;
    }

    public boolean isCommChannel(String channelName){
        return this.commChannels.contains(channelName);
    }
    public boolean isProjChannel(String channelName){
        return this.projChannels.contains(channelName);
    }

    public MathSet getInputBuses(){
        return this.inputBuses;
    }
    
    public MathSet getOutputBuses(){
        return this.outputBuses;
    }

    public MathSet getProjectionBuses(){
        return this.projectionBuses;
    }
    
    public MathSet getBuses(){
        return this.getInputBuses().union(this.getOutputBuses());
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        PExpression expression = definition.getExpression();
        this.currentProcess = definition.getName().getText().trim();
        //System.out.println("caseACspProcessDefinition:: "+this.currentProcess);
        expression.apply(this);
    }
    
    public void caseACspAbstractDefinition(ACspAbstractDefinition definition) {
    	System.out.println("caseACspAbstractDefinition:: "+definition);
    	throw new AbstractDefinitionException();
    }
    

    public void caseACspChannelDefinition(ACspChannelDefinition node){
        this.currentChannelDefinition = node;
        
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        // Retrieve Channel Names
        Vector names = new Vector();
        
        this.checkingChannelNames = true;
        this.channelNames = new Vector();
        channels.apply(this);
        this.checkingChannelNames = false;
        
        this.checkingChannelTypes = true;
         this.channelTypes = new Vector();
        if (type != null) {
            type.apply(this);
        }
        this.checkingChannelTypes = false;
        
        for(int i=0; i < this.channelNames.size(); i++) {
            this.channelTypesMap.put(this.channelNames.get(i),this.channelTypes);
        }
    }
    
    public void caseACspChannel(ACspChannel node){
        this.channelNames.add(node.getName().toString().trim());
    }
    
    public void caseACspType(ACspType node){
    	String nodeTypeName = node.getName().getText().trim(); 
    	String typeName = nodeTypeName;

    	if (HandelC.getInstance().isSetType(typeName)) {
    		typeName = HandelC.getInstance().getSetType(typeName);
    	}
    	
    	if (this.checkingChannelTypes){
        	if (typeName.equals(CspUtil.getBoolType())
        		|| typeName.equals(CspUtil.getIntType())
        		|| this.types.isDatatype(typeName)){
        		// It is an existing type
            	this.channelTypes.add(node);
            } else if (this.constants.containsConstant(typeName)) {
            	// It is a constant
            	Node constantDefinition = this.constants.getConstantsDefinitions(typeName);
            	MaximalTypesInferenceVisitor.applyToMaximalType(this, constantDefinition);
        	}
        }
    	
        this.setHcType(nodeTypeName);
    }
    
    public void caseACspAbstractType(ACspAbstractType node) {
    	if (this.checkingChannelTypes) {
        	MaximalTypesInferenceVisitor.applyToMaximalType(this, node);
        }
    }
    
    
    public void caseACspSetType(ACspSetType node){
        if (this.checkingChannelTypes){
        	MaximalTypesInferenceVisitor.applyToMaximalType(this, node);
        }
    }

    private void setHcType(String typeName) {
        String hcType = HandelC.getInstance().getHandelCTypeName(typeName);
        this.typeCard = HandelC.getInstance().getCardVarOf(hcType);
        this.typeHandelC = hcType;
    }
    
    public void caseACspTupleType(ACspTupleType node){
        if (this.checkingChannelTypes){
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentChannelDefinition.toString());
        }
    }

    public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)event).getComunication();
        PHelp proc = node.getProcess();
        LinkedList projection = ((ACspEvent)event).getProjection();

        String channelName = ((ACspEvent)event).getChannelName().toString().trim();
        
        // Checking Projections
        if (projection != null) {
            if (!this.projChannels.contains(channelName)) {
                this.projChannels.add(channelName);
            }    
            
            // Storing buses
            if (this.directives.getBuses().contains(channelName) && projection.size() > 0) {
                this.getProjectionBuses().add(channelName);
            }
        }

        // Checking communications
        ListIterator iterator = comm.listIterator(0);
        boolean isCommunication = false;
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if ((next instanceof ACspInputComunication) || (next instanceof ACspOutputComunication)) {
                if (!this.commChannels.contains(channelName)) {
                    this.commChannels.add(channelName);
                }
                
                // Storing buses
                if (this.directives.getBuses().contains(channelName)) {
                    if (next instanceof ACspInputComunication) {
                    	this.getInputBuses().add(channelName);
                    }
                    if (next instanceof ACspOutputComunication) {
                    	this.getOutputBuses().add(channelName);
                    }
                }
            }
        }
        
        // AddIvan: Criando uma lista de canais e associando-a ao seu respectivo processo
        //System.out.println(currentProcess);
        if (this.processChannels.containsKey(this.currentProcess)) {
        	List<String> channelList = this.processChannels.get(this.currentProcess);
        	if (!channelList.contains(channelName)){
        		channelList.add(channelName);
        		this.processChannels.put(this.currentProcess, channelList);
        	}
        } else {
        	List<String> channelList = new Vector <String>();
        	channelList.add(channelName);
        	this.processChannels.put(this.currentProcess, channelList);
        }
        
        proc.apply(this);
    }

    // Returns the dimension of an array of channels
    // This is the number of type expression used in the channel declaration
    // (minus one if this is a communication channel)
    private int getArrayDimension(String channelName) {
        int size = this.getChannelType(channelName).size();
        if (size > 0 && this.isCommChannel(channelName)) {
            size = size - 1;
        }
        return size;
    }

    public boolean containsChannelType(String channelName){
    	
    	//System.out.println("ChannelsVisitor :: containsChannelType -> Channel Types = "+this.channelTypes);
    	
        return this.channelTypesMap.containsKey(channelName);
        
        
    }
    
    public List getChannelType(String channelName){
        return (List)this.channelTypesMap.get(channelName);
    }

    public String getChannelCommType(String chan) {
        String commType = "";
        List types = this.getChannelType(chan);
        int arrayDim = this.getArrayDimension(chan);
        if(types.size() == arrayDim) {
            // There is no actual value being communicated. Type of communication is SYNC
             commType= HandelC.SYNC_TYPE;
        } else {
            // There is a value being communicated. Type of communication is the last type
            PType type = (PType)ListUtil.last(types);
            type.apply(this);
            commType = this.typeHandelC;
        }
        
        return commType;
    }
    
    public String getChannelDeclaration(String chan) {
        String decl = "";
        decl = decl + HandelC.CHAN_DECL + HandelC.SPACE;
        
        // Declaring the value communicated via this channel
        if (this.containsChannelType(chan)) {
            List types = this.getChannelType(chan);
            int arrayDim = this.getArrayDimension(chan);

            if(types.size() == arrayDim) {
                // There is no actual value being communicated. Type of communication is SYNC
                decl = decl + HandelC.SYNC_TYPE;
            } else {
                // There is a value being communicated. Type of communication is the last type
                PType type = (PType)ListUtil.last(types);
                decl = decl + this.getChannelCommType(chan);
            }
            decl = decl + HandelC.SPACE + chan;
            
            // Declaring possible arrays
            
            if(this.isProjChannel(chan)){
                for (int i = 0; i < arrayDim; i++) {
                    PType type = (PType)types.get(i);
                    decl = decl + HandelC.ARRAY_BEGIN;
                    type.apply(this);
                    decl = decl + this.typeCard;
                    decl = decl + HandelC.ARRAY_END;
                }
            }
        }
        return decl;
    }

    public String getBusVarDeclaration(String bus) {
        String decl = "";
		String busVarName = "";

        // <TYPE> bus_<in|out>_<NAME>;

		if(this.inputBuses.contains(bus)) {
    		busVarName = HandelC.getInstance().getInBusVarName((String)bus);
    	} else {
    		busVarName = HandelC.getInstance().getOutBusVarName((String)bus);
    	}
    	
        // Declaring the value communicated via this channel
        if (this.containsChannelType(bus)) {
            List types = this.getChannelType(bus);
            if (types.size() == 1) {
                decl = decl + this.getChannelCommType(bus);
            } 
            decl = decl + HandelC.SPACE + busVarName;
        }
        
        return decl;
    }
    
    public String getBusInterfaceDeclaration(String bus) {
        // Declaring the value communicated via this channel
    	String type = "";
    	if (this.containsChannelType(bus)) {
            List types = this.getChannelType(bus);
            if (types.size() == 1) {
        		if(this.getBuses().contains(bus)) {
                    type = this.getChannelCommType(bus);
            	}
            } 
        }

		String decl = "";
		if(this.inputBuses.contains(bus)) {
			decl = this.getInBusInterfaceDeclaration(bus,type);
    	} else {
			decl = this.getOutBusInterfaceDeclaration(bus,type);
    	}
        return decl;
    }
    
    public String getInBusInterfaceDeclaration(String bus, String type) {
        // interface bus_in(integer value_read) InBus_read() with {data={"read4","read3","read2","read1"}};
    	
    	String decl = HandelC.INTERFACE + HandelC.SPACE + HandelC.BUS_IN
						+ HandelC.PAR_BEGIN + type + HandelC.SPACE + HandelC.getInstance().getBusArgName(bus) + HandelC.PAR_END
						+ HandelC.SPACE + HandelC.getInstance().getInBusMethodName(bus) + HandelC.PAR_BEGIN + HandelC.PAR_END
						+ HandelC.SPACE + HandelC.WITH + HandelC.SPACE
						+ HandelC.BLOCK_BEGIN
						+ HandelC.DATA + HandelC.ASSIGNMENT
						+ HandelC.BLOCK_BEGIN + this.getBusPinNamesDecl(bus,type) + HandelC.BLOCK_END
						+ HandelC.BLOCK_END;
        
        return decl;
    }

    public String getOutBusInterfaceDeclaration(String bus, String type) {
    	// interface bus_out() OutBus_write(integer OutPort=bus_out_write) with {data={"write4","write3","write2","write1"}}; 
    	
    	String decl = HandelC.INTERFACE + HandelC.SPACE + HandelC.BUS_OUT
						+ HandelC.PAR_BEGIN + HandelC.PAR_END
						+ HandelC.SPACE + HandelC.getInstance().getOutBusMethodName(bus) 
						+ HandelC.PAR_BEGIN + type + HandelC.SPACE + HandelC.OUTPORT 
						+ HandelC.ASSIGNMENT + HandelC.getInstance().getOutBusVarName(bus) + HandelC.PAR_END
						+ HandelC.SPACE + HandelC.WITH + HandelC.SPACE
						+ HandelC.BLOCK_BEGIN
						+ HandelC.DATA + HandelC.ASSIGNMENT
						+ HandelC.BLOCK_BEGIN + this.getBusPinNamesDecl(bus,type) + HandelC.BLOCK_END
						+ HandelC.BLOCK_END;
        return decl;
    }
    
    public String getBusPinNamesDecl(String bus, String type) {
    	int length = this.types.getTypeLength(type);

    	String decl="";
    	for(int i=length; i > 1; i--) {
    		decl = decl + "\"" + HandelC.getInstance().getBusPinName(bus,i) + "\"" + HandelC.SEP_ARGUMENTS;
    	}
		decl = decl + "\"" + HandelC.getInstance().getBusPinName(bus,1) + "\"";
			
    	return decl;
    }
    
    public String getBusSimulationDeclaration(String bus) {
        // Declaring the value communicated via this channel
    	String type = "";
    	if (this.containsChannelType(bus)) {
            List types = this.getChannelType(bus);
            if (types.size() == 1) {
        		if(this.getBuses().contains(bus)) {
                    type = this.getChannelCommType(bus);
            	}
            } 
        }

		String decl = "";
		if(this.inputBuses.contains(bus)) {
			decl = this.getInBusSimulationDeclaration(bus,type);
    	} else {
			decl = this.getOutBusSimulationDeclaration(bus,type);
    	}
        return decl;
    }
    
    public String getInBusSimulationDeclaration(String bus, String type) {
        // chanin integer bus_name with {infile = "bus_name.txt"};
    	
    	String decl = HandelC.CHANIN + HandelC.SPACE + type + HandelC.SPACE + "simulate_" + bus + HandelC.SPACE + HandelC.WITH + HandelC.SPACE
						+ HandelC.BLOCK_BEGIN
						+ "infile" + HandelC.ASSIGNMENT
						+ HandelC.SPACE + "\"" + bus + ".txt\"" + HandelC.BLOCK_END;
        
        return decl;
    }

    public String getOutBusSimulationDeclaration(String bus, String type) {
    	// chanout integer bus_name with {outfile = "bus_name.txt"};
    	
    	String decl = HandelC.CHANOUT + HandelC.SPACE + type + HandelC.SPACE + "simulate_" + bus + HandelC.SPACE + HandelC.WITH + HandelC.SPACE
						+ HandelC.BLOCK_BEGIN
						+ "outfile" + HandelC.ASSIGNMENT
						+ HandelC.SPACE + "\"" + bus + ".txt\"" + HandelC.BLOCK_END;
        return decl;
    }
    
    // AddIvan
    public Hashtable<String, List<String>> getProcessChannels() {
		return this.processChannels;
	}
    
    public void caseACspAssert(ACspAssert node) {
    }
    
    public void insertCommChannel(String channelName){
    	this.commChannels.add(channelName);
    }
    
    public void insertProjChannel(String channelName){
    	this.projChannels.add(channelName);
    }

}

