package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;
import mathCollection.HashMathSet;
import mathCollection.MathSet;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class DirectivesVisitor extends ExtendedDepthFirstAdapter {
	
	// DEFAULT VALUES
	private static int DEFAULT_INT_BITS = 1;
	private static String DEFAULT_MAIN_PROC = "";
	private static String DEFAULT_INT_CHOICE = Directives.COMPILE_TIME;
	
    // Directives Information

    // Channel Input or Output -- procName:String -> (channelName:String -> A_WRITE/A_READ:String)
    private Hashtable channelInOut;

    // Argument types -- procName:String -> (argumentName:String -> type:String)
    private Hashtable arguments;

    // Empty sets types -- procName:String -> type:String
    private Hashtable setTypes;
    
    // Main process
    private String mainProc;

    // Number of bits for the CSP type Int
    private int intBits;

    // Type of internal choice
    private String intChoice;

    // Sets of buses
    private MathSet buses;
    
    // Current directive
    private String currentDirective;
    
    // Current line
    private int currentLine;

    // Indicates that we are reading the directive main
    private boolean isReadingMain;

    public DirectivesVisitor() {
        this.channelInOut = new Hashtable();
        this.arguments = new Hashtable();
        this.setTypes = new Hashtable();
        this.mainProc = DEFAULT_MAIN_PROC;
        this.currentDirective = "";
        this.currentLine = 1;
        this.intBits = DEFAULT_INT_BITS;
        this.intChoice = DEFAULT_INT_CHOICE;
        this.isReadingMain = false;
        this.buses = new HashMathSet();
    }
    
    public void printChannelInOut (){
    	System.out.println(this.channelInOut);
    }

    public boolean containsProcChannelInOut(String procName) {
        return this.channelInOut.containsKey(procName);
    }
    
    public Hashtable getProcChannelInOut(String procName){
        return (Hashtable)this.channelInOut.get(procName);
    }

    public void putProcChannelInOut(String procName, String name, String type){
        if (this.channelInOut.containsKey(procName)) {
            Hashtable hashInOut = (Hashtable)this.channelInOut.get(procName);
            hashInOut.put(name,type);
        } else {
            Hashtable hashInOut = new Hashtable();
            hashInOut.put(name,type);
            this.channelInOut.put(procName,hashInOut);
        }
    }
    
    public boolean containsProcArguments(String procName) {
        return this.arguments.containsKey(procName);
    }
    
    public Hashtable getProcArguments(String procName){
        return (Hashtable)this.arguments.get(procName);
    }
    
    public void putProcArguments(String procName, String name, String type){
    	if (this.arguments.containsKey(procName)) {
            Hashtable hashArgs = (Hashtable)this.arguments.get(procName);
            hashArgs.put(name,type);
        } else {
            Hashtable hashArgs = new Hashtable();
            hashArgs.put(name,type);
            this.arguments.put(procName,hashArgs);
        }
    }

    public boolean containsSetType(String def) {
        return this.setTypes.containsKey(def);
    }
    
    public String getSetType(String def){
        return (String)this.setTypes.get(def);
    }
    
    public String getMain(){
        return this.mainProc;
    }
    
    public int getIntBits() { return this.intBits; }

    public String getIntChoice() { return this.intChoice; }

    public MathSet getBuses() { return this.buses; }
    
    public void outStart(Start node) {
        // If directive for main method is not declared or is empty, set to default SKIP
        if (this.mainProc.trim().equals("")) {
            this.mainProc = CspUtil.getSkip();
        }
        
        //System.out.println("DirectivesVisitor :: outStart -> buses="+this.buses);
    }
    
    public void caseACspLineComment(ACspLineComment node){
        TSingleLineComment lineComment = node.getSingleLineComment();
        if (lineComment.toString().trim().startsWith(Directives.DIRECTIVE)){
            this.currentDirective = lineComment.toString();
            this.currentLine = lineComment.getLine();
            this.processDirective(lineComment);
        }
    }

    private void processDirective(TSingleLineComment lineComment){
        this.currentLine = lineComment.getLine();
        this.currentDirective = lineComment.toString();
        
        String directive = this.currentDirective;
        directive = StringUtil.removeBreakLines(directive);
        directive = StringUtil.removeReturns(directive);
        directive = StringUtil.removeTabs(directive);
        directive = StringUtil.removeExtraSpaces(directive);
        
        StringTokenizer tokenizer = new StringTokenizer(directive,HandelC.SPACE);
        
        // throws the directive header away
        tokenizer.nextToken();
        if (tokenizer.hasMoreTokens()) {
            // gets the directive type
            String directiveType = tokenizer.nextToken();
            if (directiveType.equals(Directives.DIR_CHANNEL_IN_OUT)){
            	this.isReadingMain = false;
                this.processChannelInOut(tokenizer);
            } else if (directiveType.equals(Directives.DIR_ARGUMENT)){
            	this.isReadingMain = false;
                this.processArgumentType(tokenizer);
            } else if (directiveType.equals(Directives.DIR_MAIN)){
            	this.isReadingMain = true;
                this.processMain(tokenizer);
            } else if (directiveType.equals(Directives.DIR_INT_BITS)){
            	this.isReadingMain = false;
                this.processIntBits(tokenizer);
            } else if (directiveType.equals(Directives.DIR_SET_TYPE)){
            	this.isReadingMain = false;
                this.processSetType(tokenizer);
            } else if (directiveType.equals(Directives.DIR_INT_CHOICE)){
            	this.isReadingMain = false;
                this.processIntChoice(tokenizer);
            } else if (directiveType.equals(Directives.DIR_BUSES)){
            	this.isReadingMain = false;
                this.processBuses(tokenizer);
            } else {
            	if (this.isReadingMain) {
                    this.processMain(directiveType,tokenizer);
            	} else {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.INEXISTENT_DIRECTIVE,this.currentLine);
            	}
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.INEXISTENT_DIRECTIVE,this.currentLine);
        }
    }

    private void processChannelInOut(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String name = null;
        String type = null;
        String within = null;
        String proc = null;
        while (tokenizer.hasMoreTokens()) {
            if (name == null) {
                name = tokenizer.nextToken();
            } else if (type == null) {
                type = tokenizer.nextToken();
            } else if (within == null) {
                within = tokenizer.nextToken();
                if (!within.equals(Directives.WITHIN)) {
                    throw new InvalidDirectiveException(this.currentDirective, Messenger.EXPECTED+Directives.WITHIN, this.currentLine);
                }
            } else if (proc == null) {
                proc = tokenizer.nextToken();
            } else {
                throw new InvalidDirectiveException(this.currentDirective, Messenger.CHANNEL_IN_OUT_TOO_MUCH_INFO, this.currentLine);
            }
        }
        if (name != null && type != null && within != null && proc != null) {
            if (type.equals(Directives.CHANNEL_IN) || type.equals(Directives.CHANNEL_OUT)) {
            	String action = HandelC.getInstance().getActionFromDirective(type);
                if (this.channelInOut.containsKey(proc)) {
                    Hashtable hashInOut = (Hashtable)this.channelInOut.get(proc);
                    hashInOut.put(name,action);
                } else {
                    Hashtable hashInOut = new Hashtable();
                    hashInOut.put(name,action);
                    this.channelInOut.put(proc,hashInOut);
                }
            } else {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.CHANNEL_IN_OUT_WRONG_TYPE,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.CHANNEL_IN_OUT_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processArgumentType(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String name = null;
        String type = null;
        String within = null;
        String proc = null;
        while (tokenizer.hasMoreTokens()) {
            if (name == null) {
                name = tokenizer.nextToken();
            } else if (type == null) {
                type = "";
                while (tokenizer.hasMoreTokens() && within == null) {
                    String token = tokenizer.nextToken();
                    if (!token.equals(Directives.WITHIN)) {
                        type = type + token + HandelC.SPACE;
                    } else {
                        within = token;
                    }
                }
                if (within == null) {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Directives.WITHIN,this.currentLine);
                } else {
                    if (tokenizer.hasMoreTokens()) {
                        proc = tokenizer.nextToken();
                    } else {
                        throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Messenger.PROCESS_NAME,this.currentLine);
                    }
                }
            }
        }
        if (name != null && type != null && within != null && proc != null) {
            Hashtable hashArguments = new Hashtable();
            if (this.arguments.containsKey(proc)) {
                Hashtable hashArgs = (Hashtable)this.arguments.get(proc);
                hashArgs.put(name,type);
            } else {
                Hashtable hashArgs = new Hashtable();
                hashArgs.put(name,type);
                this.arguments.put(proc,hashArgs);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.ARGUMENT_TOO_FEW_INFO,this.currentLine);
        }
    }
    
    private void processMain(String s,StringTokenizer tokenizer) throws InvalidDirectiveException{
        this.mainProc = this.mainProc + s;
        this.processMain(tokenizer);
    }

    private void processMain(StringTokenizer tokenizer) throws InvalidDirectiveException{
        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();
            this.mainProc = this.mainProc + token;
        }
    }
    
    private void processIntBits(StringTokenizer tokenizer) throws InvalidDirectiveException{
        if (tokenizer.hasMoreElements()) {
            String sizeStr = tokenizer.nextToken();
            try {
                this.intBits = (new Integer(sizeStr)).intValue();
                if (this.intBits < 1) {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_WRONG_TYPE,this.currentLine);
                }
            } catch (NumberFormatException e) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_WRONG_TYPE,this.currentLine);
            }
            if (tokenizer.hasMoreElements()) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_TOO_MUCH_INFO,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processSetType(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String type = null;
        String def = null;
        while (tokenizer.hasMoreTokens()) {
            if (type == null) {
                type = "";
                if (tokenizer.hasMoreTokens()) {
                    String token = tokenizer.nextToken();
                    type = token.trim();
                }
                if (tokenizer.hasMoreTokens()) {
                    String token = tokenizer.nextToken();
                    if (!token.equals(Directives.WITHIN)) {
                        throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Directives.WITHIN,this.currentLine);
                    }
                }
                
                if (tokenizer.hasMoreTokens()) {
                    def = tokenizer.nextToken().trim();
                } else {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Messenger.DEFINITION_NAME,this.currentLine);
                }
            }
        }
        if (type != null && def != null) {
            if (this.setTypes.containsKey(def)) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.SET_TYPE_REDEFINED,this.currentLine);
            } else {
                this.setTypes.put(def,type);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.SET_TYPE_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processIntChoice(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String at = null;
        String type = null;
        while (tokenizer.hasMoreTokens()) {
            if (at == null) {
                at = tokenizer.nextToken();
                if (!at.equals(Directives.AT)) {
                    throw new InvalidDirectiveException(this.currentDirective, Messenger.EXPECTED+Directives.AT, this.currentLine);
                }
            } else if (type == null) {
                type = tokenizer.nextToken();
            } else {
                throw new InvalidDirectiveException(this.currentDirective, Messenger.INT_CHOICE_TOO_MUCH_INFO, this.currentLine);
            }
        }
        if (at != null && type != null) {
            if (type.equals(Directives.COMPILE_TIME) || type.equals(Directives.RUN_TIME)) {
            	this.intChoice = type;
            } else {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_CHOICE_WRONG_TYPE,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_CHOICE_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processBuses(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String name = null;
        while (tokenizer.hasMoreTokens()) {
            if (name == null) {
                name = tokenizer.nextToken();
            } else {
                throw new InvalidDirectiveException(this.currentDirective, Messenger.BUSES_TOO_MUCH_INFO, this.currentLine);
            }
        }
        if (name != null) {
    		this.buses.add(name);
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.BUSES_TOO_FEW_INFO,this.currentLine);
        }
    }

    public void caseACspAssert(ACspAssert node) {
    }
    
}

