package csp2hc;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;
import lmf.formula.csp.semantic.*;
import lmf.formula.csp.semantic.exception.*;

import csp2hc.core.*;
import csp2hc.gui.*;
import csp2hc.util.*;
import csp2hc.messages.*;
import csp2hc.exception.*;

public class Csp2HC {
    private Config config;
    private String csp_files_dir;
    private Start start;
    private boolean useTypeChecker;
    private boolean displayTree;
    
    public Csp2HC() {
        this.loadConfigurations();
    }
    
    public boolean getUseTypeChecker() { return this.useTypeChecker; }
    public boolean getDisplayTree() { return this.displayTree; }
        
    private void loadConfigurations(){
        Messenger.getInstance().start(Messenger.LOADING_CONFIG);
        this.config = Config.getInstance();
        this.csp_files_dir = config.getConfig("CSP_FILES_DIR");
        this.useTypeChecker = (new Boolean(config.getConfig("TYPE_CHECKER"))).booleanValue();
        this.displayTree = (new Boolean(config.getConfig("DISPLAY_TREE"))).booleanValue();
        Messenger.getInstance().finish(Messenger.LOADING_CONFIG);
    }
    
    public String openFile() throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.OPENING_FILE);
        String file = "";
        String fileName = "";

        int val = 0;
        JFileChooser chooser = new JFileChooser();
        ExampleFileFilter filter = new ExampleFileFilter(HandelC.CSP_EXTENSION);
        chooser.setFileFilter(filter);
        chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
        chooser.setDialogType(JFileChooser.OPEN_DIALOG);
        chooser.setCurrentDirectory(new File(csp_files_dir));

        val = chooser.showDialog(null,Messenger.OPEN);
        if(val == JFileChooser.APPROVE_OPTION){
            fileName = chooser.getSelectedFile().toString();
            Messenger.getInstance().finish(Messenger.OPENING_FILE);
            this.rememberDirectory(fileName);
        } else {
            Messenger.getInstance().cancel(Messenger.OPENING_FILE);
        }
        return fileName;
    }

    public void saveFile(String text) throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.SAVING_OUTPUT);
        String fileName = null;

        boolean choosed = false;
        while (!choosed) {
            fileName = null;
            int val = 0;
            JFileChooser chooser = new JFileChooser();

            ExampleFileFilter filter = new ExampleFileFilter(HandelC.EXTENSION);
            chooser.setFileFilter(filter);
            chooser.setFileSelectionMode(JFileChooser.FILES_AND_DIRECTORIES);
            chooser.setDialogType(JFileChooser.SAVE_DIALOG);
            chooser.setCurrentDirectory(new File(csp_files_dir));

            val = chooser.showDialog(null,Messenger.SAVE);
        
            if(val == JFileChooser.APPROVE_OPTION){
                File file = chooser.getSelectedFile(); 
                fileName = file.toString();
                
                if(!fileName.endsWith(HandelC.DOT+HandelC.EXTENSION)) {
                    fileName = fileName + HandelC.DOT + HandelC.EXTENSION;
                    file = new File(fileName);
                }
                
                if(file.exists()) {
                    int answer = JOptionPane.showConfirmDialog (chooser,
                        Messenger.OVERWRITE_1 + fileName + Messenger.OVERWRITE_2,
                        Messenger.OVERWRITE_TITLE,
                        JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
                    if (answer == JOptionPane.OK_OPTION){
                        choosed = true;
                    }
                } else {
                    choosed = true;
                }
            } else if (val == JFileChooser.CANCEL_OPTION){
                choosed = true;
            }
        }
        if (fileName != null) {
            FileWriter writer = new FileWriter(fileName,false);
            writer.write(text);
            writer.close();
            Messenger.getInstance().finish(Messenger.SAVING_OUTPUT);
        } else {
            Messenger.getInstance().cancel(Messenger.SAVING_OUTPUT);
        }
    }
    
    private void rememberDirectory(String fileName) throws FileNotFoundException, IOException{
        Messenger.getInstance().start(Messenger.STORING_DIRECTORY);
        Properties properties = new Properties();
        
        File f = new File("./config/config_en.properties");
        try{
            FileInputStream propFile = new FileInputStream(f);
            properties.load(propFile);
        } catch (FileNotFoundException e) {
            Messenger.getInstance().error(Messenger.NO_CONFIGURATION_FILE);
        }

        try{
            String directory = fileName.substring(0,fileName.lastIndexOf("\\")).replace('\\','/');
            FileOutputStream propFileOut = new FileOutputStream(f);
            properties.setProperty("CSP_FILES_DIR",directory);
            properties.store(propFileOut,"Configuration File");
            this.csp_files_dir = fileName;
        } catch (FileNotFoundException e) {
            Messenger.getInstance().error(Messenger.NO_CONFIGURATION_FILE);
        }
        Messenger.getInstance().finish(Messenger.STORING_DIRECTORY);
    }
    
    private Start parser(String fileName) throws FileNotFoundException, 
                                                IOException, 
                                                ParserException,
                                                LexerException{

        FileInputStream in = new FileInputStream(fileName);
        InputStreamReader isr = new InputStreamReader(in);
        PushbackReader pb = new PushbackReader(isr, 1024);
        Lexer lexer = new Lexer(pb);
        Parser p = new Parser(lexer);
        Start tree = p.parse();
        return tree;
    }
    
    public String translate(String fileName, boolean withTypeChecker) 
                                            throws FileNotFoundException, 
                                                    IOException,
                                                    ParserException,
                                                    LexerException{
                                                        
        Messenger.getInstance().start(Messenger.PARSING);
        Node start = this.parser(fileName);
        Messenger.getInstance().finish(Messenger.PARSING);
        
        if (withTypeChecker) {
            Messenger.getInstance().start(Messenger.TYPE_CHECKING);
            start.apply(new CspTypeChecker());
            Messenger.getInstance().finish(Messenger.TYPE_CHECKING);
        }

        //////////////////////
        // LOAD DIRECTIVE INFO
        ////////////////////// 
        Messenger.getInstance().start(Messenger.LOADING_DIRECTIVES);
        DirectivesVisitor directives = new DirectivesVisitor();
        start.apply(directives);
        Messenger.getInstance().finish(Messenger.LOADING_DIRECTIVES);

        //////////////////
        // PRE-PROCESSING
        ////////////////// 

        // Constants
        Messenger.getInstance().start(Messenger.LOADING_CONSTANTS);
        ConstantsVisitor constants = new ConstantsVisitor(directives);
        start.apply(constants);
        Messenger.getInstance().finish(Messenger.LOADING_CONSTANTS);
        
        // Processes Arguments and Definitions
        Messenger.getInstance().start(Messenger.LOADING_PROCESSES_INFO);
        ProcessesInfoVisitor procInfo = new ProcessesInfoVisitor();
        start.apply(procInfo);
        Messenger.getInstance().finish(Messenger.LOADING_PROCESSES_INFO);
        
        // Datatypes
        Messenger.getInstance().start(Messenger.LOADING_DATATYPES);
        TypesVisitor datatypes = new TypesVisitor(directives,constants);
        start.apply(datatypes);
        Messenger.getInstance().finish(Messenger.LOADING_DATATYPES);

        // Channels Information
        Messenger.getInstance().start(Messenger.LOADING_CHANNELS_INFO);
        ChannelsVisitor channels = new ChannelsVisitor(datatypes,directives);
        start.apply(channels);
        Messenger.getInstance().finish(Messenger.LOADING_CHANNELS_INFO);
        
        ////////////////////////
        // PRIMARY RESTRICTIONS
        //////////////////////// 
        Messenger.getInstance().start(Messenger.CHECKING_PRIMARY_RESTRICTIONS);
        RestrictionsVisitor restrictions = new RestrictionsVisitor(directives, datatypes);
        start.apply(restrictions);
        Messenger.getInstance().finish(Messenger.CHECKING_PRIMARY_RESTRICTIONS);

        //////////////////////////
        // SECONDARY RESTRICTIONS
        //////////////////////////
        Messenger.getInstance().start(Messenger.CHECKING_SECONDARY_RESTRICTIONS);
        TailRecursionVisitor recursion = new TailRecursionVisitor(procInfo,directives);
        start.apply(recursion);
        Messenger.getInstance().finish(Messenger.CHECKING_SECONDARY_RESTRICTIONS);

        ///////////////////////
        // GENERATE MAIN METHOD
        ///////////////////////
        Messenger.getInstance().start(Messenger.GENERATING_MAIN);

        String tempFileName = fileName+HandelC.TEMPORARY_FILE;

        // Read Original File
        BufferedReader in = new BufferedReader(new FileReader(fileName));
        String original = "";
        String str;
        while ((str = in.readLine()) != null) {
            original = original + str + "\n";
        }
        in.close();

        // Write Temporary File (original file + main process)
        Messenger.getInstance().start(Messenger.CREATING_TEMP_FILE);
        String mainProc = HandelC.MAIN + HandelC.ASSIGNMENT + directives.getMain();
        String contents = original + mainProc;
        File file = new File(tempFileName);
        file.deleteOnExit();
        FileWriter writer = new FileWriter(file);
        writer.write(contents);
        writer.close();
        Messenger.getInstance().finish(Messenger.CREATING_TEMP_FILE);

        // Parser Temporary File
        Messenger.getInstance().start(Messenger.PARSING_TEMP_FILE);
        Node mainStart = this.parser(tempFileName);
        Messenger.getInstance().finish(Messenger.PARSING_TEMP_FILE);

        // Type Check Temporary File
        if (withTypeChecker) {
            Messenger.getInstance().start(Messenger.TYPE_CHECKING_TEMP_FILE);
            mainStart.apply(new CspTypeChecker());
            Messenger.getInstance().finish(Messenger.TYPE_CHECKING_TEMP_FILE);
        }

        // Translate Main Method
        ProcessesTranslatorVisitor mainTranslator = 
            new ProcessesTranslatorVisitor(procInfo,directives,datatypes,
                                            recursion,channels,constants);
        
        MainMethodVisitor mainVisitor = new MainMethodVisitor(mainTranslator,recursion.getMutualRecursion());
        mainStart.apply(mainVisitor);
        
        Messenger.getInstance().finish(Messenger.GENERATING_MAIN);

        //////////////////
        // TRANSLATE
        ////////////////// 
        String code = "";
        
        // Processes
        Messenger.getInstance().start(Messenger.TRANSLATING_PROCESSES);
        ProcessesTranslatorVisitor translator = 
            new ProcessesTranslatorVisitor(procInfo,directives,datatypes,
                                            recursion,channels,constants);
        
        // As we do not have parallel in the specification, we are allowed to consider
        // that the value of the parallelism counter will not be altered in
        // the translation of the specification
        translator.setParallelismCounter(mainVisitor.getParallelismCounter());
        start.apply(translator);
        Messenger.getInstance().finish(Messenger.TRANSLATING_PROCESSES);

        //////////////////
        // FORMAT CODE
        ////////////////// 
        String finalCode = translator.getHcCode() + mainVisitor.getHcCode();

        Messenger.getInstance().start(Messenger.FORMATTING_OUTPUT);
        code = CodeFormatting.format(finalCode);
        Messenger.getInstance().finish(Messenger.FORMATTING_OUTPUT);

        return code;
    }
    
    public void displayTree(String fileName) throws FileNotFoundException, 
                                                    IOException,
                                                    ParserException,
                                                    LexerException{
            JFrame frame = new JFrame("CSP Syntax Trees");
            TreeView treeViewCst = new TreeView();
            JScrollPane scrollCst = new JScrollPane(treeViewCst);

            TreeView treeViewAst = new TreeView();
            JScrollPane scrollAst = new JScrollPane(treeViewAst);

            treeViewCst.setModel(new DefaultTreeModel(
            		new DefaultMutableTreeNode("Constructing the tree.")));
            treeViewAst.setModel(new DefaultTreeModel(
            		new DefaultMutableTreeNode("Constructing the tree.")));

            JTabbedPane tabs = new JTabbedPane();
            tabs.addTab("CST - Concrete Syntax Tree", scrollCst);
            tabs.addTab("AST - Abstract Syntax Tree", scrollAst);

            frame.getContentPane().add(tabs);

            frame.setDefaultCloseOperation(JFrame.EXIT_ON_CLOSE);
    		frame.setSize(800, 400);
            frame.setVisible(true);

            try {
                Start tree = this.parser(fileName);
                tree.apply(new CspTypeChecker());
                tree.apply(new CspCstFactory(treeViewCst));
                tree.apply(new CspAstFactory(treeViewAst));
            } catch (Exception e) {
                treeViewCst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                treeViewAst.setRoot(new DefaultMutableTreeNode(e.getMessage()));
                e.printStackTrace();
            }
    }
    
    private boolean askForNewTranslation(){
        int answer = JOptionPane.showConfirmDialog (null,
            Messenger.NEW_TRANSLATION,
            Messenger.NEW_TRANSLATION_TITLE,
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        return (answer == JOptionPane.OK_OPTION);
    }

    private boolean askTryAgain(){
        int answer = JOptionPane.showConfirmDialog (null,
            Messenger.TRY_AGAIN,
            Messenger.TRY_AGAIN_TITLE,
            JOptionPane.YES_NO_OPTION, JOptionPane.QUESTION_MESSAGE);
        return (answer == JOptionPane.OK_OPTION);
    }

    public boolean showKnowException(Throwable e) {
        Messenger.getInstance().error(e.getMessage());
        e.printStackTrace();
        return askTryAgain();
    }

    public boolean showUnknowException(Throwable e) {
        Messenger.getInstance().uerror(e.getMessage());
        e.printStackTrace();
        return askTryAgain();
    }
    
    public static void main (String[] argv) {
        boolean keepGoing = true;
        Csp2HC translator = new Csp2HC();
        while(keepGoing) {
            try {
                String fileName = translator.openFile();
                if (!fileName.equals("")){
                    String message = Messenger.TRANSLATING_FILE + fileName;
                    Messenger.getInstance().addComment(message);
                    String hcCode = translator.translate(fileName,translator.getUseTypeChecker());
                    translator.saveFile(hcCode);
                    if(translator.getDisplayTree()) {
                        translator.displayTree(fileName);
                    }
                    keepGoing = translator.askForNewTranslation();
                } else {
                    keepGoing = false;
                }
            } catch (FileNotFoundException e) {
                keepGoing = translator.showKnowException(e);
            } catch (IOException e) {
                keepGoing = translator.showKnowException(e);
            } catch (ParserException e) {
                keepGoing = translator.showKnowException(e);
            } catch (LexerException e) {
                keepGoing = translator.showKnowException(e);
            } catch (CspSemanticException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UnsupportedConstructorException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UndefinedChannelInOutException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UndefinedArgumentTypeException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UndefinedChannelTypeException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UndefinedIntBitsException e) {
                keepGoing = translator.showKnowException(e);
            } catch (WrongNumberOfArgumentsException e) {
                keepGoing = translator.showKnowException(e);
            } catch (InvalidDefinitionException e) {
                keepGoing = translator.showKnowException(e);
            } catch (InvalidDirectiveException e) {
                keepGoing = translator.showKnowException(e);
            } catch (InconsistentChannelProjectionUseException e) {
                keepGoing = translator.showKnowException(e);
            } catch (RepeatedConstantException e) {
                keepGoing = translator.showKnowException(e);
            } catch (UndefinedNameReferenceException e) {
                keepGoing = translator.showKnowException(e);
            } catch (RuntimeException e) {
                keepGoing = translator.showUnknowException(e);
            }
        }
        Messenger.getInstance().message(Messenger.PROGRAM_TERMINATED);
    }
}
