package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ChannelsVisitor extends ExtendedDepthFirstAdapter {

    // Type of the channels
    // Name:String -> Types:PType Vector
    private Hashtable channelTypesMap;
    
    // Indicate which channels communicate values using projections c.*
    private List projChannels; 

    // Indicate which channels communicate values using either ! or ?
    // String Vector
    private List commChannels; 

    // Indicates to the visitor that we are checking the names of
    // the channels being declared
    private boolean checkingChannelNames;

    // Stores the channel names of the current channel definition
    private List channelNames;
    
    // Indicates to the visitor that we are checking the type of
    // the channels being declared
    private boolean checkingChannelTypes;

    // String containing the type cardinality
    private String typeCard;
    
    // String containing the type handel-c equivalent
    private String typeHandelC;
    
    // Stores the channel types of the current channel definition
    private List channelTypes;
    
    // Visitor containing the information about the datatype
    private TypesVisitor datatypes;

    // Visitor containing the information about the directives
    private DirectivesVisitor directives;
    
    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Current Channel Definition
    private ACspChannelDefinition currentChannelDefinition;
    
    public ChannelsVisitor(TypesVisitor datatypes, DirectivesVisitor directives) {
        this.channelTypesMap = new Hashtable();
        this.commChannels = new Vector();
        this.projChannels = new Vector();
        
        this.datatypes = datatypes;
        this.directives = directives;
        this.typeCard = "";
        this.typeHandelC = "";
    }
    
    public boolean isCommChannel(String channelName){
        return this.commChannels.contains(channelName);
    }
    public boolean isProjChannel(String channelName){
        return this.projChannels.contains(channelName);
    }

    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        PExpression expression = definition.getExpression();
        expression.apply(this);
    }

    public void caseACspChannelDefinition(ACspChannelDefinition node){
        this.currentChannelDefinition = node;
        
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        // Retrieve Channel Names
        Vector names = new Vector();
        
        this.checkingChannelNames = true;
        this.channelNames = new Vector();
        channels.apply(this);
        this.checkingChannelNames = false;
        
        this.checkingChannelTypes = true;
        this.channelTypes = new Vector();
        if (type != null) {
            type.apply(this);
        }
        this.checkingChannelTypes = false;
        
        for(int i=0; i < this.channelNames.size(); i++) {
            this.channelTypesMap.put(this.channelNames.get(i),this.channelTypes);
        }
    }
    
    public void caseACspChannel(ACspChannel node){
        this.channelNames.add(node.getName().toString().trim());
    }

    public void caseACspReferencedType(ACspReferencedType node){
        if (this.checkingChannelTypes){
            this.channelTypes.add(node);
        }
        this.typeCard = node.toString().trim() + HandelC.TYPE_CARDINALITY;
        this.typeHandelC = node.toString().trim();
    }
    
    public void caseACspIntType(ACspIntType node){
        if (this.checkingChannelTypes){
            this.channelTypes.add(node);
        }
        this.typeCard = HandelC.INTEGERL + HandelC.TYPE_CARDINALITY;
        this.typeHandelC = HandelC.INTEGERL;
    }
    
    public void caseACspBoolType(ACspBoolType node){
        if (this.checkingChannelTypes){
            this.channelTypes.add(node);
        }
        this.typeCard = HandelC.BOOLEAN + HandelC.TYPE_CARDINALITY;
        this.typeHandelC = HandelC.BOOLEAN;
    }

    public void caseACspTupleType(ACspTupleType node){
        if (this.checkingChannelTypes){
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentChannelDefinition.toString());
        }
    }

    public void caseACspSetType(ACspSetType node){
        if (this.checkingChannelTypes){
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentChannelDefinition.toString());
        }
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        
        PEvent event = node.getEvent();
        LinkedList comm = node.getComunication();
        PPrefixProcess proc = node.getProcess();
        PProjection projection = ((ACspEvent)event).getProjection();

        String channelName = ((ACspEvent)event).getChannelName().toString().trim();
        
        // Checking Projections
        if (projection != null) {
            if (!this.projChannels.contains(channelName)) {
                this.projChannels.add(channelName);
            }    
        }

        // Checking communications
        ListIterator iterator = comm.listIterator(0);
        boolean isCommunication = false;
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if ((next instanceof ACspInputComunication) 
                    || (next instanceof ACspOutputComunication)) {
                if (!this.commChannels.contains(channelName)) {
                    this.commChannels.add(channelName);
                }    
            }
        }
        proc.apply(this);
    }

    // Returns the dimension of an array of channels
    // This is the number of type expression used in the channel declaration
    // (minus one if this is a communication channel)
    private int getArrayDimension(String channelName) {
        int size = this.getChannelType(channelName).size();
        if (size > 0 && this.isCommChannel(channelName)) {
            size = size - 1;
        }
        return size;
    }

    public boolean containsChannelType(String channelName){
        return this.channelTypesMap.containsKey(channelName);
    }
    
    public List getChannelType(String channelName){
        return (List)this.channelTypesMap.get(channelName);
    }

    public String getChannelCommType(String chan) {
        String commType = "";
        List types = this.getChannelType(chan);
        int arrayDim = this.getArrayDimension(chan);
        if(types.size() == arrayDim) {
            // There is no actual value being communicated. Type of communication is SYNC
             commType= HandelC.SYNC_TYPE;
        } else {
            // There is a value being communicated. Type of communication is the last type
            PType type = (PType)ListUtil.last(types);
            type.apply(this);
            commType = this.typeHandelC;
        }
        
        return commType;
    }
    
    public String getChannelDeclaration(String chan) {
        String decl = "";
        decl = decl + HandelC.CHANNEL + HandelC.SPACE;
        
        // Declaring the value communicated via this channel
        if (this.containsChannelType(chan)) {
            List types = this.getChannelType(chan);
            int arrayDim = this.getArrayDimension(chan);

            if(types.size() == arrayDim) {
                // There is no actual value being communicated. Type of communication is SYNC
                decl = decl + HandelC.SYNC_TYPE;
            } else {
                // There is a value being communicated. Type of communication is the last type
                PType type = (PType)ListUtil.last(types);
                decl = decl + this.getChannelCommType(chan);
            }
            decl = decl + HandelC.SPACE + chan;
            
            // Declaring possible arrays
            
            if(this.isProjChannel(chan)){
                for (int i = 0; i < arrayDim; i++) {
                    PType type = (PType)types.get(i);
                    decl = decl + HandelC.ARRAY_BEGIN;
                    type.apply(this);
                    decl = decl + this.typeCard;
                    decl = decl + HandelC.ARRAY_END;
                }
            }
        }
        return decl;
    }

}

