package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ExpressionsVisitor extends ExtendedDepthFirstAdapter {

    // Visitor that has the information about the directies
    protected DirectivesVisitor directives;

    // Resulting handel-c code
    protected String hcCode = "";

    // Current definition being read
    protected String currentDefinition;

    // Indicates that we are analysing a set
    boolean isSet;

    // Indicates that we are analysing an array index
    boolean isArrayIndex;
    
    // Indicates if this is being used to translate macro expressions
    boolean isMacroExpression;
    
    public ExpressionsVisitor(DirectivesVisitor directives, boolean isMacroExpression) {
        this.directives = directives;
        this.currentDefinition = "";
        this.hcCode = "";
        
        this.isSet = false;
        this.isArrayIndex = false;
        this.isMacroExpression = isMacroExpression;
    }
    
    public String getHcCode() { return this.hcCode; }

    public void setCurrentDefinition(String def){ 
        this.currentDefinition = def; 
    }
    
    // Booleans

    public void caseACspBooleanValue(ACspBooleanValue node) {
        this.hcCode = this.hcCode + node.toString();
    }

    public void caseACspAndLogicalExpr(ACspAndLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.LOGICAL_AND + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }

    public void caseACspOrLogicalExpr(ACspOrLogicalExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.LOGICAL_OR + HandelC.SPACE;
        node.getRightExpr().apply(this);
    }
    
    public void caseACspNotPrefixExpr(ACspNotPrefixExpr node) {
        this.hcCode = this.hcCode + HandelC.LOGICAL_NOT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    // Numbers 

    public void caseACspNumberValue(ACspNumberValue node) {
        this.hcCode = this.hcCode + node.toString().trim();
    }

    public void caseACspAddMathExpr(ACspAddMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_ADD;
        node.getRightExpr().apply(this);
    }

    public void caseACspDiffMathExpr(ACspDiffMathExpr node) {
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.MATH_DIFF;
        node.getRightExpr().apply(this);
    }

    public void caseACspMulMath1Expr(ACspMulMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_MULT;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspDivMath1Expr(ACspDivMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_DIV;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspRemainderMath1Expr(ACspRemainderMath1Expr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.MATH_REMAINDER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspMinusPrefixExpr(ACspMinusPrefixExpr node) {
        this.hcCode = this.hcCode + HandelC.MATH_NEG;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getExpression().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    // Relations
    public void caseACspEqualRelationalExpr(ACspEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspNotEqualRelationalExpr(ACspNotEqualRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_NOT_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGtRelationalExpr(ACspGtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspGteRelationalExpr(ACspGteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_GREATER_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void caseACspLtRelationalExpr(ACspLtRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }

    public void caseACspLteRelationalExpr(ACspLteRelationalExpr node) {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getLeftExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.REL_LESS_EQUALS;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        node.getRightExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    
    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
    }

    // Sets
    public void caseACspSetValue(ACspSetValue node) {
        this.isSet = true;
        node.getSet().apply(this);
        this.isSet = false;
    }

    public void caseACspEmptySet(ACspEmptySet node) {
        if (this.directives.containsEmptySetType(this.currentDefinition)) {
            String type = (String)this.directives.getEmptySetType(this.currentDefinition);
            this.hcCode = this.hcCode + type + HandelC.EMPTY_SET_SUFFIX + HandelC.SET_SUFFIX;
        } else {
            throw new UndefinedEmptySetTypeException(this.currentDefinition);
        }
    }
    
    // Sequences
    public void caseACspSequenceValue(ACspSequenceValue node) {
    }

    
    public void caseACspExprArgument(ACspExprArgument node) {
        String macro = HandelC.SINGLETON + HandelC.SET_SUFFIX + HandelC.UNDERSCORE + HandelC.MACRO;
        if(this.isSet) {
            this.hcCode = this.hcCode + macro + HandelC.PAR_BEGIN;
        }
        node.getExpression().apply(this);
        if(this.isSet) {
            this.hcCode = this.hcCode + HandelC.PAR_END;
        }

    }

    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node) {
        PArgument argument = node.getArgument();
        argument.apply(this);
        
        PArgumentList next = node.getNext();
        if(this.isSet) {
            if (next != null){
                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_BITWISE_OR + HandelC.SPACE;
                next.apply(this);
            }
        }
    }
    
    public void caseACspCallPostfixExpr(ACspCallPostfixExpr node){
        this.hcCode = this.hcCode + node.toString().trim();
    }
    
    public void caseACspProductionsSet(ACspProductionsSet node){
        this.hcCode = this.hcCode + HandelC.CHANNELS + HandelC.EMPTY_SET_SUFFIX + HandelC.SET_SUFFIX;
    }
    
    public void caseACspIfExpr(ACspIfExpr node){
        if (this.isMacroExpression) {
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getIfBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_COND + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getThenBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.CONDITIONAL_SEP + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getElseBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.PAR_END;
        } else {
            this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
            node.getIfBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
            node.getThenBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.ELSE + HandelC.SPACE + HandelC.BLOCK_BEGIN;
            node.getElseBranch().apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.SPACE;
        }
    }

}
