package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ProcessesTranslatorVisitor extends ExpressionsVisitor {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the types
    private TypesVisitor types;

    // Visitor that has the information about recursion
    private TailRecursionVisitor recursion;

    // Visitor that has the information about the channels
    private ChannelsVisitor channels;

    // Visitor that has the information about the constants
    private ConstantsVisitor constants;

    // Current prefixing being read
    private String currentPrefixing;

    // Current List of the remaining types for the current prefixing
    private List currentPrefixingTypes;

    // Current type for the current prefixing
    private String currentType;
    
    // Identity of Processes usedin the translation of Mutual Recursion
    private List processesIds;
    
    // List of current input variables in context
    private List inputVariables;

    // Indicates if the translator is translating a channel projection
    private boolean isProjection;

    // Stack of number of arguments to be translated for each tag
    // Changes the values according to the translation
    // Stack of int
    private Stack tagsArgs;

    // Stack of number of arguments to be translated for each tag
    // Never changes the values (either push or pop)
    // Stack of int
    private Stack tags;

    // Stack of types of arguments to be translated for each tag
    // Stack of (List of String)
    private Stack tagsTypes;
    
    // Number of parallel operators
    private int parallelismCounter;
    
    // Current branch in the parallel composition in the current process
    // Used to indicated which element of the array of local copies
    // and auxiliary variables must be accessed in the case of a mutual recursion
    // with a parallel composition in the main method
    private int currentParallelBranchInProcess;

    // Stack of sets of events whose interleaving must be forced
    private Stack forcedInterleaving;

    // Set of channel semaphors used
    private Set usedSemaphors;

    private static boolean IS_ARGUMENT = true;
    private static boolean IS_NOT_ARGUMENT = false;
    private static boolean IS_DECLARATION = true;
    private static boolean IS_NOT_DECLARATION = false;
    private static boolean DECLARE_INPUT_VAR = true;
    private static boolean DO_NOT_DECLARE_INPUT_VAR = false;
    
    public ProcessesTranslatorVisitor(ProcessesInfoVisitor procsInfo, 
                                        DirectivesVisitor directives, 
                                        TypesVisitor types, 
                                        TailRecursionVisitor recursion,
                                        ChannelsVisitor channels,
                                        ConstantsVisitor constants) {
        
        super(directives,false);
        
        this.procsInfo = procsInfo;
        this.types = types;
        this.recursion = recursion;
        this.channels = channels;
        this.constants = constants;
        
        this.currentPrefixing = "";
        this.currentPrefixingTypes = new Vector();
        this.currentType = "";
        
        // Creating Ids for each process
        this.processesIds = new Vector();
        Enumeration keys = this.procsInfo.keys();
        while(keys.hasMoreElements()) {
            this.processesIds.add(keys.nextElement());
        }
        
        this.inputVariables = new Vector();
        this.isProjection = false;
        this.parallelismCounter = 0;
        this.currentParallelBranchInProcess = 0;
        this.tagsArgs = new Stack();
        this.tags = new Stack();
        this.tagsTypes = new Stack();
        this.forcedInterleaving = new Stack();
        this.usedSemaphors = new HashSet();
    }

    public int getParallelismCounter(){ return this.parallelismCounter; }
    public void setParallelismCounter(int i){ this.parallelismCounter = i; }

    public void setCurrentDefinition(String def){ 
        super.setCurrentDefinition(def); 
        this.currentParallelBranchInProcess = 0;
    }
    
    public void caseStart(Start node){
        if (!this.recursion.getMutualRecursion()) {
            node.getPSpecification().apply(this);
        } else {
            this.translateProcessesDefinitionsMutualRecursion();
        }
        
        // Preamble of file
        if (this.recursion.getMutualRecursion()) {
            this.declareMutualRecConstants();
        } else {
            this.declareProcessesPrototypes();
        }
        this.declareAuxiliaryMacros();
        this.declareChannels();
        this.declareConstants();
        this.declareDatatypes();
        this.declareAuxiliaryConstants();
        this.declareClock();
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        if (!this.recursion.getMutualRecursion()) {
            TName name = definition.getName();
            ACspParameters parameters = (ACspParameters)definition.getParameters();
            PExpression expression = definition.getExpression();

            this.setCurrentDefinition(name.toString().trim());
            if (parameters == null) {
                parameters = new ACspParameters();
            }
            this.hcCode = this.hcCode + HandelC.generateComment(Messenger.PROCESS + HandelC.SPACE + name);
            this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + name;
            this.hcCode = this.hcCode + this.translateFormalArguments(parameters,this.currentDefinition);
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

            if(this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
                // It is a tail recursion

                // int 1 KEEP_LOOPING;
                this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
                this.hcCode = this.hcCode + this.getVariableArray(HandelC.KEEP_LOOPING,IS_DECLARATION);
                this.hcCode = this.hcCode + HandelC.END_COMMAND;
                
                // Declare argument copies
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_DECLARATION,IS_NOT_ARGUMENT);
                
                // Initialise them
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_NOT_DECLARATION,IS_NOT_ARGUMENT);
                
                // KEEP_LOOPING = true;
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
                
                // Loop
                this.hcCode = this.hcCode + HandelC.WHILE ;
                this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.KEEP_LOOPING + HandelC.PAR_END;
                this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
                this.hcCode = this.hcCode + HandelC.SEQUENCE + HandelC.BLOCK_BEGIN;
                expression.apply(this);
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
            } else {
                expression.apply(this);
            }
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
        }
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
    }

    public void caseACspSkipProcess(ACspSkipProcess node){
    }
    public void caseACspStopProcess(ACspStopProcess node){
        this.hcCode = this.hcCode + HandelC.WHILE 
                                    + HandelC.PAR_BEGIN
                                    + HandelC.LOGICAL_TRUE
                                    + HandelC.PAR_END
                                    + HandelC.BLOCK_BEGIN
                                    + HandelC.BLOCK_END;
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
        this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        PEvent event = node.getEvent();
        LinkedList comm = node.getComunication();
        PPrefixProcess proc = node.getProcess();
        
        String inputVar = this.translatePrefixWithForcedSync(event,comm,DECLARE_INPUT_VAR);
        
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        proc.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;

        if (inputVar != null) {
            this.inputVariables.remove(inputVar);
        }

    }

    public void caseACspEvent(ACspEvent node){
        String chan = node.getChannelName().toString().trim();

        this.hcCode = this.hcCode + chan;

        int startingIndex = this.hcCode.length();
        
        if (node.getProjection() != null) {
            this.isProjection = true;
            this.currentType = "";
            node.getProjection().apply(this);

            this.closeArrayAccessChannelSync();
            
            this.isProjection = false;
        }
        
    }
    
    public void caseACspProjection(ACspProjection node) {
        if (this.tags.empty()) {
            // this is a real projection
            this.nextArrayAccessChannelSync();

            node.getValue().apply(this);
            
        } else {
            // this may be a tag access
            // Check how many arguments left for the current tag
            int argsLeft = ((Integer)this.tagsArgs.pop()).intValue();
            // If it is zero, we should close the parenthesis
            // and remove the top of the stack of total number of arguments
            if (argsLeft == 0) {
                // For every finished argument of a datatype tag
                while(argsLeft == 0) {
                    // Close the array invocation
                    this.closeArrayAccessTypeConstructor();

                    // If there are still arguments left
                    if (!tagsArgs.isEmpty()) {
                        // We have finished one tag constructor, remove it from the top of the stack
                        this.tags.pop();
                        
                        // Let us check the next one
                        argsLeft = ((Integer)this.tagsArgs.pop()).intValue();
                    } else {
                        // There are no arguments left, indicate this by setting 
                        // argsLeft to a negative number
                        argsLeft = -1;
                    }
                }

                // Furthermore, If there are any further tags being analysed
                // and there are yet arguments left
                if (!this.tags.isEmpty() && argsLeft > 0) {
                    // we have done one more argument from the previous tag
                    // Reduce the top of the stack of analysed arguments by one
                    this.tagsArgs.push(new Integer(argsLeft-1));
                } else if (argsLeft < 0) {
                    // Otherwise, if we finished the all the arguments, just remove the 
                    // last tag from the tags stack
                    this.tags.pop();
                }
            } else {
                // If it is not zero, we do not need to close any parenthesis
                // But we have done one more argument from the current tag
                // Reduce the top of the stack of analysed arguments by one
                this.tagsArgs.push(new Integer(argsLeft - 1));
            }
            
            // If we are analysing tags
            if (!this.tags.isEmpty()) {
                int argsTotal = ((Integer)this.tags.peek()).intValue();
                
                // And there are still other arguments to be analysed
                if (argsLeft < argsTotal) {
                    
                    // we Should include a argument separator (comma)
                    //this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;

                    this.nextArrayAccessTypeConstructor();
                }
            }

            // If we are not analysing any tags anymore
            // this is actually a real projection
            if (this.tags.empty()) {
                this.nextArrayAccessChannelSync();
            }
            
            // Analyse the remaining of the projection
            node.getValue().apply(this);

            // After analysing the projection, let us check if we have finished
            // any tag analysis
            
            // If there are tags to be analysed yet
            if (!this.tagsArgs.isEmpty()) {
                // For every tag whose analysys has finished
                while (!this.tagsArgs.isEmpty() && ((Integer)this.tagsArgs.pop()).intValue() == 0) {
                    // Remove the corresponding value from the stack of total 
                    // number of arguments, and close its parenthesis
                    this.tags.pop();
                    //this.hcCode = this.hcCode + HandelC.PAR_END;
                    
                    this.closeArrayAccessTypeConstructor();
                    
                    // If there are still tags being analysed
                    // Reduce the top value by one, because we just finished one argument
                    if (!this.tagsArgs.isEmpty()) {
                        int temp = ((Integer)this.tagsArgs.pop()).intValue();
                        this.tagsArgs.push(new Integer(temp-1));
                    }
                }
            }
        }
    }
    
    
    public void caseACspInputComunication(ACspInputComunication node) {
        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
        this.hcCode = this.hcCode + HandelC.SPACE; 
        this.translateExpression(node.getExpression());
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,Directives.CHANNEL_IN); 
    }
    
    public void caseACspOutputComunication(ACspOutputComunication node) {
        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
        this.hcCode = this.hcCode + HandelC.SPACE;
        this.translateExpression(node.getExpression());
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,Directives.CHANNEL_OUT); 
    }

    public void caseACspCallPostfixExpr(ACspCallPostfixExpr node){
        String name = ((ACspCallPostfixExpr)node).getName().toString().trim();
        
        // Is it a Process invocation
        
        if (this.procsInfo.containsProcArguments(name)){
            ACspArguments arguments = (ACspArguments)node.getArguments();
            if (!this.recursion.getMutualRecursion()) {
                // It is not mutual recursion, behave normally
                if(name.equals(this.currentDefinition) && 
                    this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
                    // It is a tail recursion invocation
                    // Save values of the copies
                    this.translateRealArgumentsCopies(name,arguments,this.currentDefinition);
                    this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
                } else {
                    this.hcCode = this.hcCode + name;
                    this.translateRealArguments(name,arguments,this.currentDefinition);
                    this.hcCode = this.hcCode + HandelC.END_COMMAND;
                }
            } else {
                // It is a mutual recursion

                // Save values of the copies
                this.translateRealArgumentsCopies(name,arguments,this.currentDefinition);

                // Change the value of the program counter
                if (!this.isTranslatingMain()) {
                    this.hcCode = this.hcCode + this.getVariableArray(HandelC.PROGRAM_COUNTER,IS_NOT_DECLARATION);
                    this.hcCode = this.hcCode + HandelC.SPACE;
                    this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE;
                    this.hcCode = this.hcCode + name + HandelC.END_COMMAND;

                    // Indicate the change by setting KEEP_LOOPING to true
                    this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
                } else {
                    // Translating the main method Invoke MUTUAL_REC
                    this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION + HandelC.PAR_BEGIN + name;
                    if (this.needsLocalVariablesAsArrays()) {
                        this.hcCode = this.hcCode + this.getVariableArrayIndex();
                    }
                    this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.END_COMMAND;
                }
            }
        } else if (this.inputVariables.contains(name)) {
            // It is a input variable, do not change it
            this.hcCode = this.hcCode + name;
        } else if (this.procsInfo.isArgumentOf(name,this.currentDefinition)) {
            // It is an argument of this process
            
            if (this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)
                || this.recursion.getMutualRecursion()) {
                // We are translating a tail recursive process or a mutual recursion
                // Replace this argument by its copy
                this.hcCode = this.hcCode + this.copyName(this.currentDefinition,name,IS_NOT_DECLARATION);
            } else {
                // Keep original name
                this.hcCode = this.hcCode + name;
            }
        } else if (this.types.isATypeBranch(name)) {
            // this is a reference to an element of a datatype
            this.hcCode = this.hcCode + name;
        } else if (this.types.isTag(name)) {
            // this is a reference to a tag of a datatype
            
            String parent = this.types.getTagParent(name);
            String lookupTableName = this.types.getLookupTableName(parent,name);
            
            Integer tagSize = new Integer(this.types.getTagValues(name).size());
            
            this.tagsArgs.push(tagSize);
            this.tags.push(tagSize);
            this.tagsTypes.push((List)this.types.getTagTypes(name));
            
            this.hcCode = this.hcCode + lookupTableName;
            this.openArrayAccessTypeConstructor();
        } else if (this.constants.containsConstant(name)) {
            // this is a constant
            this.hcCode = this.hcCode + name;
            // it might have arguments, so...
            if (node.getArguments() != null) {
                this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
                node.getArguments().apply(this);
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }

        } else {
            throw new UndefinedNameReferenceException(name,this.currentDefinition);
        }
        
        // this might as well be a projection, so...
        if (node.getProjection() != null) {
            node.getProjection().apply(this);
        }

    }
    
    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node) {
        node.getArgument().apply(this);
        this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
        node.getNext().apply(this);
    }

    public void caseACspExprArgument(ACspExprArgument node) {
        this.hcCode = this.hcCode + node.getExpression().toString().trim();
    }
    

    private String keepLooping(String value) {
        String answer = this.getVariableArray(HandelC.KEEP_LOOPING,IS_NOT_DECLARATION) + HandelC.SPACE;
        answer = answer + HandelC.ASSIGNMENT + HandelC.SPACE + value + HandelC.END_COMMAND;
        
        return answer;
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();
        
        // Checking the synchronisation channel set
        PExpression syncExp = ((ACspSharingParallelProcess)node).getExpression();
        UsedChannelsVisitor usedChannels = new UsedChannelsVisitor(this.procsInfo);
    	syncExp.apply(usedChannels);

        this.translateConcurrency(left,right,this.getForcedInterleavedEvents(left, right, usedChannels.getUsedChannels()));
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        Node left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
        Node right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();

        // Checking the synchronisation channel set
        
        PExpression syncExpLeft = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
        UsedChannelsVisitor usedChannelsLeft = new UsedChannelsVisitor(this.procsInfo);
        syncExpLeft.apply(usedChannelsLeft);

        PExpression syncExpRight = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
        UsedChannelsVisitor usedChannelsRight = new UsedChannelsVisitor(this.procsInfo);
        syncExpRight.apply(usedChannelsRight);
        
        // Intersection
        Set leftSet = usedChannelsLeft.getUsedChannels();
        Set rightSet = usedChannelsRight.getUsedChannels();
        leftSet.retainAll(rightSet);
        
        this.translateConcurrency(left,right,this.getForcedInterleavedEvents(left, right, leftSet));
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        Node left = ((ACspInterleaveProcess)node).getLeftProcess();
        Node right = ((ACspInterleaveProcess)node).getRightProcess();
        this.translateConcurrency(left,right,this.getForcedInterleavedEvents(left, right, new HashSet()));
    }
    
    private Set getForcedInterleavedEvents(Node left, Node right, Set sync){
        UsedChannelsVisitor usedChannels = new UsedChannelsVisitor(this.procsInfo);
        
        left.apply(usedChannels);
        Set leftUsedChannels = usedChannels.getUsedChannels();
        usedChannels.resetUsedChannels();
        right.apply(usedChannels);
        Set rightUsedChannels = usedChannels.getUsedChannels();
        usedChannels.resetUsedChannels();
        
        rightUsedChannels.retainAll(leftUsedChannels); // rightUsedChannels intersection leftUsedChannels
        rightUsedChannels.removeAll(sync); // rightUSedChannels minus sync
        
        return rightUsedChannels;
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);

        List keys = initialsExtractor.keyList();

        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);
            this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
            this.declareInputVar(comm);
        }
        
        // Declare the channel set that stores the events that happened
    	int nextId = MathUtil.getNextInt();
    	String controlVar = this.getForcedInterleavingControlVarName(nextId);
    	this.hcCode = this.hcCode + types.getTypeSetOf(HandelC.CHANNEL_TYPE)
					+ HandelC.SPACE + controlVar + HandelC.SPACE + HandelC.END_COMMAND;
    	this.hcCode = this.hcCode + controlVar + HandelC.SPACE 
					+ HandelC. ASSIGNMENT + HandelC.SPACE 
					+ types.getEmptySetOf(HandelC.CHANNEL_TYPE)
					+ HandelC.END_COMMAND;

    	this.translateExternalChoice(initialsExtractor,controlVar);
    }

    private void translateExternalChoice(InitialsExtractorVisitor initials, String controlVar) {
    	List<String> cnames = initials.getKeyNames();
    	System.out.println("ProcessesTranslatorVisitor :: buildExternalChoice -> "+cnames);

    	List semaphorsTaken = new Vector<String>();
    	for (int i=0; i<cnames.size(); i++) {
    		this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_CHANNEL_SET
    					+ HandelC.PAR_BEGIN + controlVar + HandelC.PAR_END + HandelC.PAR_END
    					+ HandelC.BLOCK_BEGIN;
    			
    		this.hcCode = this.hcCode + HandelC.WHILE + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_CHANNEL_SET
    				+ HandelC.PAR_BEGIN + controlVar + HandelC.PAR_END + HandelC.SPACE
    				+ HandelC.LOGICAL_AND + HandelC.SPACE 
    				+ this.getTrySema(cnames.get(i)) + HandelC.PAR_END + HandelC.BLOCK_BEGIN;

    		semaphorsTaken.add(cnames.get(i));
    		this.translateCases(initials,
    							new Vector(),
    							0,
    							controlVar,semaphorsTaken);
    				
    		this.hcCode = this.hcCode + HandelC.BLOCK_END;
    		this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    }

    private void translateCases(InitialsExtractorVisitor initials, 
    							List unblockedEvents,
    							int checked,
    							String controlVar, 
    							List semaphorsTaken) {
    	if (!(checked == initials.getKeyNames().size())) {
			this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN + HandelC.CHANNEL_MEMBER
				+ HandelC.PAR_BEGIN + initials.getKeyNames().get(checked)
				+ HandelC.SEP_ARGUMENTS + HandelC.FORCED_SYNC
				+ HandelC.PAR_END + HandelC.PAR_END
				+ HandelC.BLOCK_BEGIN;
			
			this.translateCases(initials, unblockedEvents,
								checked+1, controlVar, semaphorsTaken); 
			
			this.hcCode = this.hcCode + HandelC.BLOCK_END;
			
			this.hcCode = this.hcCode + HandelC.ELSE + HandelC.BLOCK_BEGIN;

			Object currentEvent = initials.keyList().get(checked); 
			unblockedEvents.add(currentEvent);
			this.translateCases(initials, unblockedEvents,
								checked+1, controlVar, semaphorsTaken); 
			
			this.hcCode = this.hcCode + HandelC.BLOCK_END;

			unblockedEvents.remove(currentEvent);
			checked--;
    	} else {
        	// Translate the cases
            this.hcCode = this.hcCode + HandelC.ALTERNATION;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            
            // Declare the cases
            for(int i=0; i < unblockedEvents.size(); i++) {
                this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                PEvent event = (PEvent)unblockedEvents.get(i);
                LinkedList comm = initials.getCommunication(event);
                
                this.translatePrefix(event,comm,DO_NOT_DECLARE_INPUT_VAR);
                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;

            	// Adds the channel to the set of channels that happened
            	this.hcCode = this.hcCode + HandelC.SET_UNION
            				+ HandelC.PAR_BEGIN
            				+ controlVar
            				+ HandelC.SEP_ARGUMENTS
            				+ types.getTypeSetOf(this.getChanName(this.currentPrefixing))
            				+ HandelC.PAR_END + HandelC.END_COMMAND;
                
            	// Release the semaphors that have been taken
            	for (int j=0; j<semaphorsTaken.size(); j++) {
                	this.hcCode = this.hcCode + this.getReleaseSema((String)semaphorsTaken.get(j));
            	}
            	
                Node process = initials.getProcess(event);
                process.apply(this);
                this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
            }
            this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP + HandelC.DELAY + HandelC.END_COMMAND;
            
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    }
/*
    private void translateCase(InitialsExtractorVisitor initials,
    							List toCheckEvents,
    							String controlVar) {
    	// Translate the cases
        this.hcCode = this.hcCode + HandelC.ALTERNATION;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        // Declare the cases
        for(int i=0; i < toCheckEvents.size(); i++) {
            this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
            PEvent event = (PEvent)toCheckEvents.get(i);
            LinkedList comm = initials.getCommunication(event);
            this.translatePrefix(event,comm,DO_NOT_DECLARE_INPUT_VAR);
            this.hcCode = this.hcCode + HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;

            Node process = initials.getProcess(event);
            process.apply(this);
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
        }
        this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP + HandelC.DELAY + HandelC.END_COMMAND;
        
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
 */   
    private String getForcedInterleavingControlVarName(int diff){
    	return HandelC.DONE_EVENTS + HandelC.UNDERSCORE + diff;
    }
    
    public void caseACspSequentialProcess(ACspSequentialProcess node){
        
        this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        node.getLeftProcess().apply(this);
        node.getRightProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    private String typeArgument(String argument, String process){
        String code = "";
        if (this.directives.containsProcArguments(process) 
            && this.directives.getProcArguments(process).containsKey(argument)) {
            String type = (String)this.directives.getProcArguments(process).get(argument);
            code = code + type + argument;
        } else {
            throw new UndefinedArgumentTypeException(argument,process);
        }
        
        return code;
    }

    // if declaration is true it declares the variable, otherwise it initialises it
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String typeArgumentCopy(String argument, String process, 
                                        boolean declaration, boolean isArgument){
        String code = "";
        if (this.directives.containsProcArguments(process) 
            && this.directives.getProcArguments(process).containsKey(argument)) {
            String type = (String)this.directives.getProcArguments(process).get(argument);
            String copy = this.copyName(process,argument,declaration);
            if (declaration) {
                // Declare copy
                String decl = type + HandelC.SPACE + copy;
                if (isArgument) {
                    // As a function argument
                    code = code + HandelC.SEP_ARGUMENTS + HandelC.SPACE + decl;
                } else {
                    // As a local variable
                    code = code + decl + HandelC.END_COMMAND;
                }
            } else {
                // Initialise Copy
                code = code + copy + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE;
                code = code + argument + HandelC.END_COMMAND;
            }
        } else {
            throw new UndefinedArgumentTypeException(argument,process);
        }
        return code;
    }
    
    // Return the name of the copy of the local variable together with any
    // possible array access
    // Receives the process name P, the variable name N, and a boolean that indicates
    // if this is a declaration or not
    // Returns the name P_local_N[
    private String copyName(String procName, String varName, boolean declaration) {
        String copyName = procName + HandelC.LOCAL_COPY + varName;
        return this.getVariableArray(copyName,declaration);
    }  
    
    private void declareChannels(){
    	// Declaring the semaphors
    	Iterator elements = this.usedSemaphors.iterator(); 
    	while(elements.hasNext()) {
    		String semaphor = (String)elements.next();
            this.hcCode = HandelC.SEMAPHOR + HandelC.SPACE + semaphor + HandelC.END_COMMAND + this.hcCode;
    	}
    	
    	// Declaring the channels
        List<String> names = this.channels.getChannelNames();
        List<String> newnames = new Vector();
        for(int i=0; i < names.size(); i++) {
        	String name = names.get(i);
        	newnames.add(this.getChanName(name));
        }
        TypesVisitor typesVisitor = new TypesVisitor(directives,this.constants);
        typesVisitor.declareType(HandelC.CHANNEL_TYPE, newnames);
        this.hcCode = typesVisitor.getHcCode() + this.hcCode;

    	Enumeration keys = this.directives.getAllChannels();
        
        while(keys.hasMoreElements()) {
            String chan = (String)keys.nextElement();
            String decl = this.channels.getChannelDeclaration(chan);
            this.hcCode = decl + HandelC.END_COMMAND + this.hcCode;
        }
        this.hcCode = HandelC.generateComment(Messenger.CHANNELS_DECL) + this.hcCode;
        
    }
    
    private String getChanName(String name) {
    	return HandelC.CHANNEL + HandelC.UNDERSCORE + name;
    }
    
    private void declareProcessesPrototypes(){
        Enumeration keys = this.procsInfo.keys();
        
        while(keys.hasMoreElements()) {
            String process = ((String)keys.nextElement()).trim();
        String declaration = HandelC.INLINE + HandelC.SPACE            
                             + HandelC.VOID + HandelC.SPACE + process;
            if (this.procsInfo.containsProcArguments(process)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(process);
                declaration = declaration + translateFormalArguments(parameters,process);
            }
            declaration = declaration + HandelC.END_COMMAND;
            this.hcCode = declaration + this.hcCode;
        }
        this.hcCode = HandelC.generateComment(Messenger.PROCESSES_PROTOTYPES) + this.hcCode;
    }

    private void declareDatatypes(){
        this.hcCode = this.types.getHcCode() + this.hcCode;
    }
    
    private void declareClock(){
        String declaration = HandelC.generateComment(Messenger.CLOCK_DECLARATION);
        declaration = declaration + HandelC.CLOCK_DEF;
        declaration = declaration + HandelC.COMMENT_LINE_BREAK;
        
        this.hcCode = declaration + this.hcCode;
    }

    
    private void declareAuxiliaryConstants(){
        this.hcCode = HandelC.SYNC_DEF + HandelC.SYNC_OUT_DEF + HandelC.SYNC_IN_DEF + this.hcCode;
        this.hcCode = HandelC.generateComment(Messenger.SYNC_VALUE_DECL) + this.hcCode;
        
        /*
        String emptyChannelSetDecl = HandelC.CHANNELS + HandelC.DATATYPE_EMPTY_SET + HandelC.DATATYPE_MEMBER_SET
                                    + HandelC.SPACE + HandelC.BINARY_PREFIX + HandelC.MATH_ZERO;
        String nextLine = HandelC.generateDefine(emptyChannelSetDecl);
        this.hcCode = nextLine + this.hcCode;
        this.hcCode = HandelC.generateComment(Messenger.EMPTY_CHANNEL_SET_DECL) + this.hcCode;
		*/
    }
    
    public void declareAuxiliaryMacros() {

        /*
		inline void LEAVE_SEMAPHOR(CHANNEL_TYPE c, sema *s, CHANNEL_TYPE_set S){
		    if (CHANNEL_MEMBER(c,S)){
		        releasesema(*s);
		    }
		}
        */
    	String leaveSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.CHANNEL_TYPE + HandelC.SPACE + "c" + HandelC.SEP_ARGUMENTS 
                        + HandelC.SEMAPHOR + HandelC.SPACE + HandelC.REFERENCE + "s" + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL_TYPE + HandelC.SET_SUFFIX + HandelC.SPACE + "S"
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.IF + HandelC.PAR_BEGIN
                        + HandelC.CHANNEL_MEMBER
                        + HandelC.PAR_BEGIN
                        + "c" + HandelC.SEP_ARGUMENTS  + "S"
                        + HandelC.PAR_END
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.RELEASE_SEMAPHOR
                        + HandelC.PAR_BEGIN
                        + HandelC.REFERENCE + "s"
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        this.hcCode = leaveSema + this.hcCode;

        /*
		inline void ENTER_SEMAPHOR(CHANNEL_TYPE c, sema *s, CHANNEL_TYPE_set S){
		    if (CHANNEL_MEMBER(c,S)){
		        while(trysema(*s)==0) delay;
		    }
		}
        */
    	String enterSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.ENTER_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.CHANNEL_TYPE + HandelC.SPACE + "c" + HandelC.SEP_ARGUMENTS 
                        + HandelC.SEMAPHOR + HandelC.SPACE + HandelC.REFERENCE + "s" + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL_TYPE + HandelC.SET_SUFFIX + HandelC.SPACE + "S"
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.IF + HandelC.PAR_BEGIN
                        + HandelC.CHANNEL_MEMBER
                        + HandelC.PAR_BEGIN
                        + "c" + HandelC.SEP_ARGUMENTS  + "S"
                        + HandelC.PAR_END
                        + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.WHILE	
                        + HandelC.PAR_BEGIN
                        + HandelC.TRY_SEMAPHOR
                        + HandelC.PAR_BEGIN
                        + HandelC.REFERENCE + "s"
                        + HandelC.PAR_END
                        + HandelC.REL_EQUALS + HandelC.MATH_ZERO
                        + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.DELAY + HandelC.END_COMMAND
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        this.hcCode = enterSema + this.hcCode;

        /*
        macro expr unsigned_MACRO(integer_local_) = ((unsigned)(integer_local_ + integer_card + integer_offset) % integer_card);
        */
    	String memberMacro = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.CHANNEL_MEMBER + HandelC.PAR_BEGIN
                        + "e,s" + HandelC.PAR_END 
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN 
                        + HandelC.PAR_BEGIN                        
                        + types.getLookupTableName(HandelC.CHANNEL_TYPE,HandelC.SET_SUFFIX)
                        + HandelC.ARRAY_BEGIN
                        + HandelC.PAR_BEGIN
                        + HandelC.UNSIGNED
                        + HandelC.PAR_END +"e"
                        + HandelC.ARRAY_END
                        + HandelC.SPACE
                        + HandelC.LOGICAL_BITWISE_OR
                        + HandelC.SPACE + "s"
                        + HandelC.PAR_END
                        + HandelC.REL_EQUALS
                        + "s" + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = memberMacro + this.hcCode;

        /*
        macro expr SET_UNION(SET_UNION_S,SET_UNION_T) = (SET_UNION_S | SET_UNION_T);
        */
    	String setUnion = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.SET_UNION + HandelC.PAR_BEGIN
                        + HandelC.SET_UNION + "_S" + HandelC.SEP_ARGUMENTS 
                        + HandelC.SET_UNION + "_T " + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.SET_UNION + "_S" 
                        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_OR + HandelC.SPACE
                        + HandelC.SET_UNION + "_T" 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = setUnion + this.hcCode;

        /*
        macro expr IS_EMPTY_CHANNEL_SET(SET) = (SET == CHANNEL_TYPE_nil_set);
        */
    	String isEmptyChannelSet = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.IS_EMPTY_CHANNEL_SET + HandelC.PAR_BEGIN
                        + "S" + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + "S" 
                        + HandelC.SPACE + HandelC.REL_EQUALS + HandelC.SPACE
                        + types.getEmptySetOf(HandelC.CHANNEL_TYPE) 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = isEmptyChannelSet + this.hcCode;

        
        this.hcCode = HandelC.generateComment(Messenger.AUX_MACROS) + this.hcCode;

    }

    private void declareConstants(){
        this.hcCode = this.constants.getHcCode() + this.hcCode;
        this.hcCode = HandelC.generateComment(Messenger.CONSTANTS) + this.hcCode;
        this.declareConstantsPrototypes();
    }

    private void declareConstantsPrototypes(){
        Enumeration keys = this.constants.getConstantsNames();
        
        while(keys.hasMoreElements()) {
            String name = (String)keys.nextElement();
            List args = this.constants.getArguments(name);
            
            // macro expr name = expression;
            String macro = HandelC.MACRO_EXP + HandelC.SPACE + name;
            
            if (args.size() > 0) {
                macro = macro + HandelC.PAR_BEGIN;
                for(int i=0; i < args.size(); i++){
                   macro = macro + args.get(i);
                   if (i < args.size() - 1) {
                       macro = macro + HandelC.SEP_ARGUMENTS;
                   }
                }
                macro = macro + HandelC.PAR_END;
            }
            macro = macro + HandelC.END_COMMAND;
            this.hcCode = macro + this.hcCode;
        }
        
        this.hcCode = HandelC.generateComment(Messenger.CONSTANTS_PROTOTYPES) + this.hcCode;
    }
    
    
    private void declareMutualRecConstants(){
        for (int i=0; i<this.processesIds.size(); i++) {
            String nextLine = HandelC.generateDefine(this.processesIds.get(i) + HandelC.SPACE + i);
            this.hcCode = nextLine + this.hcCode;
        }
        this.hcCode = HandelC.generateComment(Messenger.MUTUAL_REC_STATE_DECL) + this.hcCode;
    }
    
    
    private String translateFormalArguments(ACspParameters parameters, String process){
        String code = HandelC.PAR_BEGIN;

        code = code + HandelC.CHANNEL_TYPE + HandelC.SET_SUFFIX
        		+ HandelC.SPACE + HandelC.FORCED_SYNC;
        if (parameters.getParamList() != null){
            code = code + ",";
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
            } else if (paramsList instanceof ACspLinkedParamList) {
                String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
                code = code + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    code = code + this.typeArgument(parameter,process);
                    code = code + HandelC.SEP_ARGUMENTS;
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
            }
        }
        code = code + HandelC.PAR_END;
        return code;
    }

    // if declaration is true it declares the variables, otherwise it initialises them
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String translateFormalArgumentsCopies(ACspParameters parameters, String process, 
                                                    boolean declaration, boolean argument){
        String code = "";
        if (parameters.getParamList() != null){
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = this.typeArgumentCopy(parameter,process,declaration,argument);
            } else if (paramsList instanceof ACspLinkedParamList) {
                String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                code = this.typeArgumentCopy(parameter,process,declaration,argument);
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    code = code + this.typeArgumentCopy(parameter,process,declaration,argument);
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgumentCopy(parameter,process,declaration,argument);
            }
        }
        
        return code;
    }
    
    public void translateRealArguments(String name, ACspArguments arguments, String current){
        this.check(name,arguments,current);
        
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;

        if (!this.forcedInterleaving.isEmpty()) {
        	this.hcCode = this.hcCode + this.translateForcedSyncSet();
        } else {
        	this.hcCode = this.hcCode + types.getEmptySetOf(HandelC.CHANNEL_TYPE);
        }

        if (arguments != null && arguments.getArgumentList() != null){
            if (!this.forcedInterleaving.isEmpty()) {
            	this.hcCode = this.hcCode + ",";
            }
            
    		// Writing the parameters
            Node argsList = ((ACspArguments)arguments).getArgumentList();

            if (argsList instanceof ACspBaseArgumentList) {
                this.translateArgument(((ACspBaseArgumentList)argsList).getArgument());
            } else if (argsList instanceof ACspLinkedArgumentList) {
                this.translateArgument(((ACspLinkedArgumentList)argsList).getArgument());
                this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    this.translateArgument(((ACspLinkedArgumentList)argsList).getArgument());
                    this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                }
                argsList = ((ACspLinkedArgumentList)argsList).getNext();
                this.translateArgument(((ACspBaseArgumentList)argsList).getArgument());
            }
        }
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void translateRealArgumentsCopies(String procName, ACspArguments arguments, String current){
        this.check(procName,arguments,current);
        
        if (arguments != null && arguments.getArgumentList() != null){
            if (this.procsInfo.containsProcArguments(procName)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(procName);
                
                PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
                PParamList paramsList = parameters.getParamList();
                
                if (argsList instanceof ACspBaseArgumentList) {
                    this.translateRealArgumentCopy(procName, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                } else if (argsList instanceof ACspLinkedArgumentList) {
                    this.translateRealArgumentCopy(procName, 
                                                    ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                    ((ACspLinkedParamList)paramsList).getParam());
                    while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                        argsList = ((ACspLinkedArgumentList)argsList).getNext();
                        paramsList = ((ACspLinkedParamList)paramsList).getNext();
                        this.translateRealArgumentCopy(procName, 
                                                        ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                        ((ACspLinkedParamList)paramsList).getParam());
                    }
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    this.translateRealArgumentCopy(procName, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                }
            } else {
                throw new UndefinedProcessArgumentsException(procName);
            }
        }
    }

    private void translateRealArgumentCopy(String procName, PArgument arg, PParam param) {
        this.hcCode = this.hcCode + this.copyName(procName,param.toString().trim(),IS_NOT_DECLARATION) + HandelC.SPACE; 
        this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE; 
        this.translateArgument(arg);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
    }

    private String translateForcedSyncSet(){
    	Set<String> forcedChannels = (Set)this.forcedInterleaving.peek();
    	List setNames = new Vector();
    	for(String c:forcedChannels){
    		String singletonSet = this.getChanName(c) + HandelC.SET_SUFFIX;
    		setNames.add(singletonSet);
    	}
    	return this.getSetUnionCode(setNames);
    }
    
    private String getSetUnionCode(List<String> elements) {
		String retorno = HandelC.CHANNEL_TYPE + HandelC.EMPTY_SET_SUFFIX + HandelC.SET_SUFFIX;
    	if (elements.size() > 0) {
    		String head = elements.get(0);
    		elements.remove(0);
			retorno = HandelC.SET_UNION + HandelC.PAR_BEGIN
					+ head + HandelC.SEP_ARGUMENTS + this.getSetUnionCode(elements) 
					+ HandelC.PAR_END;
		}
		return retorno;
    }
    
    private void check(String name, ACspArguments arguments, String current){
        if (this.procsInfo.containsProcArguments(name)) {
            ACspParameters parameters = this.procsInfo.getProcArguments(name);
            if(!this.check(parameters,arguments)) {
                throw new WrongNumberOfArgumentsException(current,name);
            }
        } else {
            if (arguments != null && arguments.getArgumentList() != null){
                throw new WrongNumberOfArgumentsException(current,name);
            }
        }
    }

    private boolean check(ACspParameters parameters, ACspArguments arguments) {
        boolean answer = true;
        
        if (arguments != null && arguments.getArgumentList() != null 
            && parameters != null && parameters.getParamList() != null) {
            PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
            PParamList paramsList = parameters.getParamList();

            if ((paramsList instanceof ACspBaseParamList) && (argsList instanceof ACspBaseArgumentList)) {
                // do nothing!!! They are ok
            } else if ((paramsList instanceof ACspLinkedParamList) 
                        && (argsList instanceof ACspLinkedArgumentList)) {
                while( ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList)
                        && ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList) ){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                }
                boolean moreParams = ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList);
                boolean moreArgs = ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList);
                if (moreParams != moreArgs) {
                    answer = false;
                }
            } else {
                answer = false;
            }
        } else {
            boolean anyParams = (parameters != null && parameters.getParamList() != null);
            boolean anyArguments = (arguments != null && arguments.getArgumentList() != null);
            if (anyParams != anyArguments) {
                answer = false;
            }
        }
        return answer;
    }
    
    /*
     * Checks if there is an input in the list of communications.
     * Returns the name of the variable if yes and null otherwise.
     */
    private String checkInput(LinkedList comms) {
        String var = null;
        ListIterator iterator = comms.listIterator(0);
        
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                var = ((ACspInputComunication)next).getExpression().toString();
            }
        }
        return var;
    }
    
    private String translatePrefixWithForcedSync(PEvent event, LinkedList comm, boolean declareInputVar) {
    	String retorno;
        this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();

    	this.hcCode = this.hcCode + this.getTrySema(this.currentPrefixing) + HandelC.END_COMMAND;
    	retorno = this.translatePrefix(event, comm, declareInputVar);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
    	this.hcCode = this.hcCode + this.getReleaseSema(this.currentPrefixing);

    	return retorno;
    }
    
    private String getTrySema(String chan){
    	
        return HandelC.ENTER_SEMAPHOR + HandelC.PAR_BEGIN
		+ this.getChanName(chan) + HandelC.SEP_ARGUMENTS
		+ HandelC.ADDRESS + this.getChannelSemaphor(chan) + HandelC.SEP_ARGUMENTS
		+ HandelC.FORCED_SYNC
		+ HandelC.PAR_END;
    }

    private String getReleaseSema(String chan){
        return  HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
		+ this.getChanName(chan) + HandelC.SEP_ARGUMENTS
		+ HandelC.ADDRESS + this.getChannelSemaphor(chan) + HandelC.SEP_ARGUMENTS
		+ HandelC.FORCED_SYNC
		+ HandelC.PAR_END + HandelC.END_COMMAND;
    }
    
    private String getChannelSemaphor(String chan) {
    	String semaphor = chan + HandelC.SEMAPHOR_SUFFIX;
    	this.usedSemaphors.add(semaphor);
    	
    	//Hashtable semas = (Hashtable)this.semaphors.peek();
    	//semaphor = (String)semas.get(chan);
    	
        return semaphor;
    }

    private String translatePrefix(PEvent event, LinkedList comm, boolean declareInputVar) {
        this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
        ListIterator iterator = comm.listIterator(0);
        
        String inputVar = null;
        
        this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));

        if (!iterator.hasNext()) {
            event.apply(this);
            if (this.directives.containsProcChannelInOut(this.currentDefinition)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentDefinition);
                if (channelInOut.containsKey(this.currentPrefixing)){
                    String inOut = (String)channelInOut.get(this.currentPrefixing);
                    if (inOut.equals(Directives.CHANNEL_IN)) {
                        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
                        this.hcCode = this.hcCode + HandelC.SYNC_IN;
                    } else if (inOut.equals(Directives.CHANNEL_OUT)) {
                        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
                        this.hcCode = this.hcCode + HandelC.SYNC_OUT;
                    }
                } else {
                    throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
                }
            } else {
                throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
            }

        } else {
            if (declareInputVar) {
                inputVar = this.declareInputVar(comm);
            } 
            event.apply(this);
            
            /* 
             * The Restrictions Visitor guarantess that we have communications
             * of the form c[.exp]*[?name | !exp] 
            */
            while(iterator.hasNext()) {
                Object next = iterator.next();
                if (next instanceof ACspInputComunication) {
                    ((ACspInputComunication)next).apply(this);
                } else if (next instanceof ACspOutputComunication) {
                    ((ACspOutputComunication)next).apply(this);
                } else if (next instanceof ACspProjection) {
                    ((ACspProjection)next).apply(this);
                }
            }
        }
        return inputVar;
    }

    // Declares the input variable (if there is any)
    private String declareInputVar(LinkedList comm) {
        ListIterator iterator = comm.listIterator(0);
        
        String var = this.checkInput(comm);
        if (var != null) {
            if (this.channels.containsChannelType(this.currentPrefixing)) {
                String type = this.channels.getChannelCommType(this.currentPrefixing);
                this.hcCode = this.hcCode + type + HandelC.SPACE + var + HandelC.END_COMMAND;
            } else {
                throw new UndefinedChannelTypeException(this.currentPrefixing,this.currentDefinition);
            }
            var = var.trim();
            this.inputVariables.add(var);
        }
        return var;
    }

    // Translates all the processes definitions as an action system like
    // Does not support parametrised processes
    private void translateProcessesDefinitionsMutualRecursion() {
        int numberOfProcesses = this.procsInfo.size();
        int numberOfBits = MathUtil.getNumberOfBits(numberOfProcesses);

        // Declare local copies for all arguments of every parametrised process
        this.hcCode = this.hcCode + HandelC.generateComment(Messenger.MUTUAL_REC_LOCAL_VARS);
        
        
        for (int i=0; i< this.processesIds.size(); i++) {
            String name = (String)this.processesIds.get(i);
            if (this.procsInfo.containsProcArguments(name)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(name);
                if (parameters == null) {
                    parameters = new ACspParameters();
                }
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,name,IS_DECLARATION,IS_NOT_ARGUMENT);
            }
        }

        // inline void MUTUAL_REC(
        this.hcCode = this.hcCode + HandelC.generateComment(Messenger.MUTUAL_REC_METHOD);
        this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION + HandelC.PAR_BEGIN;

        // int numberOfBits MUTUAL_REC_i
        //this.hcCode = this.hcCode + this.getMutualRecIndexDecl();

        // int numberOfBits PROGRAM_COUNTER
        //this.hcCode = this.hcCode + HandelC.generateComment(Messenger.MUTUAL_REC_AUX_VARS);
        this.hcCode = this.hcCode + HandelC.INT + HandelC.SPACE;
        this.hcCode = this.hcCode + numberOfBits + HandelC.SPACE;
        this.hcCode = this.hcCode + this.getVariableArray(HandelC.PROGRAM_COUNTER,IS_DECLARATION);
        //this.hcCode = this.hcCode + HandelC.END_COMMAND;

        // ){
        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
        
        // int 1 KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
        this.hcCode = this.hcCode + this.getVariableArray(HandelC.KEEP_LOOPING,IS_DECLARATION);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;

        // KEEP_LOOPING "[MUTUAL_REC_i]" = true;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
        
        // while(KEEP_LOOPING "[MUTUAL_REC_i]") {
        this.hcCode = this.hcCode + HandelC.WHILE;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + this.getVariableArray(HandelC.KEEP_LOOPING,IS_NOT_DECLARATION);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        // KEEP_LOOPING "[MUTUAL_REC_i]" = false;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
        
        // switch(PROGRAM_COUNTER "[MUTUAL_REC_i]") {
        this.hcCode = this.hcCode + HandelC.SWITCH + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + this.getVariableArray(HandelC.PROGRAM_COUNTER,IS_NOT_DECLARATION);
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        // case PROC_NAME: 
        for (int i = 0; i < this.processesIds.size(); i++) {
            this.setCurrentDefinition(((String)this.processesIds.get(i)).trim());

            this.hcCode = this.hcCode + HandelC.generateComment(Messenger.PROCESS + HandelC.SPACE + this.currentDefinition);
            this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
            this.hcCode = this.hcCode + this.currentDefinition + HandelC.SPACE + HandelC.CASE_SEP;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            if(this.procsInfo.containsProcDefinition(this.currentDefinition)) {
                this.procsInfo.getProcDefinition(this.currentDefinition).apply(this);
            }
            this.hcCode = this.hcCode + HandelC.BREAK + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
        }

        // default: KEEP_LOOPING "[MUTUAL_REC_i]" = false;
        this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP + HandelC.SPACE;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);

        // }}}
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    private void translateExpression(PExpr expression) {
        expression.apply(this);
    }

    private void translateArgument(PArgument argument) {
        argument.apply(this);
    }

    private int getNumberOfParallelismBranches(){
        return this.parallelismCounter + 1;
    }
    
    private String getMutualRecIndexDecl() {
        String answer = "";
        if (this.needsLocalVariablesAsArrays()) {
            answer = answer + HandelC.UNSIGNED + HandelC.SPACE;
            answer = answer + HandelC.INT + HandelC.SPACE;
            answer = answer + MathUtil.getNumberOfBits(this.getNumberOfParallelismBranches()) + HandelC.SPACE;
            answer = answer + this.getVariableArrayIndex();
        }
        
        return answer;
    }

    // Indicates the necessity of representing the local variables
    // and the variables related to mutual recursion as arrays 
    private boolean needsLocalVariablesAsArrays() {
        /*
        boolean parallelism = (this.getNumberOfParallelismBranches() > 1);
        boolean mutualRecursion = (this.recursion.getMutualRecursion());
        return (parallelism && mutualRecursion);
         */
        
        return false;
    }

    // Indicates that we are translating the main method's process 
    private boolean isTranslatingMain() {
        return this.currentDefinition.equals(HandelC.MAIN);
    }

    // Returns the string representing the index of the array of variables
    // that must be accessed.  This is either the variable MUTUAL_REC_i (if we are
    // translating the process) or the curernt branch in the parallelism (if
    // we are translating the main process
    private String getVariableArrayIndex() {
        String index = "";
        if (this.isTranslatingMain()) {
            index = index + this.currentParallelBranchInProcess;
        } else {
            index = HandelC.MUTUAL_REC_INDEX;
        }
        return index;
    }
    
    // Returns the string representing the access to the array of variables
    // that must be accessed.
    private String getVariableArray(String variable, boolean declaration) {
        String arrayRef = variable; 
        if (this.needsLocalVariablesAsArrays()) {
            arrayRef = arrayRef + HandelC.ARRAY_BEGIN;
            if (declaration) {
                arrayRef = arrayRef + this.getNumberOfParallelismBranches();
            }  else {
                arrayRef = arrayRef + this.getVariableArrayIndex();
            }
            arrayRef = arrayRef + HandelC.ARRAY_END;
        }
        return arrayRef;
    }
    
    /*
    private String getMutualRecursionInvocation() {
        String answer = HandelC.MUTUAL_RECURSION + HandelC.PAR_BEGIN;
        if (this.needsLocalVariablesAsArrays()) {
            answer = answer + this.getVariableArrayIndex();
        }
        answer = answer + HandelC.PAR_END + HandelC.END_COMMAND;
        return answer;
    }
    */

    // Translates both branches of a parallelism or interleaving
    private void translateConcurrency(Node left, Node right, Set forcedInterleaving){
		
		Set allChannels = new HashSet<String>(this.channels.getChannelNames());
    	if (this.forcedInterleaving.size() >= 1) {
    		// We are at a secondary parallel composition
        	// i.e. Some P in (P||Q) || (R || S), or deeper
    		Set forcedInterleavingTop = (Set)this.forcedInterleaving.peek();
    		Set noLongerForced = forcedInterleavingTop;
    		noLongerForced.removeAll(forcedInterleaving);
    		System.out.println("ProcessesTranslatorVisitor :: translateConcurrency -> No longer forced: "+noLongerForced);
    		if (!noLongerForced.isEmpty()) {
    		    //throw new UnsupportedConstructorException(Messenger.PARALLEL_COMPOSITION_FORCED_INTERLEAVING,this.currentDefinition);
    		}
    	}
		// Push the new forced interleaving into the stack
		this.forcedInterleaving.push(forcedInterleaving);
		
		this.parallelismCounter++;
    
        this.hcCode = this.hcCode + HandelC.PARALLEL;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        left.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.currentParallelBranchInProcess++;
        
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        right.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        
        this.hcCode = this.hcCode + HandelC.BLOCK_END;

		this.forcedInterleaving.pop();
    }

    /*
    private void translateSemaphors() {
        this.hcCode = this.hcCode + HandelC.END_COMMAND + HandelC.PARALLEL;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        Iterator events = ((Set)this.forcedInterleaving.peek()).iterator();
    	while(events. hasNext()) {
    		String event = (String)events.next();
    		this.translateSemaphor(event);
    		this.hcCode = this.hcCode + HandelC.END_COMMAND;
    	}

    	this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;
    }

    private void translateSemaphor(String event) {
    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    	this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
    	this.hcCode = this.hcCode + this.getSemaphorCondVar();
    	this.hcCode = this.hcCode + HandelC.SPACE + HandelC.ASSIGNMENT;
    	this.hcCode = this.hcCode + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND;
    	this.hcCode = this.hcCode + HandelC.WHILE + HandelC.PAR_BEGIN;
    	this.hcCode = this.hcCode + this.getSemaphorCondVar() + HandelC.PAR_END;
    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        this.hcCode = this.hcCode + HandelC.ALTERNATION;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        // Declare the cases
        this.translateSemaphorChannelChoice(event, HandelC.LEFT);
        this.translateSemaphorChannelChoice(event, HandelC.RIGHT);
        this.translateSemaphorChannelChoice(event, HandelC.KILL);

        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }

    private String getSemaphorCondVar(){
    	return HandelC.SEMAPHOR + HandelC.UNDERSCORE + HandelC.KEEP_LOOPING
    			+ HandelC.UNDERSCORE + this.getParallelismCounter();
    }
    
    private void translateSemaphorChannelChoice(String event, String side) {
    	if (side.equals(HandelC.KILL)) {
		    this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
		    this.hcCode = this.hcCode + this.getSemaphorChannelComm(event,side,HandelC.KILL) 
		    	+ HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;
		    this.hcCode = this.hcCode + this.getSemaphorCondVar() + HandelC.SPACE 
		    			+ HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.LOGICAL_FALSE 
		    			+ HandelC.END_COMMAND;
	        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
    	} else {
		    this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
		    this.hcCode = this.hcCode + this.getSemaphorChannelComm(event,side,HandelC.ENTER) 
		    	+ HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;
		    this.hcCode = this.hcCode + this.getSemaphorChannelComm(event,side,HandelC.LEAVE) + HandelC.END_COMMAND;
	        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
    	}
    }
    
	private String getSemaphorChannelComm(String event, String side, String action) {
		return this.getSemaphorChannelName(event, side, action) 
			+ HandelC.INPUT_COMMUNICATION
	    	+ HandelC.SYNC_IN;
	}

	private String getSemaphorChannelName(String event, String side, String action) {
		return HandelC.SEMAPHOR + HandelC.UNDERSCORE + this.getParallelismCounter()
					+ HandelC.UNDERSCORE + event
					+ HandelC.UNDERSCORE + side 
					+ HandelC.UNDERSCORE + action;
	}
	*/
    
    // when translating the accesses to an element of an array of channel
    // this function is invoked in order to guarantee that, if the type
    // is an integer, it uses the unsigned representation of that
    // integer; otherwise, we could have access to negative numbers
    private void unsignInteger() {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.UNSIGNED + HandelC.PAR_END + HandelC.PAR_BEGIN;
    }
    
    // Closes the access to the arrays of channels
    private void nextArrayAccessChannelSync() {
        this.closeArrayAccessChannelSync();
        this.openArrayAccessChannelSync();
    }

    // Closes the access to the arrays of channels
    private void closeArrayAccessChannelSync() {

        // this is the first access, do nothing
        if(!this.currentType.equals("")) {
            if(this.currentType.equals(CspUtil.getIntType())) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            this.hcCode = this.hcCode + HandelC.ARRAY_END;
        }
        
    }

    // Opens the access to the arrays of channels
    private void openArrayAccessChannelSync() {

        this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
        if (this.intArrayAccessChannelSync()) { 
            this.unsignInteger(); 
        }
        
        this.removeCurrentPrefixingTypes();
    }
    
    // Indicates that we were accessing an array using integers 
    private boolean intArrayAccessChannelSync() {
        boolean answer = false;
        
        boolean empty = this.currentPrefixingTypes.isEmpty();
        boolean lastType = this.currentPrefixingTypes.size()==1;
        boolean isComm = this.channels.isCommChannel(this.currentPrefixing);
        boolean isInt = this.currentPrefixingTypes.get(0).toString().trim().equals(CspUtil.getIntType());
        
        if (!empty) {
            if (lastType) {
                if (!isComm && isInt) {
                    answer = true;
                }
            } else {
                if (isInt) {
                    answer = true;
                }
            }
        }
        
        return answer;
    }
    
    private void removeCurrentPrefixingTypes() {
        if (!this.currentPrefixingTypes.isEmpty()) {
            this.currentType = this.currentPrefixingTypes.get(0).toString().trim();
            this.currentPrefixingTypes.remove(0);
        }
    }

    private void openArrayAccessTypeConstructor(){
        this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
        if(!this.tagsTypes.isEmpty()) {
            String type = (String)ListUtil.head(((List)this.tagsTypes.peek()));
            if(type.equals(HandelC.INTEGERL)) {
                this.unsignInteger();
            }
        }
    }
    
    private void closeArrayAccessTypeConstructor(){
        if(!this.tagsTypes.isEmpty() && !(((List)this.tagsTypes.peek()).size() == 0)) {
            List types = (List)this.tagsTypes.pop();
            String type = (String)ListUtil.head(types);
            if(type.equals(HandelC.INTEGERL)) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            if (types.size() > 1) {
                this.tagsTypes.push(ListUtil.tail(types));
            }
        }
        this.hcCode = this.hcCode + HandelC.ARRAY_END;
        
    }

    private void nextArrayAccessTypeConstructor(){
        this.closeArrayAccessTypeConstructor();
        this.openArrayAccessTypeConstructor();
    }

}
