package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class RestrictionsVisitor extends ExtendedDepthFirstAdapter {
    
    /*
     * The name of the current definition being checked
    */ 
    String currentDefinition;

    /*
     * Used in the verification of the types of communication accepted. Indicates that a 
     * communication of a value has been found and therefore, this event should neither
     * output nor synchornise in anything else.
    */ 
    boolean commPrefixFound;
    
    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Visitor containing info about the directives
    private DirectivesVisitor directives;

    // Visitor containing info about types
    private TypesVisitor types;

    // Stores the number of projections used for each channel
    private Hashtable channelProjections;

    // Counts the number of projections for the current channel
    private int projectionCounter;

    // Counts the number of arguments left in a projection that uses datatypes tags
    private int tagsArgsLeft;
    
    public RestrictionsVisitor(DirectivesVisitor directives, TypesVisitor types) {
        this.commPrefixFound = false;
        this.directives = directives;
        this.channelProjections = new Hashtable();
        this.types = types;
        this.tagsArgsLeft = 0;
    }
    
    public void caseACspSubtypeDefinition(ACspSubtypeDefinition node) {
        throw new InvalidDefinitionException(node.toString());
    }
    public void caseACspNametypeDefinition(ACspNametypeDefinition node) {
        throw new InvalidDefinitionException(node.toString());
    }
    public void caseACspExternalDefinition(ACspExternalDefinition node) {
        throw new InvalidDefinitionException(node.toString());
    }
    public void caseACspTransparentDefinition(ACspTransparentDefinition node) {
        throw new InvalidDefinitionException(node.toString());
    }
    public void caseACspFunctionDefinition(ACspFunctionDefinition node) {
        throw new InvalidDefinitionException(node.toString());
    }

    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        TName name = definition.getName();
        this.currentDefinition = name.toString().trim();
        ACspParameters parameters = (ACspParameters)definition.getParameters();
        PExpression expression = definition.getExpression();

        if (parameters == null) {
            parameters = new ACspParameters();
        }
        expression.apply(this);
    }

    public void caseACspDatatypeDefinition(ACspDatatypeDefinition definition) {
        String name = definition.getName().toString().trim();
        this.currentDefinition = name;
        definition.getBranchList().apply(this);
    }

    public void caseACspChannelDefinition(ACspChannelDefinition node){
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        this.currentDefinition = channels.toString();
        if (type != null) {
            type.apply(this);
        }
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
        this.currentDefinition = node.getName().toString().trim();
        node.getExpression().apply(this);
    }
    
    
    /*
    public void caseACspReferencedType(ACspReferencedType node){
        String referencedName = node.getName().toString().trim();
        
    }
    */
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        this.projectionCounter = 0;
        
        PEvent event = node.getEvent();
        PProjection projection = ((ACspEvent)event).getProjection();
        LinkedList comm = node.getComunication();
        PPrefixProcess proc = node.getProcess();

        
        if (projection != null) {
            projection.apply(this);
        }
        ListIterator iterator = comm.listIterator(0);
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                ((ACspInputComunication)next).apply(this);
            } else if (next instanceof ACspOutputComunication) {
                ((ACspOutputComunication)next).apply(this);
            } else if (next instanceof ACspProjection) {
                ((ACspProjection)next).apply(this);
            }
        }
        
        this.commPrefixFound = false;
        
        // Checking the consistent use of the channel projections
        String channelName = ((ACspEvent)this.currentPrefix.getEvent()).getChannelName().toString().trim();
        if (this.channelProjections.containsKey(channelName)) {
            int counter = ((Integer)this.channelProjections.get(channelName)).intValue();
            
            if(this.projectionCounter != counter) {
                throw new InconsistentChannelProjectionUseException(channelName, this.currentDefinition);
            }
        } else {
            this.channelProjections.put(channelName, new Integer(this.projectionCounter));
        }
        
        proc.apply(this);
    }
    
    public void caseACspInputComunication(ACspInputComunication node) {
        if (this.commPrefixFound) {
            throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
        } else {
            this.commPrefixFound = true;
            node.getExpression().apply(this);
        }
    }

    public void caseACspOutputComunication(ACspOutputComunication node) {
        if (this.commPrefixFound) {
            throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
        } else {
            this.commPrefixFound = true;
            node.getExpression().apply(this);
        }
    }

    public void caseACspProjection(ACspProjection node) {
        if (this.tagsArgsLeft > 0) {
            this.tagsArgsLeft--;
        } else {
            this.projectionCounter++;
            if (this.commPrefixFound) {
                throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
            }
        }
        node.getValue().apply(this);
    }
    
    public void caseACspCallPostfixExpr(ACspCallPostfixExpr node){
        String name = ((ACspCallPostfixExpr)node).getName().toString().trim();
        if (this.types.isTag(name)) {
            this.tagsArgsLeft = this.tagsArgsLeft + this.types.getTagValues(name).size();
        }
        // this might as well be a projection, so...
        if (node.getProjection() != null) {
            node.getProjection().apply(this);
        }
    }

    public void caseACspAlphabetisedReplicatedProcess(ACspAlphabetisedReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.ALPHABETISED_REPLICATED_PROCESS,this.currentDefinition);
    }
    
    public void caseACspLinkedParallelProcess(ACspLinkedParallelProcess node) {
        throw new UnsupportedConstructorException(Messenger.LINKED_PARALLEL,this.currentDefinition);
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
        // Verify that there are only prefixed processes in the choice
        
        // Verify that there are no two input variables with the same name
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);

        List keys = initialsExtractor.keyList();

        Vector inputVars = new Vector();
        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);

            ListIterator iterator = comm.listIterator(0);

            while(iterator.hasNext()) {
                Object next = iterator.next();
                if (next instanceof ACspInputComunication) {
                    String var = ((ACspInputComunication)next).getExpression().toString();
                    if (inputVars.contains(var)) {
                        throw new UnsupportedConstructorException(Messenger.INPUT_PREFIX_SAME_VAR_NAME + var,this.currentDefinition);
                    } else {
                        inputVars.add(var);
                    }
                }
            }
        }
        
        node.getLeftProcess().apply(this);
        node.getRightProcess().apply(this);
    }

    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
        throw new UnsupportedConstructorException(Messenger.TUPLE_VALUE,this.currentDefinition);
    }

    // Sets
    public void caseACspComprehensionSet(ACspComprehensionSet node) {
        throw new UnsupportedConstructorException(Messenger.SET_COMPREHENSION,this.currentDefinition);
    }
    public void caseACspClosedRange(ACspClosedRange node) {
        throw new UnsupportedConstructorException(Messenger.CLOSED_RANGE,this.currentDefinition);
    }
    public void caseACspOpenRange(ACspOpenRange node) {
        throw new UnsupportedConstructorException(Messenger.OPEN_RANGE,this.currentDefinition);
    }

    // Sequences
    public void caseACspSequenceValue(ACspSequenceValue node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_VALUE,this.currentDefinition);
    }
    public void caseACspCatSequenceExpr(ACspCatSequenceExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_CONCATENATION,this.currentDefinition);
    }
    public void caseACspLengthPrefixExpr(ACspLengthPrefixExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_LENGTH,this.currentDefinition);
    }

}

