package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class ChannelsVisitor extends ExtendedDepthFirstAdapter {

    // Resulting handel-c code
    private String hcCode;

    // Type of the channels
    // Name:String -> Types:PType Vector
    private Hashtable channelTypesMap;
    
    // Indicate which channels communicate values using projections c.*
    private List projChannels; 

    // Indicate which channels communicate values using either ! or ?
    // String Vector
    private List commChannels; 

    // Indicates to the visitor that we are checking the names of
    // the channels being declared
    private boolean checkingChannelNames;

    // Stores the channel names of the current channel definition
    private List channelNames;
    
    // Indicates to the visitor that we are checking the type of
    // the channels being declared
    private boolean checkingChannelTypes;

    // String containing the type cardinality
    private String typeCard;
    
    // String containing the type handel-c equivalent
    private String typeHandelC;
    
    // Stores the channel types of the current channel definition
    private List channelTypes;
    
    // Visitor containing the information about the datatype
    private TypesVisitor datatypes;

    // Visitor containing the information about the directives
    private DirectivesVisitor directives;

    // Visitor containing the information about the processes
    private ProcessesInfoVisitor procInfo;

    // Visitor containing the information about the constants
    private ConstantsAndFunctionsVisitor constants;

    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Current Channel Definition
    private ACspChannelDefinition currentChannelDefinition;
    
    // Indicates that there are events in which we will need to force the interleaving
    private boolean needsForcedInterleaving;
    
    public ChannelsVisitor(TypesVisitor datatypes, DirectivesVisitor directives,
    		ProcessesInfoVisitor procInfo, ConstantsAndFunctionsVisitor constants) {
        this.channelTypesMap = new Hashtable();
        this.commChannels = new Vector();
        this.projChannels = new Vector();
        
        this.datatypes = datatypes;
        this.directives = directives;
        this.procInfo = procInfo;
        this.constants = constants;
        this.typeCard = "";
        this.typeHandelC = "";
        this.needsForcedInterleaving = false;
    }
    
    public List getChannelNames(){
    	List names = new Vector();
        Iterator iterator = this.channelTypesMap.keySet().iterator();
        while(iterator.hasNext()) {
        	names.add(iterator.next());
        }
        return names;
    }

    public List getProjChannels(){
        return this.projChannels;
    }

    public boolean isCommChannel(String channelName){
        return this.commChannels.contains(channelName);
    }
    public boolean isProjChannel(String channelName){
        return this.projChannels.contains(channelName);
    }

    public boolean needsForcedInterleaving(){
        return this.needsForcedInterleaving;
    }

    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        PExpression expression = definition.getExpression();
        expression.apply(this);
    }
    

    public void caseACspChannelDefinition(ACspChannelDefinition node){
        this.currentChannelDefinition = node;
        
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        // Retrieve Channel Names
        Vector names = new Vector();
        
        this.checkingChannelNames = true;
        this.channelNames = new Vector();
        channels.apply(this);
        this.checkingChannelNames = false;
        
        this.checkingChannelTypes = true;
        this.channelTypes = new Vector();
        if (type != null) {
            type.apply(this);
        }
        this.checkingChannelTypes = false;
        
        for(int i=0; i < this.channelNames.size(); i++) {
            this.channelTypesMap.put(this.channelNames.get(i),this.channelTypes);
        }
    }
    
    public void caseACspChannel(ACspChannel node){
        this.channelNames.add(node.getName().toString().trim());
    }
    
    public void caseACspType(ACspType node){
        if (this.checkingChannelTypes){
            this.channelTypes.add(node);
        }
    	if (node.getName().getText().equals(CspUtil.getBoolType())) {
            this.typeCard = HandelC.BOOLEAN + HandelC.TYPE_CARDINALITY;
            this.typeHandelC = HandelC.BOOLEAN;
    	} else if (node.getName().getText().equals(CspUtil.getIntType())) {
	        this.typeCard = HandelC.INTEGERL + HandelC.TYPE_CARDINALITY;
	        this.typeHandelC = HandelC.INTEGERL;
    	} else {
            this.typeCard = node.toString().trim() + HandelC.TYPE_CARDINALITY;
            this.typeHandelC = node.toString().trim();
    	}	
    }

    public void caseACspTupleType(ACspTupleType node){
        if (this.checkingChannelTypes){
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentChannelDefinition.toString());
        }
    }

    public void caseACspSetType(ACspSetType node){
        if (this.checkingChannelTypes){
            throw new UnsupportedConstructorException(Messenger.INVALID_TYPE_REFERENCE,this.currentChannelDefinition.toString());
        }
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)event).getComunication();
        PHelp proc = node.getProcess();
        LinkedList projection = ((ACspEvent)event).getProjection();

        String channelName = ((ACspEvent)event).getChannelName().toString().trim();
        
        // Checking Projections
        if (projection != null) {
            if (!this.projChannels.contains(channelName)) {
                this.projChannels.add(channelName);
            }    
        }

        // Checking communications
        ListIterator iterator = comm.listIterator(0);
        boolean isCommunication = false;
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if ((next instanceof ACspInputComunication) 
                    || (next instanceof ACspOutputComunication)) {
                if (!this.commChannels.contains(channelName)) {
                    this.commChannels.add(channelName);
                }    
            }
        }
        proc.apply(this);
    }

    // Returns the dimension of an array of channels
    // This is the number of type expression used in the channel declaration
    // (minus one if this is a communication channel)
    private int getArrayDimension(String channelName) {
        int size = this.getChannelType(channelName).size();
        if (size > 0 && this.isCommChannel(channelName)) {
            size = size - 1;
        }
        return size;
    }

    public boolean containsChannelType(String channelName){
        return this.channelTypesMap.containsKey(channelName);
    }
    
    public List getChannelType(String channelName){
        return (List)this.channelTypesMap.get(channelName);
    }

    public String getChannelCommType(String chan) {
        String commType = "";
        List types = this.getChannelType(chan);
        int arrayDim = this.getArrayDimension(chan);
        if(types.size() == arrayDim) {
            // There is no actual value being communicated. Type of communication is SYNC
             commType= HandelC.SYNC_TYPE;
        } else {
            // There is a value being communicated. Type of communication is the last type
            PType type = (PType)ListUtil.last(types);
            type.apply(this);
            commType = this.typeHandelC;
        }
        
        return commType;
    }
    
    public String getChannelDeclaration(String chan) {
        String decl = "";
        decl = decl + HandelC.CHANNEL + HandelC.SPACE;
        
        // Declaring the value communicated via this channel
        if (this.containsChannelType(chan)) {
            List types = this.getChannelType(chan);
            int arrayDim = this.getArrayDimension(chan);

            if(types.size() == arrayDim) {
                // There is no actual value being communicated. Type of communication is SYNC
                decl = decl + HandelC.SYNC_TYPE;
            } else {
                // There is a value being communicated. Type of communication is the last type
                PType type = (PType)ListUtil.last(types);
                decl = decl + this.getChannelCommType(chan);
            }
            decl = decl + HandelC.SPACE + chan;
            
            // Declaring possible arrays
            
            if(this.isProjChannel(chan)){
                for (int i = 0; i < arrayDim; i++) {
                    PType type = (PType)types.get(i);
                    decl = decl + HandelC.ARRAY_BEGIN;
                    type.apply(this);
                    decl = decl + this.typeCard;
                    decl = decl + HandelC.ARRAY_END;
                }
            }
        }
        return decl;
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
    	this.checkConcurrency(node);
    }

    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
    	this.checkConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
    	this.checkConcurrency(node);
    }

    public void checkConcurrency(Node node) {
        Set forcedInterleaving = null;

        if (node instanceof ACspSharingParallelProcess) {
            forcedInterleaving = this.getForcedInterleavingEvents((ACspSharingParallelProcess)node);
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            forcedInterleaving = this.getForcedInterleavingEvents((ACspAlphabetisedParallelProcess)node);
    	} else if (node instanceof ACspInterleaveProcess) {
            forcedInterleaving = this.getForcedInterleavingEvents((ACspInterleaveProcess)node);
    	}

        if (forcedInterleaving != null && !forcedInterleaving.isEmpty()) {
        	this.needsForcedInterleaving = true;
        }
    }

    public Set getForcedInterleavingEvents(ACspSharingParallelProcess node) {
        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();

        // Checking the synchronisation channel set
        PExpression syncExp = ((ACspSharingParallelProcess)node).getExpression();
        UsedChannelsVisitor usedChannels = new UsedChannelsVisitor(this.procInfo);
    	syncExp.apply(usedChannels);

        return this.getForcedInterleavingEvents(left, right, usedChannels.getUsedChannels());
    }

    public Set getForcedInterleavingEvents(ACspAlphabetisedParallelProcess node) {
        Node left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
        Node right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();

        // Checking the synchronisation channel set
        PExpression syncExpLeft = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
        UsedChannelsVisitor usedChannelsLeft = new UsedChannelsVisitor(this.procInfo);
        syncExpLeft.apply(usedChannelsLeft);

        PExpression syncExpRight = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
        UsedChannelsVisitor usedChannelsRight = new UsedChannelsVisitor(this.procInfo);
        syncExpRight.apply(usedChannelsRight);
        
        // Intersection
        Set leftSet = usedChannelsLeft.getUsedChannels();
        Set rightSet = usedChannelsRight.getUsedChannels();
        leftSet.retainAll(rightSet);

        return this.getForcedInterleavingEvents(left, right, leftSet);
    }

    public Set getForcedInterleavingEvents(ACspInterleaveProcess node) {
        Node left = ((ACspInterleaveProcess)node).getLeftProcess();
        Node right = ((ACspInterleaveProcess)node).getRightProcess();
        return this.getForcedInterleavingEvents(left, right, new HashSet());
    }

    public Set getForcedInterleavingEvents(Node left, Node right, Set sync){
        UsedChannelsVisitor usedChannels = new UsedChannelsVisitor(this.procInfo);
        
        left.apply(usedChannels);
        Set leftUsedChannels = usedChannels.getUsedChannels();
        usedChannels.resetUsedChannels();
        right.apply(usedChannels);
        Set rightUsedChannels = usedChannels.getUsedChannels();
        usedChannels.resetUsedChannels();
        
        rightUsedChannels.retainAll(leftUsedChannels); // rightUsedChannels intersection leftUsedChannels
        rightUsedChannels.removeAll(sync); // rightUSedChannels minus sync
        
        return rightUsedChannels;
    }

}

