package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class DirectivesVisitor extends ExtendedDepthFirstAdapter {
    // Directives Information

    // Channel Input or Output -- procName:String -> (channelName:String -> input/output:String)
    public Hashtable channelInOut;

    // Argument types -- procName:String -> (argumentName:String -> type:String)
    public Hashtable arguments;

    // Empty sets types -- procName:String -> type:String
    public Hashtable emptySets;
    
    // Main process
    public String mainProc;

    // Number of bits for the CSP type Int
    public int intBits;
    
    // Current directive
    public String currentDirective;
    
    // Current line
    public int currentLine;

    public DirectivesVisitor() {
        this.channelInOut = new Hashtable();
        this.arguments = new Hashtable();
        this.emptySets = new Hashtable();
        this.mainProc = "";
        this.currentDirective = "";
        this.currentLine = 1;
        this.intBits = 1;
    }

    public boolean containsProcChannelInOut(String procName) {
        return this.channelInOut.containsKey(procName);
    }
    
    public Hashtable getProcChannelInOut(String procName){
        return (Hashtable)this.channelInOut.get(procName);
    }

    public void putProcChannelInOut(String procName, String name, String type){
        if (this.channelInOut.containsKey(procName)) {
            Hashtable hashInOut = (Hashtable)this.channelInOut.get(procName);
            hashInOut.put(name,type);
        } else {
            Hashtable hashInOut = new Hashtable();
            hashInOut.put(name,type);
            this.channelInOut.put(procName,hashInOut);
        }
    }
    
    public boolean containsProcArguments(String procName) {
        return this.arguments.containsKey(procName);
    }
    
    public Hashtable getProcArguments(String procName){
        return (Hashtable)this.arguments.get(procName);
    }

    public boolean containsEmptySetType(String def) {
        return this.emptySets.containsKey(def);
    }
    
    public String getEmptySetType(String def){
        return (String)this.emptySets.get(def);
    }
    
    public Enumeration getAllChannels() {
        Vector channels = new Vector();
        Enumeration procNames = this.channelInOut.keys();
        while(procNames.hasMoreElements()) {
            Object proc = procNames.nextElement();
            Hashtable procInfo = (Hashtable)this.channelInOut.get(proc);
            
            Enumeration channelsNames = procInfo.keys();
            while(channelsNames.hasMoreElements()) {
                Object c = channelsNames.nextElement();
                if (!channels.contains(c)) {
                    channels.addElement(c);
                }
            }
        }
        
        return channels.elements();
    }

    public String getMain(){
        return this.mainProc;
    }
    
    public int getIntBits() { return this.intBits; }

    public void outStart(Start node) {
        // If directive for main method is not declared or is empty, set to default SKIP
        if (this.mainProc.trim().equals("")) {
            this.mainProc = CspUtil.getSkip();
        }
    }
    
    public void caseACspLineComment(ACspLineComment node){
        TSingleLineComment lineComment = node.getSingleLineComment();
        if (lineComment.toString().trim().startsWith(Directives.DIRECTIVE)){
            this.currentDirective = lineComment.toString();
            this.currentLine = lineComment.getLine();
            this.processDirective(lineComment);
        }
    }

    private void processDirective(TSingleLineComment lineComment){
        this.currentLine = lineComment.getLine();
        this.currentDirective = lineComment.toString();
        
        String directive = this.currentDirective;
        directive = StringUtil.removeBreakLines(directive);
        directive = StringUtil.removeReturns(directive);
        directive = StringUtil.removeTabs(directive);
        directive = StringUtil.removeExtraSpaces(directive);
        
        StringTokenizer tokenizer = new StringTokenizer(directive,HandelC.SPACE);
        
        // throws the directive header away
        tokenizer.nextToken();
        if (tokenizer.hasMoreTokens()) {
            // gets the directive type
            String directiveType = tokenizer.nextToken();
            if (directiveType.equals(Directives.DIR_CHANNEL_IN_OUT)){
                this.processChannelInOut(tokenizer);
            } else if (directiveType.equals(Directives.DIR_ARGUMENT)){
                this.processArgumentType(tokenizer);
            } else if (directiveType.equals(Directives.DIR_MAIN)){
                this.processMain(tokenizer);
            } else if (directiveType.equals(Directives.DIR_INT_BITS)){
                this.processIntBits(tokenizer);
            } else if (directiveType.equals(Directives.DIR_EMPTY_SET_TYPE)){
                this.processEmptySetType(tokenizer);
            } else {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INEXISTENT_DIRECTIVE,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.INEXISTENT_DIRECTIVE,this.currentLine);
        }
    }

    private void processChannelInOut(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String name = null;
        String type = null;
        String within = null;
        String proc = null;
        while (tokenizer.hasMoreTokens()) {
            if (name == null) {
                name = tokenizer.nextToken();
            } else if (type == null) {
                type = tokenizer.nextToken();
            } else if (within == null) {
                within = tokenizer.nextToken();
                if (!within.equals(Directives.WITHIN)) {
                    throw new InvalidDirectiveException(this.currentDirective, Messenger.EXPECTED+Directives.WITHIN, this.currentLine);
                }
            } else if (proc == null) {
                proc = tokenizer.nextToken();
            } else {
                throw new InvalidDirectiveException(this.currentDirective, Messenger.CHANNEL_IN_OUT_TOO_MUCH_INFO, this.currentLine);
            }
        }
        if (name != null && type != null && within != null && proc != null) {
            if (type.equals(Directives.CHANNEL_IN) || type.equals(Directives.CHANNEL_OUT)) {
                if (this.channelInOut.containsKey(proc)) {
                    Hashtable hashInOut = (Hashtable)this.channelInOut.get(proc);
                    hashInOut.put(name,type);
                } else {
                    Hashtable hashInOut = new Hashtable();
                    hashInOut.put(name,type);
                    this.channelInOut.put(proc,hashInOut);
                }
            } else {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.CHANNEL_IN_OUT_WRONG_TYPE,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.CHANNEL_IN_OUT_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processArgumentType(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String name = null;
        String type = null;
        String within = null;
        String proc = null;
        while (tokenizer.hasMoreTokens()) {
            if (name == null) {
                name = tokenizer.nextToken();
            } else if (type == null) {
                type = "";
                while (tokenizer.hasMoreTokens() && within == null) {
                    String token = tokenizer.nextToken();
                    if (!token.equals(Directives.WITHIN)) {
                        type = type + token + HandelC.SPACE;
                    } else {
                        within = token;
                    }
                }
                if (within == null) {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Directives.WITHIN,this.currentLine);
                } else {
                    if (tokenizer.hasMoreTokens()) {
                        proc = tokenizer.nextToken();
                    } else {
                        throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Messenger.PROCESS_NAME,this.currentLine);
                    }
                }
            }
        }
        if (name != null && type != null && within != null && proc != null) {
            Hashtable hashArguments = new Hashtable();
            if (this.arguments.containsKey(proc)) {
                Hashtable hashArgs = (Hashtable)this.arguments.get(proc);
                hashArgs.put(name,type);
            } else {
                Hashtable hashArgs = new Hashtable();
                hashArgs.put(name,type);
                this.arguments.put(proc,hashArgs);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.ARGUMENT_TOO_FEW_INFO,this.currentLine);
        }
    }
    
    private void processMain(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String mainP = "";
        while (tokenizer.hasMoreTokens()) {
            String token = tokenizer.nextToken();
            mainP = mainP + token + HandelC.SPACE;
        }
        this.mainProc = mainP;
    }
    
    private void processIntBits(StringTokenizer tokenizer) throws InvalidDirectiveException{
        if (tokenizer.hasMoreElements()) {
            String sizeStr = tokenizer.nextToken();
            try {
                this.intBits = (new Integer(sizeStr)).intValue();
                if (this.intBits < 1) {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_WRONG_TYPE,this.currentLine);
                }
            } catch (NumberFormatException e) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_WRONG_TYPE,this.currentLine);
            }
            if (tokenizer.hasMoreElements()) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_TOO_MUCH_INFO,this.currentLine);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.INT_BITS_TOO_FEW_INFO,this.currentLine);
        }
    }

    private void processEmptySetType(StringTokenizer tokenizer) throws InvalidDirectiveException{
        String type = null;
        String def = null;
        while (tokenizer.hasMoreTokens()) {
            if (type == null) {
                type = "";
                if (tokenizer.hasMoreTokens()) {
                    String token = tokenizer.nextToken();
                    type = token.trim();
                }
                if (tokenizer.hasMoreTokens()) {
                    String token = tokenizer.nextToken();
                    if (!token.equals(Directives.WITHIN)) {
                        throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Directives.WITHIN,this.currentLine);
                    }
                }
                
                if (tokenizer.hasMoreTokens()) {
                    def = tokenizer.nextToken().trim();
                } else {
                    throw new InvalidDirectiveException(this.currentDirective,Messenger.EXPECTED+Messenger.DEFINITION_NAME,this.currentLine);
                }
            }
        }
        if (type != null && def != null) {
            if (this.emptySets.containsKey(def)) {
                throw new InvalidDirectiveException(this.currentDirective,Messenger.EMPTY_SET_TYPE_REDEFINED,this.currentLine);
            } else {
                this.emptySets.put(def,type);
            }
        } else {
            throw new InvalidDirectiveException(this.currentDirective,Messenger.EMPTY_SET_TYPE_TOO_FEW_INFO,this.currentLine);
        }
    }
    
}
