package csp2hc.core;

import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspAlphabetisedReplicatedProcess;
import lmf.formula.csp.node.ACspCatSequenceExpr;
import lmf.formula.csp.node.ACspChannelDefinition;
import lmf.formula.csp.node.ACspClosedRange;
import lmf.formula.csp.node.ACspComprehensionSet;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspConstantDefinition;
import lmf.formula.csp.node.ACspDatatypeDefinition;
import lmf.formula.csp.node.ACspDotExpr;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspExternalDefinition;
import lmf.formula.csp.node.ACspFunctionCallExpr;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspLengthExpr;
import lmf.formula.csp.node.ACspLinkedParallelProcess;
import lmf.formula.csp.node.ACspNametypeDefinition;
import lmf.formula.csp.node.ACspOpenRange;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspParameters;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspRenamingProcess;
import lmf.formula.csp.node.ACspSequenceValue;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.ACspSubtypeDefinition;
import lmf.formula.csp.node.ACspTransparentDefinition;
import lmf.formula.csp.node.ACspTupleValue;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PChannelList;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.PTypeSpecifier;
import lmf.formula.csp.node.TName;
import csp2hc.exception.InconsistentChannelProjectionUseException;
import csp2hc.exception.InconsistentParallelCompositionException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.util.CspUtil;

public class RestrictionsVisitor extends ExtendedDepthFirstAdapter {
    
    // Visitor containing info about the directives
    private DirectivesVisitor directives;

    // Visitor containing info about the constants and functions
    private ConstantsAndFunctionsVisitor constants;

    // Visitor containing info about types
    private TypesVisitor types;

    // Visitor that has the information about the channels
    private ChannelsVisitor channels;

    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    /*
     * The name of the current definition being checked
    */ 
    String currentDefinition;

    /*
     * Used in the verification of the types of communication accepted. Indicates that a 
     * communication of a value has been found and therefore, this event should neither
     * output nor synchornise in anything else.
    */ 
    boolean commPrefixFound;
    
    /*
     * Used in the verification of the types of communication accepted. Stores the current
     * prefixing being analysed
    */ 
    ACspPrefixProcess currentPrefix;

    // Stores the number of non-projections used for each channel
    private Hashtable channelNonProjections;

    // Counts the number of non-projections for the current channel
    private int nonProjectionCounter;

    // Counts the number of arguments left in a projection that uses datatypes tags
    private int tagsArgsLeft;
    
    // Indicates that we are checking synchronisation channel sets
    private boolean checkingSyncSet;

    // Stack of sets of events whose interleaving must be forced
    private Stack forcedInterleaving;

    // Set of processes names that already been analysed while 
    // checking a concurrent process
    private Set<String> checkedProcesses;

    public RestrictionsVisitor(DirectivesVisitor directives, 
    							TypesVisitor types,
    							ConstantsAndFunctionsVisitor constants,
    							ChannelsVisitor channels,
    							ProcessesInfoVisitor procsInfo) {
        this.directives = directives;
        this.types = types;
        this.constants = constants;
        this.channels = channels;
        this.procsInfo = procsInfo;

        this.commPrefixFound = false;
        this.channelNonProjections = new Hashtable();
        this.tagsArgsLeft = 0;
        this.checkingSyncSet = false;
        this.forcedInterleaving = new Stack();
        this.checkedProcesses = new HashSet<String>();
    }
    
    public void caseACspSubtypeDefinition(ACspSubtypeDefinition node) {
        throw new UnsupportedConstructorException(Messenger.SUBTYPE,node.getName().toString());
    }
    public void caseACspNametypeDefinition(ACspNametypeDefinition node) {
        throw new UnsupportedConstructorException(Messenger.NAMETYPE,node.getName().toString());
    }
    public void caseACspExternalDefinition(ACspExternalDefinition node) {
        throw new UnsupportedConstructorException(Messenger.EXTERNAL,node.getNameList().toString());
    }
    public void caseACspTransparentDefinition(ACspTransparentDefinition node) {
        throw new UnsupportedConstructorException(Messenger.TRANSPARENT,node.getNameList().toString());
    }

    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
    	String pname = node.getName().toString().trim();
        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.checkedProcesses.contains(pname)) {
    		this.checkedProcesses.add(pname);
        	PExpression proc = this.procsInfo.getProcDefinition(pname);
        	proc.apply(this);
    		this.checkedProcesses.remove(pname);
    	}
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        TName name = definition.getName();
        this.currentDefinition = name.toString().trim();
        ACspParameters parameters = (ACspParameters)definition.getParameters();
        PExpression expression = definition.getExpression();

        if (parameters == null) {
            parameters = new ACspParameters();
        }
        expression.apply(this);
    }

    public void caseACspDatatypeDefinition(ACspDatatypeDefinition definition) {
        String name = definition.getName().toString().trim();
        this.currentDefinition = name;
        definition.getBranchList().apply(this);
    }

    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
    	if (this.checkingSyncSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_CONSTANT,this.currentDefinition);
    	}
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
    	if (this.checkingSyncSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_FUNCTION,this.currentDefinition);
    	}
    }
    
    public void caseACspChannelDefinition(ACspChannelDefinition node){
        PChannelList channels = node.getChannelList();
        PTypeSpecifier type = node.getTypeSpecifier();
        
        this.currentDefinition = channels.toString();
        if (type != null) {
            type.apply(this);
        }
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
        this.currentDefinition = node.getName().toString().trim();
        node.getExpression().apply(this);
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){

        this.currentPrefix = node;
        this.nonProjectionCounter = 0;
        
        PEvent event = node.getEvent();
        LinkedList projections = ((ACspEvent)event).getProjection();
        LinkedList comm = ((ACspEvent)event).getComunication();
        PHelp proc = node.getProcess();

        
        for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
        
        ListIterator iterator = comm.listIterator(0);
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                ((ACspInputComunication)next).apply(this);
            } else if (next instanceof ACspOutputComunication) {
                ((ACspOutputComunication)next).apply(this);
            } else if (next instanceof ACspProjection) {
                ((ACspProjection)next).apply(this);
            }
        }
        
        this.commPrefixFound = false;
        
        // Checking the consistent use of the channel projections
        String channelName = ((ACspEvent)this.currentPrefix.getEvent()).getChannelName().toString().trim();
        if (this.channelNonProjections.containsKey(channelName)) {
            int counter = ((Integer)this.channelNonProjections.get(channelName)).intValue();
            
            if(this.nonProjectionCounter != counter) {
                throw new InconsistentChannelProjectionUseException(channelName, this.currentDefinition);
            }
        } else {
            this.channelNonProjections.put(channelName, new Integer(this.nonProjectionCounter));
        }
        
        proc.apply(this);
    }
    
    public void caseACspInputComunication(ACspInputComunication node) {
        if (this.commPrefixFound) {
            throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
        } else {
        	this.nonProjectionCounter++;
            this.commPrefixFound = true;
            node.getParam().apply(this);
        }
    }

    public void caseACspOutputComunication(ACspOutputComunication node) {
        if (this.commPrefixFound) {
            throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
        } else {
        	this.nonProjectionCounter++;
            this.commPrefixFound = true;
            node.getExpression().apply(this);
        }
    }

    public void caseACspProjection(ACspProjection node) {
    	if (this.checkingSyncSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_PROJECTION,this.currentDefinition);
    	} else {
	        if (this.tagsArgsLeft > 0) {
	            this.tagsArgsLeft--;
	        } else {
	            if (this.commPrefixFound) {
	                throw new UnsupportedConstructorException(this.currentPrefix.toString(),this.currentDefinition);
	            }
	        }
	        node.getCallExpr().apply(this);
    	}
    }
    
    public void caseACspDotExpr(ACspDotExpr node){
        if (node.getCallExpr() instanceof ACspConstantCallExpr) {
        	ACspConstantCallExpr consCall = (ACspConstantCallExpr) node.getCallExpr();
        	String name = consCall.getName().toString().trim();
            if (this.types.isTag(name)) {
                this.tagsArgsLeft = this.tagsArgsLeft + this.types.getTagValues(name).size();
            } else {
            	consCall.apply(this);
            }
        }
        
        List projections = node.getProjection();
        for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
    }

    /*
    public void caseACspConstantCallExpr(ACspConstantCallExpr node) {
        if
    }
    */

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.checkConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.checkConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.checkConcurrency(node);
    }

    public void caseACspRenamingProcess(ACspRenamingProcess node) {
        throw new UnsupportedConstructorException(Messenger.RENAMING,this.currentDefinition);
    }
    
    public void checkConcurrency(Node node) {
        PExpression leftExp;
        PExpression rightExp;

        // Checking the synchronisation channel set
        
        this.checkingSyncSet = true;
        if (node instanceof ACspSharingParallelProcess) {
        	leftExp = ((ACspSharingParallelProcess)node).getExpression();
        	leftExp.apply(this);
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
    		leftExp = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
        	leftExp.apply(this);
    		rightExp = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
        	rightExp.apply(this);
    	} else if (node instanceof ACspInterleaveProcess) {
    	}
        this.checkingSyncSet = false;

        // Checking the branches
        Set forcedInterleaving = null;
        Node left = null;
        Node right = null;
        
        if (node instanceof ACspSharingParallelProcess) {
            forcedInterleaving = this.channels.getForcedInterleavingEvents((ACspSharingParallelProcess)node);
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            forcedInterleaving = this.channels.getForcedInterleavingEvents((ACspAlphabetisedParallelProcess)node);
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            forcedInterleaving = this.channels.getForcedInterleavingEvents((ACspInterleaveProcess)node);
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}

        if (forcedInterleaving == null || left == null || right == null) {
        	throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } else {
    		Set allChannels = new HashSet<String>(this.channels.getChannelNames());
        	if (this.forcedInterleaving.size() >= 1) {
        		// We are at a secondary parallel composition
            	// i.e. Some P in (P||Q) || (R || S), or deeper
        		Set forcedInterleavingTop = (Set)this.forcedInterleaving.peek();
        		if (!forcedInterleavingTop.equals(forcedInterleaving)) {
        		    throw new InconsistentParallelCompositionException(forcedInterleavingTop,forcedInterleaving,this.currentDefinition);
        		}
        	}
        	
    		// Push the new forced interleaving into the stack
    		this.forcedInterleaving.push(forcedInterleaving);
            left.apply(this);
            right.apply(this);
    		this.forcedInterleaving.pop();
        }
        
        /*
        if (forcedInterleaving == null || left == null || right == null) {
        	throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } else {
        	
        }
        */
        
    }
    
    public void caseACspAlphabetisedReplicatedProcess(ACspAlphabetisedReplicatedProcess node) {
        throw new UnsupportedConstructorException(Messenger.ALPHABETISED_REPLICATED_PROCESS,this.currentDefinition);
    }
    
    public void caseACspLinkedParallelProcess(ACspLinkedParallelProcess node) {
        throw new UnsupportedConstructorException(Messenger.LINKED_PARALLEL,this.currentDefinition);
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {
        // Verify that there are only prefixed processes in the choice
        
        // Verify that there are no two input variables with the same name
        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);

        List keys = initialsExtractor.keyList();

        Vector inputVars = new Vector();
        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);

            ListIterator iterator = comm.listIterator(0);

            while(iterator.hasNext()) {
                Object next = iterator.next();
                if (next instanceof ACspInputComunication) {
                    String var = ((ACspInputComunication)next).getParam().toString();
                    if (inputVars.contains(var)) {
                        throw new UnsupportedConstructorException(Messenger.INPUT_PREFIX_SAME_VAR_NAME + var,this.currentDefinition);
                    } else {
                        inputVars.add(var);
                    }
                }
            }
        }
        
        node.getLeftProcess().apply(this);
        node.getRightProcess().apply(this);
    }

    // Tuple
    public void caseACspTupleValue(ACspTupleValue node) {
        throw new UnsupportedConstructorException(Messenger.TUPLE_VALUE,this.currentDefinition);
    }

    // Sets
    public void caseACspComprehensionSet(ACspComprehensionSet node) {
        throw new UnsupportedConstructorException(Messenger.SET_COMPREHENSION,this.currentDefinition);
        /*
    	if (this.checkingSyncSet){
            throw new UnsupportedConstructorException(Messenger.SYNCHRONISATION_CHANNEL_SET_COMPREHENSION,this.currentDefinition);
    	}
    	*/
    }
    public void caseACspClosedRange(ACspClosedRange node) {
        throw new UnsupportedConstructorException(Messenger.CLOSED_RANGE,this.currentDefinition);
    }
    public void caseACspOpenRange(ACspOpenRange node) {
        throw new UnsupportedConstructorException(Messenger.OPEN_RANGE,this.currentDefinition);
    }

    // Sequences
    public void caseACspSequenceValue(ACspSequenceValue node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_VALUE,this.currentDefinition);
    }
    public void caseACspCatSequenceExpr(ACspCatSequenceExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_CONCATENATION,this.currentDefinition);
    }
    public void caseACspLengthExpr(ACspLengthExpr node) {
        throw new UnsupportedConstructorException(Messenger.SEQUENCE_LENGTH,this.currentDefinition);
    }

}

