package csp2hc.core;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class TailRecursionVisitor extends ExtendedDepthFirstAdapter {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the directives
    private DirectivesVisitor directives;

    // Stack of process definitions visited during analysis of 
    // a single process
    private Stack visitedProcesses;
    
    // Current process definition being analysed
    // If a process P invokes Q, this does not change
    // It only changes if we find a Q = ...
    private String currentDefinition;
    
    // Indicates that the visitor found a sequential composition and is
    // checking the let-hand side.
    private boolean checkingLeftBranchOfSequence;

    // Indicates that the visitor found a parallel composition and is
    // checking one of the branches.
    private boolean checkingParallelComposition;

    // List of tail recursive processes
    private List tailRecursiveProcesses;
    
    // Boolean that indicates if there is a mutual recursion in the system
    private boolean mutualRecursion;

    public TailRecursionVisitor(ProcessesInfoVisitor procInfo, DirectivesVisitor directives) {
        this.procsInfo = procInfo;
        this.directives = directives;
        this.visitedProcesses = new Stack();
        this.currentDefinition = "";
        this.checkingLeftBranchOfSequence = false;
        this.checkingParallelComposition = false;
        this.tailRecursiveProcesses = new Vector();
        this.mutualRecursion = false;
    }
    
    public List getTailRecursiveProcesses() { return this.tailRecursiveProcesses; }
    public boolean getMutualRecursion() { return this.mutualRecursion; }
    
    public void caseStart(Start node) {
        node.getPSpecification().apply(this);
    }
    
    public void caseACspDefinitionParagraph(ACspDefinitionParagraph node){
        PDefinition definition = node.getDefinition();
        if (definition instanceof ACspProcessDefinition){
            this.checkProcessDefinition((ACspProcessDefinition)definition);
        } 
    }
    
    private void checkProcessDefinition(ACspProcessDefinition definition) {
        TName name = definition.getName();
        this.currentDefinition = name.toString().trim();

        this.visitedProcesses.push(this.currentDefinition);

        PExpression expression = definition.getExpression();
        expression.apply(this);
        
        this.visitedProcesses.pop();
    }
    
    public void caseACspSequentialProcess(ACspSequentialProcess sequence) {
        PPrefixProcess left = sequence.getLeftProcess();
        PSequentialProcess right = sequence.getRightProcess();
        
        this.checkingLeftBranchOfSequence = true;
        left.apply(this);
        this.checkingLeftBranchOfSequence = false;
        right.apply(this);
    }
    
    public void caseACspProcessCallExpr(ACspProcessCallExpr node){
        String name = ((ACspProcessCallExpr)node).getName().toString().trim();

        // If it is a process name
        if (this.procsInfo.containsProcArguments(name)){

            if (this.checkingLeftBranchOfSequence && name.equals(this.currentDefinition)) {
                // Found a NON-TAIL Recursion
                throw new UnsupportedConstructorException(Messenger.NON_TAIL_RECURSION,this.currentDefinition);
            } else if (!this.checkingLeftBranchOfSequence && name.equals(this.currentDefinition)) {
                if (!this.visitedProcesses.empty() && !name.equals(this.visitedProcesses.peek())) {
                    // Found a MUTUAL Recursion
                    this.mutualRecursion = true;
                    
                    // Creating log of invocations
                    String log = name + Messenger.CLOSE_REASON;
                    Stack clone = (Stack)this.visitedProcesses.clone();
                    while (!clone.empty()) {
                        log = clone.pop() + Messenger.CALLS + log;
                    }
                    log = Messenger.OPEN_REASON + log;
                    
                    // Checking for parametrised processes
                    Enumeration keys = this.procsInfo.keys();
                    while (keys.hasMoreElements()) {
                        String next = (String)keys.nextElement();
                        if (this.procsInfo.containsProcArguments(next)) {
                            ACspParameters parameters = this.procsInfo.getProcArguments(next);
                            if(parameters != null && parameters.getParamList() != null) {
                                //throw new UnsupportedConstructorException(Messenger.PARAMETRISED_MUTUAL_RECURSION+log,next);
                            }
                        }
                    }
                } else {
                    // Found a TAIL Recursion
                    if(!this.tailRecursiveProcesses.contains(name)) {
                    	if (this.checkingParallelComposition) {
                            throw new UnsupportedConstructorException(Messenger.TAIL_RECURSION_IN_PARALLEL_COMPOSITION,this.currentDefinition);
                    	} else {
                    		this.tailRecursiveProcesses.add(name);
                    	}
                    }
                }
            } else if (!name.equals(this.currentDefinition) && !this.visitedProcesses.contains(name)){
                this.visitedProcesses.push(name);
                this.procsInfo.getProcDefinition(name).apply(this);
                this.visitedProcesses.pop();
            }
        }
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
    	this.checkConcurrency(node);
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node){
    	this.checkConcurrency(node);
    }

    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
    	this.checkConcurrency(node);
    }

    public void checkConcurrency(Node node) {
        Node left = null;
        Node right = null;
        
        if (!this.currentDefinition.equals(HandelC.MAIN)) {
            // No support for parallel composition (not in the main method) and mutual recursion together in
            // the same system
            if (this.mutualRecursion) {
                throw new UnsupportedConstructorException(Messenger.MUTUAL_RECURSION_AND_PARALLELISM,this.currentDefinition);
            }
        }
        if (node instanceof ACspSharingParallelProcess) {
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}
        
        if (left == null || right == null) {
        	throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } else {
            this.checkingParallelComposition = true;
            left.apply(this);
            this.checkingParallelComposition = true;
            right.apply(this);
            this.checkingParallelComposition = false;
        }
    }
}
