package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

import java.util.HashSet;

public class AlphabetsVisitor extends ExtendedDepthFirstAdapter {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the directives
    private DirectivesVisitor directives;

    // Stack of process definitions visited during analysis of 
    // a single process
    private Stack visitedProcesses;
    
    // Channel that are written to
    private Set<String> writtenChannels;

    // Channel that are read from
    private Set<String> readChannels;
    
    // Synchronisation channel set that is being constructed
    private Set<String> syncChannels;
    
    // Channels found when analysing a channel set
    private Set<String> channelsInSet;

    // Hidden channels that is being constructed
    private Set<String> hiddenChannels;

    // Visible channels that is being constructed
    private Set<String> visibleChannels;
    
    // Indicates if we are in a channel set
    private boolean isChannelSet;
    
    // Current process being analysed
    String currentAnalysedProcess = "";
    
    public AlphabetsVisitor(ProcessesInfoVisitor procInfo, DirectivesVisitor directives) {
        this.procsInfo = procInfo;
        this.directives = directives;
        this.writtenChannels = new HashSet<String>();
        this.readChannels = new HashSet<String>();
        this.hiddenChannels = new HashSet<String>();
        this.channelsInSet = new HashSet<String>();
        this.visibleChannels = new HashSet<String>();
        this.visitedProcesses = new Stack();
        this.isChannelSet = false;
    }
   
    public void setCurrentAnalysedProcess(String s){
    	this.currentAnalysedProcess = s;
    }
    public void resetChannels() { 
    	this.resetReferencedChannels(); 
    	this.resetHiddenChannels();
    	this.resetVisibleChannels();
    	this.resetChannelsInSet();
    }
    
    public void setIsChannelSet(boolean b) { this.isChannelSet = b; }
    
    public Set<String> getReferencedChannels() {
    	Set<String> alphabet = new HashSet<String>();
    	
    	Iterator<String> out = this.writtenChannels.iterator();
    	while(out.hasNext()) {
    		alphabet.add(out.next());
    	}
    	Iterator<String> in = this.readChannels.iterator();
    	while(in.hasNext()) {
    		alphabet.add(in.next());
    	}
    	return alphabet; 
    }
    public Set<String> getWrittenChannels() { 
    	return this.writtenChannels; 
    }
    public Set<String> getReadChannels() { 
    	return this.readChannels; 
  	}
    public Set<String> getHiddenChannels() { 
    	return this.hiddenChannels; 
    }
    public Set<String> getVisibleChannels() { 
    	return this.visibleChannels; 
    }
    public Set<String> getChannelsInSet() { 
    	return this.channelsInSet; 
    }

    private void resetReferencedChannels() { 
    	this.writtenChannels = new HashSet<String>(); 
    	this.readChannels = new HashSet<String>(); 
    }
    private void resetHiddenChannels() { this.hiddenChannels = new HashSet<String>(); }
    private void resetVisibleChannels() { this.visibleChannels = new HashSet<String>(); }
    private void resetChannelsInSet() { this.channelsInSet = new HashSet<String>(); }
    
    public Set<String> analyseChannelSet(Node node) {
        this.resetChannels();
        this.setIsChannelSet(true);
        node.apply(this);
        Set<String> channels = this.getChannelsInSet();
        this.setIsChannelSet(false);
        this.resetChannels();

        return channels;
    }

    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        String name = definition.getName().toString().trim();
        this.setCurrentAnalysedProcess(name);

        ACspParameters parameters = (ACspParameters)definition.getParameters();
        PExpression expression = definition.getExpression();
        if (parameters == null) {
            parameters = new ACspParameters();
        }
        expression.apply(this);
    }
    
    public void caseACspEvent(ACspEvent node){
        String name = node.getChannelName().toString().trim();
    	if (this.isChannelSet){
        	this.channelsInSet.add(name);
    	}
    }
    
    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
    	String pname = node.getName().toString().trim();

        //System.out.println("AlphabetsVisitor :: caseACspProcessCallExpr -> calling "+pname+" at "+this.currentAnalysedProcess);        
    	
        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.visitedProcesses.contains(pname)) {

        	String previousAnalysedProcess = this.currentAnalysedProcess;
        	this.setCurrentAnalysedProcess(pname);

        	this.visitedProcesses.add(pname);
        	PExpression proc = this.procsInfo.getProcDefinition(pname);
        	
            //System.out.println("AlphabetsVisitor :: caseACspProcessCallExpr -> analysing "+this.currentAnalysedProcess);        
        	
        	proc.apply(this);
    		this.visitedProcesses.remove(pname);

        	this.setCurrentAnalysedProcess(previousAnalysedProcess);
        
        }
    }

    public void caseACspHidingProcess(ACspHidingProcess node){
        Set visibleChannelsAbove = this.visibleChannels;
        Set hiddenChannelsAbove = this.hiddenChannels;
        this.resetVisibleChannels(); 
        this.resetHiddenChannels(); 

        Node hiddenChanExp = ((ACspHidingProcess)node).getHidenEvents();
        Set<String> hidden = this.analyseChannelSet(hiddenChanExp);
        this.hiddenChannels.addAll(hidden);

        Node process = ((ACspHidingProcess)node).getProcess();
        process.apply(this);

        this.visibleChannels.removeAll(hidden);
        this.visibleChannels.addAll(visibleChannelsAbove);
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        PHelp proc = node.getProcess();

    	String channel = ((ACspEvent)event).getChannelName().toString().trim();
        this.visibleChannels.add(channel);

        ListIterator commIter = comm.listIterator(0);

        //System.out.println("AlphabetsVisitor :: caseACspPrefixProcess -> process "+this.currentAnalysedProcess);
        //System.out.println("AlphabetsVisitor :: caseACspPrefixProcess -> channel "+channel);
        //System.out.println("AlphabetsVisitor :: caseACspPrefixProcess -> table "+this.directives.channelInOut);
        
        if (!commIter.hasNext()) {
            if (this.directives.containsProcChannelInOut(this.currentAnalysedProcess)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentAnalysedProcess);
                if (channelInOut.containsKey(channel)){
                    String inOut = (String)channelInOut.get(channel);
                    if (inOut.equals(Directives.CHANNEL_IN)) {
                        this.readChannels.add(channel);
                    } else if (inOut.equals(Directives.CHANNEL_OUT)) {
                        this.writtenChannels.add(channel);
                    }
                } else {
                    throw new UndefinedChannelInOutException(channel,this.currentAnalysedProcess);
                }
            } else {
                throw new UndefinedChannelInOutException(channel,this.currentAnalysedProcess);
            }
        } else {
            Object next = commIter.next();
            if (next instanceof ACspInputComunication) {
                this.readChannels.add(channel);
            } else if (next instanceof ACspOutputComunication) {
                this.writtenChannels.add(channel);
            }
        }

        proc.apply(this);
    }
}
