package csp2hc.visitors;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.List;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;
import java.util.Vector;

import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspArguments;
import lmf.formula.csp.node.ACspAssert;
import lmf.formula.csp.node.ACspBaseArgumentList;
import lmf.formula.csp.node.ACspBaseParamList;
import lmf.formula.csp.node.ACspConstantCallExpr;
import lmf.formula.csp.node.ACspConstantDefinition;
import lmf.formula.csp.node.ACspDotExpr;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspExternalChoiceProcess;
import lmf.formula.csp.node.ACspFunctionCallExpr;
import lmf.formula.csp.node.ACspFunctionDefinition;
import lmf.formula.csp.node.ACspGuardedProcess;
import lmf.formula.csp.node.ACspIfProcess;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspInternalChoiceProcess;
import lmf.formula.csp.node.ACspLinkedArgumentList;
import lmf.formula.csp.node.ACspLinkedParamList;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspParameters;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspProjection;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PArgument;
import lmf.formula.csp.node.PArgumentList;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.PParam;
import lmf.formula.csp.node.PParamList;
import lmf.formula.csp.node.PProjection;
import lmf.formula.csp.node.Start;
import csp2hc.exception.UndefinedArgumentTypeException;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.exception.UndefinedChannelTypeException;
import csp2hc.exception.UndefinedProcessArgumentsException;
import csp2hc.exception.UnsupportedConstructorException;
import csp2hc.exception.WrongNumberOfArgumentsException;
import csp2hc.messages.Directives;
import csp2hc.messages.HandelC;
import csp2hc.messages.Messenger;
import csp2hc.util.Branch2BranchSetLUT;
import csp2hc.util.CspUtil;
import csp2hc.util.ListUtil;
import csp2hc.util.MathUtil;

public class ProcessesTranslatorVisitor extends ExpressionsVisitor {
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;

    // Visitor that has the information about the types
    private TypesVisitor types;

    // Visitor that has the information about recursion
    private TailRecursionVisitor recursion;

    // Visitor that has the information about the channels
    private ChannelsVisitor channels;

    // Visitor that has the information about the constants
    private ConstantsAndFunctionsVisitor constants;
    
    // Visitor that has the information about Parallelism properties of the system
    private ParallelismAnalyserVisitor parallelismAnalyser;

    // Current prefixing being read
    private String currentPrefixing;

    // Current List of the remaining types for the current prefixing
    private List currentPrefixingTypes;

    // Current type for the current prefixing
    private String currentType;
    
    // Identity of Processes used in the translation of Mutual Recursion
    private List<String> processesIds;
    
    // List of current input variables in context
    private List inputVariables;

    // Indicates if the translator is translating a channel projection
    private boolean isProjection;

    // Stack of number of arguments to be translated for each tag
    // Changes the values according to the translation
    // Stack of int
    private Stack tagsArgsLeftStack;

    // Stack of number of arguments to be translated for each tag
    // Never changes the values (either push or pop)
    // Stack of int
    private Stack tagsArgsTotalStack;

    // Stack of types of arguments to be translated for each tag
    // Stack of (List of String)
    private Stack tagsTypesStack;
    
    // Stack of sets of events whose interleaving must be forced
    private Stack forcedInterleaving;

    // Flag that indicates if we should use semaphors or not
    private Boolean useSemaphors;

    private static boolean IS_ARGUMENT = true;
    private static boolean IS_NOT_ARGUMENT = false;
    private static boolean IS_DECLARATION = true;
    private static boolean IS_NOT_DECLARATION = false;
    private static boolean IS_PRI_ALT = true;
    private static boolean IS_NOT_PRI_ALT = false;
    private static boolean USE_DEFAULT = true;
    private static boolean DO_NOT_USE_DEFAULT = false;
    
    public ProcessesTranslatorVisitor(ProcessesInfoVisitor procsInfo, 
                                        DirectivesVisitor directives, 
                                        TypesVisitor types, 
                                        TailRecursionVisitor recursion,
                                        ChannelsVisitor channels,
                                        ConstantsAndFunctionsVisitor constants,
                                        ParallelismAnalyserVisitor parallalismAnalyser) {
        
        super(directives,false);
        
        this.procsInfo = procsInfo;
        this.types = types;
        this.recursion = recursion;
        this.channels = channels;
        this.constants = constants;
        this.parallelismAnalyser = parallalismAnalyser;
        
        this.currentPrefixing = "";
        this.currentPrefixingTypes = new Vector();
        this.currentType = "";
        
        // Creating Ids for each process
        this.processesIds = new Vector<String>();
        Enumeration<String> keys = this.procsInfo.keys();
        while(keys.hasMoreElements()) {
            this.processesIds.add(keys.nextElement());
        }
        
        this.inputVariables = new Vector();
        this.isProjection = false;
        this.tagsArgsLeftStack = new Stack();
        this.tagsArgsTotalStack = new Stack();
        this.tagsTypesStack = new Stack();
        this.forcedInterleaving = new Stack();
        this.useSemaphors = 
        		this.parallelismAnalyser.needsForcedInterleaving() || 
        		this.parallelismAnalyser.isMultiSync();
    }

    public String getCurrentDefinition(){ 
        return this.currentDefinition;
    }
    public void setCurrentDefinition(String def){ 
        super.setCurrentDefinition(def); 
    }
    
    public void caseStart(Start node){
    	if (!this.recursion.getMutualRecursion()) {
            node.getPSpecification().apply(this);
        } else {
            this.translateProcessesDefinitionsMutualRecursion();
        }
        
        // Preamble of file
        if (this.recursion.getMutualRecursion()) {
            this.declareMutualRecConstants();
        } else {
            this.declareProcessesPrototypes();
        }
        
        // Only if we have enforced interleaving
        if (this.useSemaphors) {
	        this.declareAuxiliarySemaphorMacrosAndFunctions();
	        this.declareForcedInterleavingVariablesSemaphors();
	        this.declareCurrentSyncStatusLUT();
	        this.declareForcedSyncLUT();
	        this.declareBranches();
        }
        this.declareAuxiliaryMacros();
        this.declareChannels();
        this.declareConstantsAndFunctions();
        this.declareDatatypes();
        this.declareAuxiliaryConstants();
        this.declareClock();
    }
    
    public void caseACspAssert(ACspAssert node) {
    	System.out.println("ProcessTranslatorVisitor :: caseACspAssert -> Ignoring Assertion...");
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition definition) {
        String name = definition.getName().toString().trim();
        this.setCurrentDefinition(name.toString().trim());
        
        if (name.equals(HandelC.MAINP)) {
            PExpression expression = definition.getExpression();
        	this.declareMainFunction(expression);
        } else if (!this.recursion.getMutualRecursion()) {
            ACspParameters parameters = (ACspParameters)definition.getParameters();
            PExpression expression = definition.getExpression();
            if (parameters == null) {
                parameters = new ACspParameters();
            }
            
            this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.PROCESS + HandelC.SPACE + name);
            this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
            this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + name;
            this.hcCode = this.hcCode + this.translateFormalArguments(parameters,this.currentDefinition);
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

            if(this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
                // It is a tail recursion

                // int 1 KEEP_LOOPING;
                this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
                this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
                this.hcCode = this.hcCode + HandelC.END_COMMAND;
                
                // Declare argument copies
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_DECLARATION,IS_NOT_ARGUMENT);
                
                // Initialise them
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,this.currentDefinition,IS_NOT_DECLARATION,IS_NOT_ARGUMENT);
                
                // KEEP_LOOPING = true;
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
                
                // Loop
                this.hcCode = this.hcCode + HandelC.WHILE ;
                this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.KEEP_LOOPING + HandelC.PAR_END;
                this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
                this.hcCode = this.hcCode + HandelC.SEQUENCE + HandelC.BLOCK_BEGIN;
                expression.apply(this);
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
            } else {
                expression.apply(this);
            }
            this.hcCode = this.hcCode + HandelC.BLOCK_END;

        }
    }
    
    public void caseACspConstantDefinition(ACspConstantDefinition node) {
    	// DO NOT REMOVE THIS CODE
    	// If you remove it, then the visitor will visit the 
    	// constant definition, instead of doing nothing 
    }

    public void caseACspFunctionDefinition(ACspFunctionDefinition node) {
    	// DO NOT REMOVE THIS CODE
    	// If you remove it, then the visitor will visit the 
    	// function definition, instead of doing nothing 
    }
    
    public void caseACspConstantCallExpr(ACspConstantCallExpr node){
		String name = node.getName().getText();
    	if (node.getCspType().toString().equals(CspUtil.getBoolType())) {
    		// Is a boolean constant
	        this.hcCode = this.hcCode + node.toString().trim();
    	} else {
    		if (this.inputVariables.contains(name)) {
                this.hcCode = this.hcCode + name;
            } else if (this.procsInfo.isArgumentOf(name,this.currentDefinition)) {
                // It is an argument of this process
                if (this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)
                    || this.recursion.getMutualRecursion()) {
                    // We are translating a tail recursive process or a mutual recursion
                    // Replace this argument by its copy
                    this.hcCode = this.hcCode + this.copyName(this.currentDefinition,name,IS_NOT_DECLARATION);
                } else {
                    // Keep original name
                    this.hcCode = this.hcCode + name;
                }
            } else if (this.types.isATypeBranch(name)) {
                // this is a reference to an element of a datatype
                this.hcCode = this.hcCode + name;
            } else if (this.types.isTag(name)) {
                // this is a reference to a tag of a datatype
                
                String parent = this.types.getTagParent(name);
                String lookupTableName = HandelC.getInstance().getLookupTableName(parent,name);
                
                Integer tagSize = new Integer(this.types.getTagValues(name).size());
                
                this.tagsArgsLeftStack.push(tagSize);
                this.tagsArgsTotalStack.push(tagSize);
                this.tagsTypesStack.push((List)this.types.getTagTypes(name));
                
                this.hcCode = this.hcCode + lookupTableName;
                this.openArrayAccessTypeConstructor();
            } else if (this.constants.isConstant(name)){
                // this is a reference to a global constant
                this.hcCode = this.hcCode + name;
            }
    	}
    }

    public void caseACspFunctionCallExpr(ACspFunctionCallExpr node){
        // this is a constant
        this.hcCode = this.hcCode + node.getName().getText();
        // it might have arguments, so...
        if (node.getArguments() != null) {
            this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
            node.getArguments().apply(this);
            this.hcCode = this.hcCode + HandelC.PAR_END;
        }
    }	
    
    public void caseACspIfProcess(ACspIfProcess node) {
        this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
        node.getCondition().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getThenProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getElseProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;
    }

    public void caseACspGuardedProcess(ACspGuardedProcess node) {
        this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN;
        node.getGuardExpr().apply(this);
        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        node.getProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        this.hcCode = this.hcCode + this.getStopImpl();
        this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;
    }
    
    public void caseACspPrefixProcess(ACspPrefixProcess node){
    	this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
        
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        LinkedList proj = ((ACspEvent)node.getEvent()).getProjection();
        PHelp proc = node.getProcess();

    	String inputVar = this.translatePrefix(event,comm,proj,IS_NOT_PRI_ALT);
        
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        proc.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;

        if (inputVar != null) {
            this.inputVariables.remove(inputVar);
        }

    }
    
    public void caseACspProjection(ACspProjection node) {
    	
        if (this.tagsArgsLeftStack.isEmpty()) {
            // this is a real projection
            this.nextArrayAccessChannelSync();
            node.getCallExpr().apply(this);
        } else {
            // this may be a tag access
            // Check how many arguments left for the current tag
            int argsLeft = ((Integer)this.tagsArgsLeftStack.pop()).intValue();
            int argsTotal = ((Integer)this.tagsArgsTotalStack.pop()).intValue();
            
        	// If it is zero, we should close the parenthesis
            // and remove the top of the stack of total number of arguments
            if (argsLeft == 0) {
                // For every finished argument of a datatype tag
                while(argsLeft == 0) {
                    // Close the array invocation
                    this.closeArrayAccessTypeConstructor();

                	// If there are still arguments left
                    if (!tagsArgsLeftStack.isEmpty()) {

                        // Let us check the next one
                        argsLeft = ((Integer)this.tagsArgsLeftStack.pop()).intValue();

                        // We have finished one tag constructor, remove it from the top of the stack
                        this.tagsArgsTotalStack.pop();
                    }
                }

                // Furthermore, If there are any further tags being analysed
                // and there are yet arguments left
                if (argsLeft > 0) {
                    // we have done one more argument from the previous tag
                    // Reduce the top of the stack of analysed arguments by one
                    this.tagsArgsLeftStack.push(new Integer(argsLeft-1));
                    this.tagsArgsTotalStack.push(new Integer(argsTotal));
                } else {
                    // Otherwise, if we finished the all the arguments, just remove the 
                    // last tag from the tags stack
                	this.tagsArgsLeftStack.pop();
                    this.tagsArgsTotalStack.pop();
                }
            } else {
                // If it is not zero, we do not need to close any parenthesis
                // But we have done one more argument from the current tag
                // Reduce the top of the stack of analysed arguments by one
                this.tagsArgsLeftStack.push(new Integer(argsLeft - 1));
                this.tagsArgsTotalStack.push(new Integer(argsTotal));
            }
            
            // If we are analysing tags
            // And there are still other arguments to be analysed
            if (argsLeft < argsTotal) {
                this.nextArrayAccessTypeConstructor();
            }

            // If we are not analysing any tags anymore
            // this is actually a real projection
            if (this.tagsArgsLeftStack.isEmpty()) {
                this.nextArrayAccessChannelSync();
            }
            
            // Analyse the remaining of the projection
            node.getCallExpr().apply(this);

            // After analysing the projection, let us check if we have finished
            // any tag analysis
            
            // If there are tags to be analysed yet
            if (!this.tagsArgsLeftStack.isEmpty()) {
                // For every tag whose analysys has finished
                while (!this.tagsArgsLeftStack.isEmpty() && ((Integer)this.tagsArgsLeftStack.peek()).intValue() == 0) {
                    // Remove the corresponding value from the stack of total 
                    // number of arguments, and close its parenthesis
                    this.tagsArgsLeftStack.pop();
                    this.tagsArgsTotalStack.pop();
                    
                    this.closeArrayAccessTypeConstructor();
                    
                }
            }
        }
    }
    
    
    public void caseACspInputComunication(ACspInputComunication node) {
        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
        this.hcCode = this.hcCode + HandelC.SPACE; 
        node.getParam().apply(this);
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,Directives.CHANNEL_IN); 
    }
    
    public void caseACspOutputComunication(ACspOutputComunication node) {
        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
        this.hcCode = this.hcCode + HandelC.SPACE;
        node.getExpression().apply(this);
        this.directives.putProcChannelInOut(this.currentDefinition,this.currentPrefixing,Directives.CHANNEL_OUT); 
    }

    private String getStopImpl() {
    	return HandelC.INEXISTENT_CHANNEL
				+ HandelC.INPUT_COMMUNICATION
					+ HandelC.SYNC_IN + HandelC.END_COMMAND;    
	}
    
    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
        String name = node.getName().getText();
        
        if (name.equals(CspUtil.getSkip())) {
    	// It is SKIP
    	} else if (name.equals(CspUtil.getStop())) {
    	// It is STOP
            this.hcCode = this.hcCode + this.getStopImpl();
    	} else if (!this.recursion.getMutualRecursion()) {
            // It is not mutual recursion, behave normally
            if(name.equals(this.currentDefinition) && 
                this.recursion.getTailRecursiveProcesses().contains(this.currentDefinition)) {
                // It is a tail recursion invocation
                // Save values of the copies
                this.translateRealArgumentsCopies(node,this.currentDefinition);
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
            } else {
                this.hcCode = this.hcCode + name;
                this.translateRealArguments(node,this.currentDefinition);
                this.hcCode = this.hcCode + HandelC.END_COMMAND;
            }
        } else {
            // It is a mutual recursion

            // Save values of the copies
            this.translateRealArgumentsCopies(node,this.currentDefinition);

            // Change the value of the program counter
            if (!this.isTranslatingMain()) {
                this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;
                this.hcCode = this.hcCode + HandelC.SPACE;
                this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE;
                this.hcCode = this.hcCode + name + HandelC.END_COMMAND;

                // Indicate the change by setting KEEP_LOOPING to true
                this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
            } else {
                // Translating the main method Invoke MUTUAL_REC
                this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION 
                			+ HandelC.PAR_BEGIN + name;
                this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.END_COMMAND;
            }
        }
    }
    
    public void caseACspDotExpr(ACspDotExpr node){
        node.getCallExpr().apply(this);
    	
    	List projections = node.getProjection();
    	for (Iterator iter = projections.iterator(); iter.hasNext();) {
			PProjection element = (PProjection) iter.next();
			element.apply(this);
		}
    }
    
    public void caseACspLinkedArgumentList(ACspLinkedArgumentList node) {
        node.getArgument().apply(this);
        this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
        node.getNext().apply(this);
    }

    /*
    public void caseACspArgument(ACspArgument node) {
        this.hcCode = this.hcCode + node.getExpression().a;
    }
    */
    

    private String keepLooping(String value) {
        String answer = HandelC.KEEP_LOOPING + HandelC.SPACE;
        answer = answer + HandelC.ASSIGNMENT + HandelC.SPACE + value + HandelC.END_COMMAND;
        
        return answer;
    }

    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.translateConcurrency(node);
    }
    
    public void caseACspInternalChoiceProcess(ACspInternalChoiceProcess node) {
		if (this.directives.getIntChoice().equals(Directives.COMPILE_TIME)) {
			node.getLeftProcess().apply(this);
		} else if (this.directives.getIntChoice().equals(Directives.RUN_TIME)) {
			this.hcCode = this.hcCode + HandelC.SEQ + HandelC.BLOCK_BEGIN
						+ this.updateRandomVar() + HandelC.END_COMMAND
						+ HandelC.IF + HandelC.PAR_BEGIN
						+ HandelC.PAR_BEGIN 
						+ HandelC.RANDOM_VAR + HandelC.MATH_REMAINDER + "2" + HandelC.PAR_END 
						+ HandelC.REL_EQUALS + HandelC.MATH_ZERO + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
			node.getLeftProcess().apply(this);
			this.hcCode = this.hcCode + HandelC.BLOCK_END 
						+ HandelC.ELSE + HandelC.BLOCK_BEGIN; 
			node.getRightProcess().apply(this);
			this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.BLOCK_END; 
		}
    }
    
    private String updateRandomVar() {
    	return HandelC.RANDOM + HandelC.PAR_BEGIN + HandelC.RANDOM_VAR 
    		+ HandelC.PAR_END;
    }
    
    public void caseACspExternalChoiceProcess(ACspExternalChoiceProcess node) {

        InitialsExtractorVisitor initialsExtractor = new InitialsExtractorVisitor();
        node.apply(initialsExtractor);
        List keys = initialsExtractor.keyList();
       
        // Declare the input variables (if any)
        for(int i=0; i < keys.size(); i++) {
            PEvent event = (PEvent)keys.get(i);
            LinkedList comm = initialsExtractor.getCommunication(event);
            this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
            this.declareInputVar(comm);
        }

        if (this.useSemaphors) {
        	this.translateExternalChoiceUsingSemaphors(initialsExtractor);
    	} else {
    		this.translateCases(
    				initialsExtractor,
    				initialsExtractor.keyList(),
    				initialsExtractor.keyList().size(),
					"",
					"",
					new Vector(),
					DO_NOT_USE_DEFAULT);
    	}
    	
    }

    private void translateExternalChoiceUsingSemaphors(InitialsExtractorVisitor initials) {
        // Declare the channel set that stores the events that happened
    	int nextId = MathUtil.getNextInt();
    	String doneEvents = HandelC.getInstance().getDoneEventsVarName(nextId);
    	String hasJoined = HandelC.getInstance().getHasJoinedVarName(nextId);

    	this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
    	
    	// Declare the local DONE_EVENTS_i
    	this.hcCode = this.hcCode + HandelC.getInstance().getSetVarOf(HandelC.CHANNEL)
					+ HandelC.SPACE + doneEvents + HandelC.SPACE + HandelC.END_COMMAND;

    	// Declare the local HAS_JOINED_i
    	this.hcCode = this.hcCode + HandelC.BOOLEAN
					+ HandelC.SPACE + hasJoined 
					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL)+ HandelC.ARRAY_END
					+ HandelC.SPACE + HandelC.END_COMMAND;

    	// Initialise the local DONE_EVENTS_i
    	this.hcCode = this.hcCode + doneEvents + HandelC.SPACE 
					+ HandelC. ASSIGNMENT + HandelC.SPACE 
					+ HandelC.getInstance().getEmptySetVarOf(HandelC.CHANNEL)
					+ HandelC.END_COMMAND;

    	List<String> allChannels = this.channels.getChannels();

    	// Initialise the local HAS_JOINED_i
    	for (int i=0; i< allChannels.size(); i++) {
    		String channel = allChannels.get(i);
        	this.hcCode = this.hcCode + this.getHasJoinedElement(hasJoined, channel) 
        				+ HandelC.SPACE 
						+ HandelC. ASSIGNMENT + HandelC.SPACE 
						+ HandelC.LOGICAL_FALSE
						+ HandelC.END_COMMAND;
    	}
    	
    	// List of channels taking part in the choice
    	List<String> choiceChannels = initials.getKeyNames();

    	List semaphorsTaken = new Vector<String>();
    	List<PEvent> unblockedEvents = new Vector<PEvent>();
    	
    	// For every channel name C that is initial
    	for (int i=0; i<choiceChannels.size(); i++) {
    		String channel = choiceChannels.get(i);
    		
            //HAS_JOINED_0[chan_c1] = TRY_JOIN_SEMAPHOR(BRANCH_ID,chan_c1);
    		this.hcCode = this.hcCode + this.getHasJoinedUpdate(hasJoined,channel);
    		
    		
            // while(IS_EMPTY_SET(DONE_EVENTS_0) && !HAS_JOINED_0[chan_c1]){
    		this.hcCode = this.hcCode + HandelC.WHILE + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_SET
    				+ HandelC.PAR_BEGIN + doneEvents + HandelC.PAR_END + HandelC.SPACE
    				+ HandelC.LOGICAL_AND + HandelC.SPACE
    				+ HandelC.LOGICAL_NOT + this.getHasJoinedElement(hasJoined, channel)
    				+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;

    		// if this is not the first channel in the choice
    		// then the previous channels is already unblocked
    		if (i > 0) {
    			unblockedEvents.add((PEvent)initials.keyList().get(i-1));
    		}

    		this.translateCases(initials,
    							unblockedEvents,
    							i + 1,
    							doneEvents,
    							hasJoined,
    							ListUtil.clone(semaphorsTaken),
    							USE_DEFAULT);

    		// Now, you have taken this channels's semaphor
			semaphorsTaken.add(channel);
			
            //HAS_JOINED_0[chan_c1] = TRY_JOIN_SEMAPHOR(BRANCH_ID,chan_c1);
    		this.hcCode = this.hcCode + this.getHasJoinedUpdate(hasJoined,channel);

    		this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}

    	// After taking all channels semaphors(if needed), offer all of them
    	
		// IF(IS_EMPTY_SET(DONE_EVENTS_0)){
    	this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN + HandelC.IS_EMPTY_SET
					+ HandelC.PAR_BEGIN + doneEvents + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
    	
    	// Add the last channel's semaphor which was taken in the prevuious loop
		unblockedEvents.add((PEvent)initials.keyList().get(choiceChannels.size()-1));
		this.translateCases(initials,
							unblockedEvents,
							choiceChannels.size(),
							doneEvents,
							hasJoined,
							semaphorsTaken,
							DO_NOT_USE_DEFAULT);
		this.hcCode = this.hcCode + HandelC.BLOCK_END;

    	this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }

    /*
     * Translates the cases of an external choice
     * @param initials The extractor containing the information of the initials taking part in the choice
     * @param unblockedEvents The list of events that may happen in the choice
     * 		  If no semaphor is being used, this is the list of keys from the initial extractor
     * 		  Otherwise it is initially true
     * @param checked The number of events that we have checked if it is a forced event or not
     * 		  If no semaphor is being used, this is the the size of the list of keys from the initial extractor
     * 		  Otherwise it is initially true
     * @param doneEvents Name of the variable that is the set of channels that happened
     * 			this variable controls the loop of delay
     * 		  If no semaphor is being used, this is not used
     * @param hasJoined Name of the variable that indicates for each channels if it has joined
     * 			the semaphor
     * 		  If no semaphor is being used, this is not used
     * @param semaphorsTaken List of semaphors that have been taken; it is used to release the
     * 			semaphors when some event happens
     * 		  If no semaphor is being used, this is always the empty list
     * 		  Otherwise it is only initially empty
     */
    private void translateCases(InitialsExtractorVisitor initials, 
    							List unblockedEvents,
    							int checked,
    							String doneEvents,
    							String hasJoined,
    							List semaphorsTaken,
    							boolean useDefault) {

    	if (!(checked == initials.getKeyNames().size())) {
    		// Build if-then-else that checks which channels are forced interleaving 

    		// IF part
    		String nextChannel = initials.getKeyNames().get(checked);
    		this.hcCode = this.hcCode + this.getHasJoinedUpdate(hasJoined,nextChannel);
    		this.hcCode = this.hcCode + HandelC.IF + HandelC.PAR_BEGIN 
    					+ HandelC.LOGICAL_NOT + this.getHasJoinedElement(hasJoined, nextChannel) 
    					+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;
			
			this.translateCases(initials, 
								unblockedEvents,
								checked+1, 
								doneEvents, 
								hasJoined,
								semaphorsTaken,
								useDefault); 
			
			this.hcCode = this.hcCode + HandelC.BLOCK_END;
			
    		// ELSE part

			this.hcCode = this.hcCode + HandelC.ELSE + HandelC.BLOCK_BEGIN;

			Object currentEvent = initials.keyList().get(checked); 
			unblockedEvents.add(currentEvent);
			semaphorsTaken.add(nextChannel);

			this.translateCases(initials, 
								unblockedEvents,
								checked+1, 
								doneEvents, 
								hasJoined,
								semaphorsTaken,
								useDefault); 
			
			this.hcCode = this.hcCode + HandelC.BLOCK_END;

			unblockedEvents.remove(currentEvent);
			checked--;
    	} else {
    		// Finished building the alternation, translate the switch now
    		
        	// Translate the cases
            this.hcCode = this.hcCode + HandelC.ALTERNATION;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            
            // Declare the cases
            for(int i=0; i < unblockedEvents.size(); i++) {
                this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                PEvent event = (PEvent)unblockedEvents.get(i);
                LinkedList comm = initials.getCommunication(event);
                LinkedList proj = initials.getProjection(event);
                
                this.translatePrefix(event,comm,proj,IS_PRI_ALT);
                this.hcCode = this.hcCode + HandelC.SPACE + HandelC.CASE_SEP + HandelC.BLOCK_BEGIN;

            	// ONLY if we are using semaphors
                if (this.useSemaphors) {
                	// Adds the channel to the set of channels that happened
                	this.hcCode = this.hcCode 
                	+ doneEvents + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                	+ HandelC.SET_UNION
    				+ HandelC.PAR_BEGIN
    				+ doneEvents
    				+ HandelC.SEP_ARGUMENTS
    				+ HandelC.getInstance().getSetVarOf(HandelC.getInstance().getChannelName(this.currentPrefixing))
    				+ HandelC.PAR_END + HandelC.END_COMMAND;
                	
                	
	            	// Release the semaphors that have been taken
	                // Notice that if we are not using semaphors, then this list is empty
	            	for (int j=0; j<semaphorsTaken.size(); j++) {
	                	this.hcCode = this.hcCode 
	                				+ this.getLeaveSema((String)semaphorsTaken.get(j))
	                				+ HandelC.END_COMMAND;
	            	}
                }
                
                Node process = initials.getProcess(event);
                process.apply(this);
                this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND + HandelC.BREAK + HandelC.END_COMMAND;
            }
            if (this.useSemaphors && useDefault) {
            	// If we are using semaphors, we have a loop and the default case is delay;
            	this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP 
            				+ HandelC.DELAY + HandelC.END_COMMAND
            				+ HandelC.BREAK + HandelC.END_COMMAND;
            }
            
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
    	}
    }

    public void caseACspSequentialProcess(ACspSequentialProcess node){
        
        this.hcCode = this.hcCode + HandelC.SEQUENCE;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        node.getLeftProcess().apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
        node.getRightProcess().apply(this);

        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }
    
    private String typeArgument(String argument, String process){
        String code = "";
        if (this.directives.containsProcArguments(process) 
            && this.directives.getProcArguments(process).containsKey(argument)) {
            String type = (String)this.directives.getProcArguments(process).get(argument);
            code = code + type + argument;
        } else {
            throw new UndefinedArgumentTypeException(argument,process);
        }
        
        return code;
    }

    // if declaration is true it declares the variable, otherwise it initialises it
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String typeArgumentCopy(String argument, String process, 
                                        boolean declaration, boolean isArgument){
        String code = "";
        if (this.directives.containsProcArguments(process) 
            && this.directives.getProcArguments(process).containsKey(argument)) {
            String type = (String)this.directives.getProcArguments(process).get(argument);
            String copy = this.copyName(process,argument,declaration);
            if (declaration) {
                // Declare copy
                String decl = type + HandelC.SPACE + copy;
                if (isArgument) {
                    // As a function argument
                    code = code + HandelC.SEP_ARGUMENTS + HandelC.SPACE + decl;
                } else {
                    // As a local variable
                    code = code + decl + HandelC.END_COMMAND;
                }
            } else {
                // Initialise Copy
                code = code + copy + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE;
                code = code + argument + HandelC.END_COMMAND;
            }
        } else {
            throw new UndefinedArgumentTypeException(argument,process);
        }
        return code;
    }
    
    // Return the name of the copy of the local variable together with any
    // possible array access
    // Receives the process name P, the variable name N, and a boolean that indicates
    // if this is a declaration or not
    // Returns the name P_local_N[
    private String copyName(String procName, String varName, boolean declaration) {
        String copyName = procName + HandelC.LOCAL_COPY + varName;
        return copyName;
    }  
    
    private void declareChannels(){
    	// Declaring the channels names chan_c
        List<String> names = this.channels.getChannelNames();
        TypesVisitor typesVisitor = new TypesVisitor(directives,this.constants);
        typesVisitor.declareType(HandelC.CHANNEL, names);
        this.hcCode = typesVisitor.getHcCode() + this.hcCode;

    	// Declaring the channel variables
    	Enumeration keys = this.directives.getAllChannels();
        while(keys.hasMoreElements()) {
            String chan = (String)keys.nextElement();
            String decl = this.channels.getChannelDeclaration(chan);
            this.hcCode = decl + HandelC.END_COMMAND + this.hcCode;
        }
    	// Declaring the inexistent channel
        String inexistent = HandelC.CHAN_DECL + HandelC.SPACE 
        				+ HandelC.SYNC_TYPE + HandelC.SPACE 
        				+ HandelC.INEXISTENT_CHANNEL + HandelC.END_COMMAND;
        this.hcCode = inexistent + this.hcCode;
    	        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CHANNELS_DECL) + this.hcCode;
        
    }
    
    private void declareBranches(){
    	// Declaring the branches
        int numberOfBranches = this.parallelismAnalyser.getNumberOfBranches();
        List<String> newnames = new Vector();
        for(int i=numberOfBranches-1; i >=0; i--) {
        	newnames.add(HandelC.getInstance().getBranchName(i));
        }
        TypesVisitor typesVisitor = new TypesVisitor(directives,this.constants);
        typesVisitor.declareType(HandelC.BRANCH, newnames);
        this.hcCode = typesVisitor.getHcCode() + this.hcCode;

        this.hcCode = HandelC.getInstance().generateComment(Messenger.FORCED_INTERLEAVING_DECLARATION) + this.hcCode;
    }

    private void declareProcessesPrototypes(){
        Enumeration keys = this.procsInfo.keys();
        
        while(keys.hasMoreElements()) {
        	
            String process = ((String)keys.nextElement()).trim();

            if (!process.equals(HandelC.MAINP)) {
                String declaration = HandelC.INLINE + HandelC.SPACE            
                					+ HandelC.VOID + HandelC.SPACE + process;
				if (this.procsInfo.containsProcArguments(process)) {
				   ACspParameters parameters = this.procsInfo.getProcArguments(process);
				   declaration = declaration + translateFormalArguments(parameters,process);
				}
				declaration = declaration + HandelC.END_COMMAND;
				this.hcCode = declaration + this.hcCode;
            }
        }
        this.hcCode = HandelC.getInstance().generateComment(Messenger.PROCESSES_PROTOTYPES) + this.hcCode;
    }

    private void declareDatatypes(){
        this.hcCode = this.types.getHcCode() + this.hcCode;
    }
    
    private void declareClock(){
        String declaration = HandelC.getInstance().generateComment(Messenger.CLOCK_DECLARATION);
        declaration = declaration + HandelC.CLOCK_DEF;
        declaration = declaration + HandelC.COMMENT_LINE_BREAK;
        
        this.hcCode = declaration + this.hcCode;
    }

    private void declareForcedInterleavingVariablesSemaphors(){
        String declaration = HandelC.getInstance().generateComment(Messenger.FORCED_INTERLEAVING_SEMAPHORS);
        declaration = declaration 
        			+ HandelC.SEMAPHOR + HandelC.SPACE + HandelC.getInstance().getSemaVarOf(HandelC.FORCED_SYNC_LUT)
        			+ HandelC.END_COMMAND;
        declaration = declaration 
					+ HandelC.SEMAPHOR + HandelC.SPACE + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT)
					+ HandelC.END_COMMAND;
        
        this.hcCode = declaration + this.hcCode;
    }
    
    private void declareForcedSyncLUT(){
        String declaration = HandelC.STATIC + HandelC.SPACE 
        					+ HandelC.getInstance().getSetVarOf(HandelC.BRANCH) + HandelC.SPACE 
        					+ HandelC.FORCED_SYNC_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.BRANCH) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // Automatic declaration of the look up table
        // FIRST: Channel names are declared in the reverse order from the list of names
        //			returned by the channels visitor
        //			For instance, if the channels visitor returns [c2, c1, c]
        //			the channels names are declared
        //			#define chan_c 0
        //			#define chan_c1 1
        //			#define chan_c2 2
        
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.BLOCK_BEGIN + HandelC.SPACE;
        	Branch2BranchSetLUT branchesLUT = 
        			this.parallelismAnalyser.getForcedInterleavingLUT().get(channelName);
        	
            for (int branch=0; branch < this.parallelismAnalyser.getNumberOfBranches(); branch++) {
            	declaration = declaration + HandelC.getInstance().generateComment(HandelC.getInstance().getBranchName(branch));
            	Set<Integer> excludedBranches = branchesLUT.get(new Integer(branch));
                List excludedSetNames = new Vector();
                for (Integer excluded : excludedBranches) {
					excludedSetNames.add(HandelC.getInstance().getSetVarOf(HandelC.getInstance().getBranchName(excluded.intValue())));
				}
				declaration = declaration + HandelC.getInstance().getSetUnionCode(excludedSetNames,HandelC.BRANCH);
	        	if (branch < this.parallelismAnalyser.getNumberOfBranches() - 1) {
	            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
	        	}
            }
            
            declaration = declaration + HandelC.SPACE + HandelC.BLOCK_END;
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
    }

    private void declareCurrentSyncStatusLUT(){
        String declaration = HandelC.getInstance().getSetVarOf(HandelC.BRANCH) + HandelC.SPACE 
        					+ HandelC.CURRENT_SYNC_STATUS_LUT 
        					+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(HandelC.CHANNEL) + HandelC.ARRAY_END
        					+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE + HandelC.BLOCK_BEGIN;
        
        // Automatic declaration of the look up table
        // FIRST: Channel names are declared in the reverse order from the list of names
        //			returned by the channels visitor
        //			For instance, if the channels visitor returns [c2, c1, c]
        //			the channels names are declared
        //			#define chan_c 0
        //			#define chan_c1 1
        //			#define chan_c2 2
        
        List<String> channelNames = ListUtil.reverse(this.channels.getChannelNames()); 
        
        for (int c=0; c < channelNames.size(); c++) {
        	String channelName = channelNames.get(c);
        	declaration = declaration + HandelC.getInstance().generateComment(channelName);
        	declaration = declaration + HandelC.getInstance().getEmptySetVarOf(HandelC.BRANCH);
        	if (c < channelNames.size() - 1) {
            	declaration = declaration + HandelC.SEP_ARGUMENTS + HandelC.SPACE;
        	}
        }
        declaration = declaration + HandelC.BLOCK_END + HandelC.END_COMMAND;

        this.hcCode = declaration + this.hcCode;
    }
    
    private void declareAuxiliaryConstants(){
        this.hcCode = HandelC.SYNC_DEF + HandelC.SYNC_OUT_DEF + HandelC.SYNC_IN_DEF + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.SYNC_VALUE_DECL) + this.hcCode;
    }
    
    public void declareAuxiliaryMacros() {

        /*
        macro expr SET_INTER(SET_S,SET_T) = (SET_S & SET_T);
        */
    	String setInter = HandelC.MACRO_EXP + HandelC.SPACE 
				        + HandelC.SET_INTER + HandelC.PAR_BEGIN
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
				        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
				        + HandelC.PAR_BEGIN
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
				        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_AND + HandelC.SPACE
				        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
				        + HandelC.PAR_END
				        + HandelC.END_COMMAND;
		this.hcCode = setInter + this.hcCode;

        /*
        macro expr SET_DIFF(SET_S,SET_T) = (SET_S & (~SET_T));
        */
    	String setDiff = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.SET_DIFF + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_AND + HandelC.SPACE
                        + HandelC.PAR_BEGIN
                        + HandelC.LOGICAL_BITWISE_NOT + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
                        + HandelC.PAR_END
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = setDiff + this.hcCode;

        /*
        macro expr SET_UNION(SET_S,SET_T) = (SET_S | SET_T);
        */
    	String setUnion = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.SET_UNION + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.LOGICAL_BITWISE_OR + HandelC.SPACE
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_T) 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = setUnion + this.hcCode;

        /*
        macro expr IS_EMPTY_SET(SET_S) = (SET_S == 0);
        */
    	String isEmptyChannelSet = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.IS_EMPTY_SET + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) + HandelC.PAR_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.SET, HandelC.SET_ARG_S) 
                        + HandelC.SPACE + HandelC.REL_EQUALS + HandelC.SPACE
                        + HandelC.MATH_ZERO 
                        + HandelC.PAR_END
                        + HandelC.END_COMMAND;
        this.hcCode = isEmptyChannelSet + this.hcCode;

        
        /*
         * 	ONLY IF WE HAVE INTERNAL CHOICE DECIDED AT RUNTIME
         *  unsigned 32 random_var = 1;
         *  macro proc random(random_arg) { random_arg = (random_arg<-21) @ (random_arg[12] ^ random_arg[30]) @ (random_arg[10:1]^random_arg[29:20]);} 
        */
        if (this.directives.getIntChoice().equals(Directives.RUN_TIME)) {
	    	String arg = HandelC.getInstance().getArgName(HandelC.INTEGERL, HandelC.INT_ARG_I);
	    	String random = HandelC.MACRO_PROC + HandelC.SPACE + HandelC.RANDOM 
	                        + HandelC.PAR_BEGIN + arg + HandelC.PAR_END
	                        + HandelC.SPACE + HandelC.BLOCK_BEGIN
	                        + arg + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
	                        + HandelC.PAR_BEGIN + arg + HandelC.TAKE_LSB + "21" + HandelC.PAR_END
	                        + HandelC.CONCAT
	                        + HandelC.PAR_BEGIN + arg + HandelC.ARRAY_BEGIN + "12" + HandelC.ARRAY_END 
	                        + HandelC.LOGICAL_BITWISE_XOR
	                        + arg + HandelC.ARRAY_BEGIN + "30" + HandelC.ARRAY_END
	                        + HandelC.PAR_END
	                        + HandelC.CONCAT
	                        + HandelC.PAR_BEGIN + arg + HandelC.ARRAY_BEGIN + "10" + HandelC.BIT_RANGE + "1" + HandelC.ARRAY_END 
	                        + HandelC.LOGICAL_BITWISE_XOR
	                        + arg + HandelC.ARRAY_BEGIN + "29" + HandelC.BIT_RANGE + "20" + HandelC.ARRAY_END
	                        + HandelC.PAR_END + HandelC.END_COMMAND + HandelC.BLOCK_END;
	        this.hcCode = random + this.hcCode;
	    	String declaration = HandelC.UNSIGNED + HandelC.SPACE 
					        + "32" + HandelC.SPACE
					        + HandelC.RANDOM_VAR 
					        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.MATH_ONE + HandelC.END_COMMAND;
	    	this.hcCode = declaration + this.hcCode;
        }
        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.AUX_MACROS) + this.hcCode;

    }

    public void declareAuxiliarySemaphorMacrosAndFunctions() {
        /*
		inline void LEAVE_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C){
		    LOCK_VARIABLE(&CURRENT_SYNC_STATUS_LUT_SEMA);
		    CURRENT_SYNC_STATUS_LUT[CHANNEL_C] = SET_DIFF(CURRENT_SYNC_STATUS_LUT[CHANNEL_C],BRANCH__set_LUT[BRANCH_B]);                   
		    releasesema(CURRENT_SYNC_STATUS_LUT_SEMA);
		}
        */
    	String leaveSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.BRANCH + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") 
                        + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                        + HandelC.SET_DIFF + HandelC.PAR_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END
                        + HandelC.SEP_ARGUMENTS
                        + HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH) + HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.ARRAY_END
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END;
        this.hcCode = leaveSema + this.hcCode;

        /*
		inline void JOIN_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C){
		    boolean HAS_JOINED;
		    HAS_JOINED = TRY_JOIN_SEMAPHOR(BRANCH_B,CHANNEL_C); 
		    while(!HAS_JOINED){
		        HAS_JOINED = TRY_JOIN_SEMAPHOR(BRANCH_B,CHANNEL_C);
		        delay;
		    }
		}
        */
    	String joinSema = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
                        + HandelC.JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.BRANCH + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.SEP_ARGUMENTS 
                        + HandelC.CHANNEL + HandelC.SPACE + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.PAR_END
                        + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.WHILE	+ HandelC.PAR_BEGIN 
                        + HandelC.LOGICAL_NOT + HandelC.HAS_JOINED + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.DELAY + HandelC.END_COMMAND	
                        + HandelC.BLOCK_END
                        + HandelC.BLOCK_END;
        this.hcCode = joinSema + this.hcCode;

        /*
		inline boolean TRY_JOIN_SEMAPHOR(BRANCH BRANCH_B,CHANNEL CHANNEL_C) {
		    boolean hasJoined;
		
		    if (MAY_JOIN(BRANCH_B,CHANNEL_C)) {
		        LOCK_VARIABLE(&FORCED_SYNC_LUT_SEMA);
		        LOCK_VARIABLE(&CURRENT_SYNC_STATUS_SEMA);
		
		        if (MAY_JOIN(BRANCH_B,CHANNEL_C)) {
		            CURRENT_SYNC_STATUS[CHANNEL_C] = SET_UNION(CURRENT_SYNC_STATUS[CHANNEL_C],PAR_BRANCH__set_LUT[BRANCH_B]);                  
		            hasJoined = true;
		        } else {
		            hasJoined = false;
		        }
		        releasesema(CURRENT_SYNC_STATUS_SEMA);
		        releasesema(FORCED_SYNC_LUT_SEMA);
		    } else {
		        hasJoined = false;
		    }
		
		    return hasJoined;
		}
        */
    	String tryJoinSema = HandelC.INLINE + HandelC.SPACE + HandelC.BOOLEAN + HandelC.SPACE
    					+ HandelC.TRY_JOIN_SEMAPHOR 
    					+ HandelC.PAR_BEGIN + HandelC.BRANCH  + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.BRANCH, "B") 
    					+ HandelC.SEP_ARGUMENTS + HandelC.CHANNEL + HandelC.SPACE 
    					+ HandelC.getInstance().getArgName(HandelC.CHANNEL, "C")	
                        + HandelC.PAR_END + HandelC.SPACE + HandelC.BLOCK_BEGIN
                        + HandelC.BOOLEAN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_JOIN + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C")
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.FORCED_SYNC_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN 
                        + HandelC.ADDRESS + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.IF + HandelC.PAR_BEGIN 
                        + HandelC.MAY_JOIN + HandelC.PAR_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.SEP_ARGUMENTS 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C")
                        + HandelC.PAR_END + HandelC.PAR_END + HandelC.BLOCK_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
                        + HandelC.SET_UNION + HandelC.PAR_BEGIN
                        + HandelC.CURRENT_SYNC_STATUS_LUT + HandelC.ARRAY_BEGIN  
                        + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END
                        + HandelC.SEP_ARGUMENTS
                        + HandelC.getInstance().getSingletonSetsLookupTableName(HandelC.BRANCH) + HandelC.ARRAY_BEGIN 
                        + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.ARRAY_END
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_TRUE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.CURRENT_SYNC_STATUS_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.RELEASE_SEMAPHOR + HandelC.PAR_BEGIN
                        + HandelC.getInstance().getSemaVarOf(HandelC.FORCED_SYNC_LUT) 
                        + HandelC.PAR_END + HandelC.END_COMMAND
                        + HandelC.BLOCK_END + HandelC.ELSE + HandelC.BLOCK_BEGIN
                        + HandelC.HAS_JOINED + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.SPACE + HandelC.LOGICAL_FALSE + HandelC.END_COMMAND 
                        + HandelC.BLOCK_END
                        + HandelC.RETURN + HandelC.SPACE + HandelC.HAS_JOINED + HandelC.END_COMMAND
                        + HandelC.BLOCK_END;
        this.hcCode = tryJoinSema + this.hcCode;
        
        /*
         * inline void LOCK_VARIABLE(sema *sema_s) {
    	 * 		while(trysema(*sema_s)==0) delay;
		 * }
         */
    	String lockVariable = HandelC.INLINE + HandelC.SPACE + HandelC.VOID + HandelC.SPACE 
				        + HandelC.LOCK_VAR + HandelC.PAR_BEGIN
				        + HandelC.SEMAPHOR + HandelC.SPACE + HandelC.REFERENCE 
				        + HandelC.getInstance().getArgName(HandelC.SEMAPHOR, HandelC.SET_ARG_S) + HandelC.PAR_END
				        + HandelC.BLOCK_BEGIN
				        + HandelC.WHILE	+ HandelC.PAR_BEGIN
				        + HandelC.TRY_SEMAPHOR + HandelC.PAR_BEGIN + HandelC.REFERENCE 
				        + HandelC.getInstance().getArgName(HandelC.SEMAPHOR, HandelC.SET_ARG_S) + HandelC.PAR_END
				        + HandelC.REL_EQUALS + HandelC.MATH_ZERO + HandelC.PAR_END
				        + HandelC.SPACE + HandelC.DELAY + HandelC.END_COMMAND
				        + HandelC.BLOCK_END;
		this.hcCode = lockVariable + this.hcCode;
        
        /*
         * macro expr MAY_JOIN(BRANCH_B, CHANNEL_C) = 
         * 	(IS_EMPTY_SET(SET_INTER(FORCED_SYNC_LUT[c][b],CURRENT_SYNC_STATUS[c])));
         */
    	String mayJoin = HandelC.MACRO_EXP + HandelC.SPACE + HandelC.MAY_JOIN 
							+ HandelC.PAR_BEGIN + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") 
							+ HandelC.SEP_ARGUMENTS + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") 
							+ HandelC.PAR_END + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
							+ HandelC.PAR_BEGIN
							+ HandelC.IS_EMPTY_SET + HandelC.PAR_BEGIN
							+ HandelC.SET_INTER + HandelC.PAR_BEGIN 
							+ HandelC.FORCED_SYNC_LUT 
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END 
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.BRANCH, "B") + HandelC.ARRAY_END + HandelC.SEP_ARGUMENTS
							+ HandelC.CURRENT_SYNC_STATUS_LUT
							+ HandelC.ARRAY_BEGIN + HandelC.getInstance().getArgName(HandelC.CHANNEL, "C") + HandelC.ARRAY_END 
							+ HandelC.PAR_END + HandelC.PAR_END  + HandelC.PAR_END  + HandelC.END_COMMAND;
		this.hcCode = mayJoin + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.AUX_SEMA_MACROS) + this.hcCode;
    }
    
    private void declareConstantsAndFunctions(){
        this.hcCode = this.constants.getHcCode() + this.hcCode;
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CONSTANTS_AND_FUNCTIONS) + this.hcCode;
        this.declareConstantsAndFunctionsPrototypes();
    }

    private void declareConstantsAndFunctionsPrototypes(){
        Enumeration keys = this.constants.getConstantsNames();
        
        while(keys.hasMoreElements()) {
            String name = (String)keys.nextElement();
            List args = this.constants.getArguments(name);
            
            // macro expr name = expression;
            String macro = HandelC.MACRO_EXP + HandelC.SPACE + name;
            
            if (args.size() > 0) {
                macro = macro + HandelC.PAR_BEGIN;
                for(int i=0; i < args.size(); i++){
                   macro = macro + args.get(i);
                   if (i < args.size() - 1) {
                       macro = macro + HandelC.SEP_ARGUMENTS;
                   }
                }
                macro = macro + HandelC.PAR_END;
            }
            macro = macro + HandelC.END_COMMAND;
            this.hcCode = macro + this.hcCode;
        }
        
        this.hcCode = HandelC.getInstance().generateComment(Messenger.CONSTANTS_AND_FUNCTIONS_PROTOTYPES) + this.hcCode;
    }
    
    
    private void declareMutualRecConstants(){
        for (int i=0; i<this.processesIds.size(); i++) {
        	String pname = this.processesIds.get(i);
        	if (!pname.equals(HandelC.MAINP)) {
	            String nextLine = HandelC.getInstance().generateDefine(this.processesIds.get(i) + HandelC.SPACE + i);
	            this.hcCode = nextLine + this.hcCode;
        	}
        }
        this.hcCode = HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_STATE_DECL) + this.hcCode;
    }
    
    public String getForcedSyncDecl() {
    	return HandelC.BRANCH + HandelC.SPACE + HandelC.BRANCH_ID;
    }
    
    private String translateFormalArguments(ACspParameters parameters, String process){
        String code = HandelC.PAR_BEGIN;

        // If we have forced interleaving, the first argument is the 
        // set FORCED_SYNC of forced interleaving events
        if (this.useSemaphors) {
            code = code + this.getForcedSyncDecl();
            if (parameters.getParamList() != null){
                code = code + HandelC.SEP_ARGUMENTS;
            }
        }
        if (parameters.getParamList() != null){
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
            } else if (paramsList instanceof ACspLinkedParamList) {
                String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
                code = code + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    code = code + this.typeArgument(parameter,process);
                    code = code + HandelC.SEP_ARGUMENTS;
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgument(parameter,process);
            }
        }
        code = code + HandelC.PAR_END;
        return code;
    }

    // if declaration is true it declares the variables, otherwise it initialises them
    // if argument is true, it declares it as a function argument, otherwise
    // it declares it as a local variable
    private String translateFormalArgumentsCopies(ACspParameters parameters, String process, 
                                                    boolean declaration, boolean argument){
        String code = "";
        if (parameters.getParamList() != null){
            // Writing the parameters
            Node paramsList = ((ACspParameters)parameters).getParamList();

            if (paramsList instanceof ACspBaseParamList) {
                String parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = this.typeArgumentCopy(parameter,process,declaration,argument);
            } else if (paramsList instanceof ACspLinkedParamList) {
                String parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                code = this.typeArgumentCopy(parameter,process,declaration,argument);
                while((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    parameter = ((ACspLinkedParamList)paramsList).getParam().toString().trim();
                    code = code + this.typeArgumentCopy(parameter,process,declaration,argument);
                }
                paramsList = ((ACspLinkedParamList)paramsList).getNext();
                parameter = ((ACspBaseParamList)paramsList).getParam().toString().trim();
                code = code + this.typeArgumentCopy(parameter,process,declaration,argument);
            }
        }
        
        return code;
    }
    
    public void translateRealArguments(ACspProcessCallExpr node, String current){
        ACspArguments arguments = (ACspArguments)node.getArguments();
        String name = node.getName().getText();

        this.check(name,arguments,current);
        
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;

        // If we have forced interleaving, the first argument is the 
        // set FORCED_SYNC of forced interleaving events
        if (this.useSemaphors) {
        	this.hcCode = this.hcCode 
        				+ HandelC.BRANCH_ID + HandelC.MATH_SUM
        				+ this.parallelismAnalyser.getNodeBranchId(node);
	        if (arguments != null && arguments.getArgumentList() != null){
	            	this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
	        }
        }
        
        if (arguments != null && arguments.getArgumentList() != null){
    		// Writing the parameters
            Node argsList = ((ACspArguments)arguments).getArgumentList();

            if (argsList instanceof ACspBaseArgumentList) {
                ((ACspBaseArgumentList)argsList).getArgument().apply(this);
            } else if (argsList instanceof ACspLinkedArgumentList) {
                ((ACspLinkedArgumentList)argsList).getArgument().apply(this);
                this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    ((ACspLinkedArgumentList)argsList).getArgument().apply(this);
                    this.hcCode = this.hcCode + HandelC.SEP_ARGUMENTS;
                }
                argsList = ((ACspLinkedArgumentList)argsList).getNext();
                ((ACspBaseArgumentList)argsList).getArgument().apply(this);
            }
        }
        this.hcCode = this.hcCode + HandelC.PAR_END;
    }
    
    public void translateRealArgumentsCopies(ACspProcessCallExpr node, String current){
        ACspArguments arguments = (ACspArguments)node.getArguments();
        String name = node.getName().getText();
        
        this.check(name,arguments,current);
        
        if (arguments != null && arguments.getArgumentList() != null){
            if (this.procsInfo.containsProcArguments(name)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(name);
                
                PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
                PParamList paramsList = parameters.getParamList();
                
                if (argsList instanceof ACspBaseArgumentList) {
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                } else if (argsList instanceof ACspLinkedArgumentList) {
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                    ((ACspLinkedParamList)paramsList).getParam());
                    while((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList){
                        argsList = ((ACspLinkedArgumentList)argsList).getNext();
                        paramsList = ((ACspLinkedParamList)paramsList).getNext();
                        this.translateRealArgumentCopy(name, 
                                                        ((ACspLinkedArgumentList)argsList).getArgument(), 
                                                        ((ACspLinkedParamList)paramsList).getParam());
                    }
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    this.translateRealArgumentCopy(name, 
                                                    ((ACspBaseArgumentList)argsList).getArgument(), 
                                                    ((ACspBaseParamList)paramsList).getParam());
                }
            } else {
                throw new UndefinedProcessArgumentsException(name);
            }
        }
    }

    private void translateRealArgumentCopy(String procName, PArgument arg, PParam param) {
        this.hcCode = this.hcCode + this.copyName(procName,param.toString().trim(),IS_NOT_DECLARATION) + HandelC.SPACE; 
        this.hcCode = this.hcCode + HandelC.ASSIGNMENT + HandelC.SPACE; 
        arg.apply(this);
        this.hcCode = this.hcCode + HandelC.END_COMMAND;
    }

    private String translateForcedSyncSet(){
    	Set<String> forcedChannels = (Set)this.forcedInterleaving.peek();
    	List setNames = new Vector();
    	for(String c:forcedChannels){
    		String singletonSet = HandelC.getInstance().getSetVarOf(HandelC.getInstance().getChannelName(c));
    		setNames.add(singletonSet);
    	}
    	return HandelC.getInstance().getSetUnionCode(setNames,HandelC.CHANNEL);
    }
    
    private void check(String name, ACspArguments arguments, String current){
        if (this.procsInfo.containsProcArguments(name)) {
            ACspParameters parameters = this.procsInfo.getProcArguments(name);
            if(!this.check(parameters,arguments)) {
                throw new WrongNumberOfArgumentsException(current,name);
            }
        } else {
            if (arguments != null && arguments.getArgumentList() != null){
                throw new WrongNumberOfArgumentsException(current,name);
            }
        }
    }

    private boolean check(ACspParameters parameters, ACspArguments arguments) {
        boolean answer = true;
        
        if (arguments != null && arguments.getArgumentList() != null 
            && parameters != null && parameters.getParamList() != null) {
            PArgumentList argsList = ((ACspArguments)arguments).getArgumentList();
            PParamList paramsList = parameters.getParamList();

            if ((paramsList instanceof ACspBaseParamList) && (argsList instanceof ACspBaseArgumentList)) {
                // do nothing!!! They are ok
            } else if ((paramsList instanceof ACspLinkedParamList) 
                        && (argsList instanceof ACspLinkedArgumentList)) {
                while( ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList)
                        && ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList) ){
                    paramsList = ((ACspLinkedParamList)paramsList).getNext();
                    argsList = ((ACspLinkedArgumentList)argsList).getNext();
                }
                boolean moreParams = ((((ACspLinkedParamList)paramsList)).getNext() instanceof ACspLinkedParamList);
                boolean moreArgs = ((((ACspLinkedArgumentList)argsList)).getNext() instanceof ACspLinkedArgumentList);
                if (moreParams != moreArgs) {
                    answer = false;
                }
            } else {
                answer = false;
            }
        } else {
            boolean anyParams = (parameters != null && parameters.getParamList() != null);
            boolean anyArguments = (arguments != null && arguments.getArgumentList() != null);
            if (anyParams != anyArguments) {
                answer = false;
            }
        }
        return answer;
    }
    
    /*
     * Checks if there is an input in the list of communications.
     * Returns the name of the variable if yes and null otherwise.
     */
    private String checkInput(LinkedList comms) {
        String var = null;
        ListIterator iterator = comms.listIterator(0);
        
        while(iterator.hasNext()) {
            Object next = iterator.next();
            if (next instanceof ACspInputComunication) {
                var = ((ACspInputComunication)next).getParam().toString();
            }
        }
        return var;
    }
    
    private String getTryJoinSema(String c){
    	//TRY_JOIN_SEMAPHOR(BRANCH_ID,chan_c)
    	return HandelC.TRY_JOIN_SEMAPHOR + HandelC.PAR_BEGIN
		+ HandelC.BRANCH_ID + HandelC.SEP_ARGUMENTS + HandelC.getInstance().getChannelName(c) + HandelC.PAR_END;
    }

    private String getJoinSema(String c){
        //JOIN_SEMAPHOR(BRANCH_ID,chan_c);
        return HandelC.JOIN_SEMAPHOR + HandelC.PAR_BEGIN
			+ HandelC.BRANCH_ID + HandelC.SEP_ARGUMENTS 
			+ HandelC.getInstance().getChannelName(c)+ HandelC.PAR_END;
    }

    private String getLeaveSema(String c){
        //LEAVE_SEMAPHOR(BRANCH_ID,chan_c);
        return HandelC.LEAVE_SEMAPHOR + HandelC.PAR_BEGIN
			+ HandelC.BRANCH_ID + HandelC.SEP_ARGUMENTS 
			+ HandelC.getInstance().getChannelName(c)+ HandelC.PAR_END;
    }
    
    private String getChannelSemaphor(String chan) {
    	String semaphor = HandelC.getInstance().getSetVarOf(chan);
        return semaphor;
    }

    private String translatePrefix(PEvent event, LinkedList comm, LinkedList proj, 
    								boolean isPriAlt) {
        this.currentPrefixing = ((ACspEvent)event).getChannelName().toString().trim();
        ListIterator commIter = comm.listIterator(0);
        ListIterator projIter = proj.listIterator(0);
        
        String inputVar = null;
        
        this.currentPrefixingTypes = ListUtil.clone(this.channels.getChannelType(this.currentPrefixing));

        /* 
         * The Restrictions Visitor guarantess that we have communications
         * of the form c[.exp]*[?name | !exp] 
        */

        if (!isPriAlt) {
            inputVar = this.declareInputVar(comm);
            if (this.useSemaphors) {
            	this.hcCode = this.hcCode + this.getJoinSema(this.currentPrefixing) + HandelC.END_COMMAND;
            }
        } 
        
        this.hcCode = this.hcCode + this.currentPrefixing;
        
        // Translate the projections
        while(projIter.hasNext()) {
            Object next = projIter.next();
            ((ACspProjection)next).apply(this);
            if(!projIter.hasNext()){
            	this.closeArrayAccessChannelSync();
            }
        }
        
        // Translate the communication
        if (!commIter.hasNext()) {
            if (this.directives.containsProcChannelInOut(this.currentDefinition)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(this.currentDefinition);
                if (channelInOut.containsKey(this.currentPrefixing)){
                    String inOut = (String)channelInOut.get(this.currentPrefixing);
                    if (inOut.equals(Directives.CHANNEL_IN)) {
                        this.hcCode = this.hcCode + HandelC.INPUT_COMMUNICATION;
                        this.hcCode = this.hcCode + HandelC.SYNC_IN;
                    } else if (inOut.equals(Directives.CHANNEL_OUT)) {
                        this.hcCode = this.hcCode + HandelC.OUTPUT_COMMUNICATION;
                        this.hcCode = this.hcCode + HandelC.SYNC_OUT;
                    }
                } else {
                    throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
                }
            } else {
                throw new UndefinedChannelInOutException(this.currentPrefixing,this.currentDefinition);
            }

        } else {
            Object next = commIter.next();
            if (next instanceof ACspInputComunication) {
                ((ACspInputComunication)next).apply(this);
            } else if (next instanceof ACspOutputComunication) {
                ((ACspOutputComunication)next).apply(this);
            }
        }

        if (!isPriAlt && this.useSemaphors) {
        	this.hcCode = this.hcCode + HandelC.END_COMMAND 
        				+ this.getLeaveSema(this.currentPrefixing) + HandelC.END_COMMAND;
        }
        
        this.currentType = "";
        return inputVar;
    }

    // Declares the input variable (if there is any)
    private String declareInputVar(LinkedList comm) {
        ListIterator iterator = comm.listIterator(0);
        
        String var = this.checkInput(comm);
        if (var != null) {
            if (this.channels.containsChannelType(this.currentPrefixing)) {
                String type = this.channels.getChannelCommType(this.currentPrefixing);
                this.hcCode = this.hcCode + type + HandelC.SPACE + var + HandelC.END_COMMAND;
            } else {
                throw new UndefinedChannelTypeException(this.currentPrefixing,this.currentDefinition);
            }
            var = var.trim();
            this.inputVariables.add(var);
        }
        return var;
    }

    // Translates all the processes definitions as an action system like
    // Does not support parametrised processes
    private void translateProcessesDefinitionsMutualRecursion() {
        int numberOfProcesses = this.procsInfo.size();
        int numberOfBits = MathUtil.getNumberOfBits(numberOfProcesses);

        // Declare local copies for all arguments of every parametrised process
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_LOCAL_VARS);
        
        for (int i=0; i< this.processesIds.size(); i++) {
            String name = (String)this.processesIds.get(i);
            if (this.procsInfo.containsProcArguments(name)) {
                ACspParameters parameters = this.procsInfo.getProcArguments(name);
                if (parameters == null) {
                    parameters = new ACspParameters();
                }
                this.hcCode = this.hcCode + this.translateFormalArgumentsCopies(parameters,name,IS_DECLARATION,IS_NOT_ARGUMENT);
            }
        }

        // inline void MUTUAL_REC(
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MUTUAL_REC_METHOD);
        this.hcCode = this.hcCode + HandelC.INLINE + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.MUTUAL_RECURSION + HandelC.PAR_BEGIN;

        // CHANNEL_TYPE_set FORCED_SYNC
        if(this.useSemaphors) {
        	this.hcCode = this.hcCode + this.getForcedSyncDecl() + HandelC.SEP_ARGUMENTS;
        }

        // int numberOfBits PROGRAM_COUNTER
        this.hcCode = this.hcCode + HandelC.INT + HandelC.SPACE;
        this.hcCode = this.hcCode + numberOfBits + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;

        // ){
        this.hcCode = this.hcCode + HandelC.PAR_END + HandelC.BLOCK_BEGIN;
        
        // int 1 KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.BOOLEAN + HandelC.SPACE;
        this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.END_COMMAND;

        // KEEP_LOOPING = true;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_TRUE);
        
        // while(KEEP_LOOPING) {
        this.hcCode = this.hcCode + HandelC.WHILE;
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + HandelC.KEEP_LOOPING;
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        // KEEP_LOOPING = false;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);
        
        // switch(PROGRAM_COUNTER) {
        this.hcCode = this.hcCode + HandelC.SWITCH + HandelC.PAR_BEGIN;
        this.hcCode = this.hcCode + HandelC.PROGRAM_COUNTER;
        this.hcCode = this.hcCode + HandelC.PAR_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;

        // case PROC_NAME: 
        for (int i = 0; i < this.processesIds.size(); i++) {
        	String pname = this.processesIds.get(i).trim();
        	
        	if (!pname.equals(HandelC.MAINP)) {
                this.setCurrentDefinition(pname);

                this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.PROCESS + HandelC.SPACE + this.currentDefinition);
                this.hcCode = this.hcCode + HandelC.CASE + HandelC.SPACE;
                this.hcCode = this.hcCode + this.currentDefinition + HandelC.SPACE + HandelC.CASE_SEP;
                this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
                if(this.procsInfo.containsProcDefinition(this.currentDefinition)) {
                    this.procsInfo.getProcDefinition(this.currentDefinition).apply(this);
                }
                this.hcCode = this.hcCode + HandelC.BREAK + HandelC.END_COMMAND;
                this.hcCode = this.hcCode + HandelC.BLOCK_END;
        	}
        }

        // default: KEEP_LOOPING" = false;
        this.hcCode = this.hcCode + HandelC.DEFAULT + HandelC.CASE_SEP + HandelC.SPACE;
        this.hcCode = this.hcCode + this.keepLooping(HandelC.LOGICAL_FALSE);

        // }}}
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
        
        // Declare the main process
        this.declareMainFunction(this.procsInfo.getProcDefinition(HandelC.MAINP));
    }
    
    // Indicates that we are translating the main method's process 
    private boolean isTranslatingMain() {
        return this.currentDefinition.equals(HandelC.MAINP);
    }

    // Translates both branches of a parallelism or interleaving
    public void translateConcurrency(Node node) {
    	Set newForcedInterleaving = null;
        Node left = null;
        Node right = null;
        
        if (node instanceof ACspSharingParallelProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspSharingParallelProcess)node, this.currentDefinition);
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspAlphabetisedParallelProcess)node, this.currentDefinition);
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
    	} else if (node instanceof ACspInterleaveProcess) {
            newForcedInterleaving = this.parallelismAnalyser.getForcedInterleavingChannels((ACspInterleaveProcess)node, this.currentDefinition);
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
    	}

        if (newForcedInterleaving == null || left == null || right == null) {
        	throw new UnsupportedConstructorException(Messenger.UNKNOWN_PARALLEL_COMPOSITION, this.currentDefinition);
        } else {
    		Set allChannels = new HashSet<String>(this.channels.getChannels());

    		// Push the new forced interleaving into the stack
    		this.forcedInterleaving.push(newForcedInterleaving);

            this.hcCode = this.hcCode + HandelC.PARALLEL;
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            
            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            left.apply(this);
            this.hcCode = this.hcCode + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END + HandelC.END_COMMAND;

            this.hcCode = this.hcCode + HandelC.BLOCK_BEGIN;
            right.apply(this);
            this.hcCode = this.hcCode + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BLOCK_END;
            
            this.hcCode = this.hcCode + HandelC.BLOCK_END;

    		this.forcedInterleaving.pop();
        }
    }

    // when translating the accesses to an element of an array of channel
    // this function is invoked in order to guarantee that, if the type
    // is an integer, it uses the unsigned representation of that
    // integer; otherwise, we could have access to negative numbers
    private void unsignInteger() {
        this.hcCode = this.hcCode + HandelC.PAR_BEGIN + HandelC.UNSIGNED + HandelC.PAR_END + HandelC.PAR_BEGIN;
    }
    
    // Closes the access to the arrays of channels
    private void nextArrayAccessChannelSync() {
        this.closeArrayAccessChannelSync();
        this.openArrayAccessChannelSync();
    }

    // Closes the access to the arrays of channels
    private void closeArrayAccessChannelSync() {

        // this is the first access, do nothing
        if(!this.currentType.equals("")) {
            if(this.currentType.equals(CspUtil.getIntType())) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            this.hcCode = this.hcCode + HandelC.ARRAY_END;
        }
        
    }

    // Opens the access to the arrays of channels
    private void openArrayAccessChannelSync() {

        this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
        if (this.intArrayAccessChannelSync()) { 
            this.unsignInteger(); 
        }
        
        this.removeCurrentPrefixingTypes();
    }
    
    // Indicates that we were accessing an array using integers 
    private boolean intArrayAccessChannelSync() {
        boolean answer = false;
        
        boolean empty = this.currentPrefixingTypes.isEmpty();
        boolean lastType = this.currentPrefixingTypes.size()==1;
        boolean isComm = this.channels.isCommChannel(this.currentPrefixing);
        
        boolean isInt = this.currentPrefixingTypes.get(0).toString().trim().equals(CspUtil.getIntType());
        
        if (!empty) {
            if (lastType) {
                if (!isComm && isInt) {
                    answer = true;
                }
            } else {
                if (isInt) {
                    answer = true;
                }
            }
        }
        
        return answer;
    }
    
    private void removeCurrentPrefixingTypes() {
        if (!this.currentPrefixingTypes.isEmpty()) {
            this.currentType = this.currentPrefixingTypes.get(0).toString().trim();
            this.currentPrefixingTypes.remove(0);
        }
    }

    private void openArrayAccessTypeConstructor(){
        this.hcCode = this.hcCode + HandelC.ARRAY_BEGIN;
        if(!this.tagsTypesStack.isEmpty()) {
            String type = (String)ListUtil.head(((List)this.tagsTypesStack.peek()));
            if(type.equals(HandelC.INTEGERL)) {
                this.unsignInteger();
            }
        }
    }
    
    private void closeArrayAccessTypeConstructor(){
        if(!this.tagsTypesStack.isEmpty() && !(((List)this.tagsTypesStack.peek()).size() == 0)) {
            List types = (List)this.tagsTypesStack.pop();
            String type = (String)ListUtil.head(types);
            if(type.equals(HandelC.INTEGERL)) {
                this.hcCode = this.hcCode + HandelC.PAR_END;
            }
            if (types.size() > 1) {
                this.tagsTypesStack.push(ListUtil.tail(types));
            }
        }
        this.hcCode = this.hcCode + HandelC.ARRAY_END;
        
    }

    private void nextArrayAccessTypeConstructor(){
        this.closeArrayAccessTypeConstructor();
        this.openArrayAccessTypeConstructor();
    }

    private String getHasJoinedUpdate(String hasJoined, String chan) {
    	return hasJoined + HandelC.ARRAY_BEGIN + HandelC.getInstance().getChannelName(chan) + HandelC.ARRAY_END
    				+ HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE
    				+ this.getTryJoinSema(chan) + HandelC.END_COMMAND;
    }

    private String getHasJoinedElement(String hasJoined, String chan) {
    	return hasJoined + HandelC.ARRAY_BEGIN + HandelC.getInstance().getChannelName(chan) + HandelC.ARRAY_END;
    }

    private void declareMainFunction(PExpression expression){
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(Messenger.MAIN_DECLARATION);
        this.hcCode = this.hcCode + HandelC.VOID + HandelC.SPACE + HandelC.MAIN + HandelC.PAR_BEGIN 
        			+ HandelC.PAR_END + HandelC.BLOCK_BEGIN;
        this.setCurrentDefinition(HandelC.MAINP);
        
        if (this.useSemaphors) {
            this.hcCode = this.hcCode + HandelC.BRANCH + HandelC.SPACE + HandelC.BRANCH_ID
            			+ HandelC.END_COMMAND;
            this.hcCode = this.hcCode + HandelC.BRANCH_ID + HandelC.SPACE + HandelC.ASSIGNMENT
							+ HandelC.SPACE + HandelC.MATH_ZERO + HandelC.END_COMMAND;
        }
        expression.apply(this);
        this.hcCode = this.hcCode + HandelC.BLOCK_END;
    }

}
