package csp2hc.visitors;

import java.io.*;
import java.util.*;
import java.awt.event.*;
import javax.swing.*;
import javax.swing.tree.*;

import lmf.formula.csp.node.*;
import lmf.formula.csp.lexer.*;
import lmf.formula.csp.parser.*;
import lmf.formula.csp.analysis.*;

import csp2hc.messages.*;
import csp2hc.util.*;
import csp2hc.exception.*;

public class TypesVisitor extends ExtendedDepthFirstAdapter {
    // Resulting handel-c code
    private String hcCode;

    // Values of the datatypes
    // Type Name:String -> Branch Names:String List
    private Hashtable typeBranches;
    
    // Used constructors in datatypes
    // Maps the name of the constructor to the list of list 
    // of possible values for each type used
    // EXAMPLE datatype T = B.Bool.{0,1}
    // tagValues = { B -> [[true,false],[0,1]] }
    
    // Tag Name:String -> Values:List (List String)
    private Hashtable tagValues;

    // Used constructors in datatypes
    // Maps the name of the constructor to the list of types used
    // EXAMPLE datatype T = B.Bool.Int
    // tagTypes = { B -> [boolean,integer] }
    // Tag Name:String -> Types:List String
    private Hashtable tagTypes;

    // Maps the constructor name to the name of the type it was used in
    // Tag Name:String -> Type where it was used:String
    private Hashtable tagParents;

    // Visitor containing info about the directives
    private DirectivesVisitor directives;

    // Visitor containing info about the constants
    private ConstantsAndFunctionsVisitor constants;
    
    // Current definition
    private String currentDefinition;

    // Branches of the current definition
    private List<String> currentBranches;
    
    // Current tag being analysed
    private String currentTag;

    // Branches created for a given tag in a complex branch
    private List<String> currentTagBranches;

    // Values used for a tag in a complex branch
    private List currentTagValues;

    // Types used for a tag in a complex branch
    private List currentTagTypes;
    
    // Indicates that we are currently analysing a datatype definition
    private boolean isDatatype;

    // List of skipped datatype definitions, which should be analysed latter
    private List skippedTypes;

    // List of simple branches
    private Set<String> simpleBranches;
    
    public TypesVisitor(DirectivesVisitor directives, 
    					ConstantsAndFunctionsVisitor constants) {
        this.hcCode = "";
        this.typeBranches = new Hashtable();
        this.tagValues = new Hashtable();
        this.tagTypes = new Hashtable();
        this.tagParents = new Hashtable();
        this.directives = directives;
        this.constants = constants;
        this.currentBranches = new Vector();
        this.currentTag = "";
        this.currentTagBranches = new Vector();
        this.isDatatype = false;
        this.skippedTypes = new Vector();
        this.simpleBranches = new HashSet<String>();
    }
    
    public String getHcCode(){ return this.hcCode; }
    
    
    // Returns true if the given name is a branch of a datatype
    public boolean isATypeBranch(String name) {
        boolean answer = false;
        Enumeration keys = this.typeBranches.keys();
        while(!answer && keys.hasMoreElements()) {
            List branches = this.getTypeBranches((String)keys.nextElement());
            for (int i=0; i<branches.size() && !answer ; i++) {
                if (name.equals(branches.get(i))){
                    answer = true;
                }
            }
        }
        return answer;
    }
    
    // Returns true if the given name is a datatype
    public boolean containsTypeBranches(String typeName){ 
        return this.typeBranches.containsKey(typeName); 
    }
    
    // Returns the list os branches of a given a datatype name
    public List getTypeBranches(String typeName){ 
        return (List)this.typeBranches.get(typeName); 
    }

    // Returns true if the name is a tag used in the definition of a datatype
    public boolean isTag(String tag) {
        return this.tagParents.containsKey(tag);
    }

    // Returns the list of values used in the use of the given tag
    public List getTagValues(String tag) {
        return (List)this.tagValues.get(tag);
    }

    // Returns the list of types used in the use of the given tag
    public List getTagTypes(String tag) {
        return (List)this.tagTypes.get(tag);
    }

    // Returns the name of the dataype in whose definition the tag was used 
    public String getTagParent(String tag) {
        return (String)this.tagParents.get(tag);
    }
    
    public void caseStart(Start node) {
        this.declareBool();
        this.declareInt();
        
        node.getPSpecification().apply(this);

        int loops = 0;
        while(this.skippedTypes.size() > 0) {
        	loops++;
            this.skippedTypes = new Vector();
            node.getPSpecification().apply(this);
        }

        // There will be one lookup table for each type
        // Change the method invoked below.
        //this.declareSingletonSetsMacro();
    }
   
    public void caseACspDatatypeDefinition(ACspDatatypeDefinition definition){
        this.currentDefinition = definition.getName().toString().trim();

        this.currentBranches = new Vector();
        this.isDatatype = true;
        
        definition.getBranchList().apply(this);

        if (!this.skippedTypes.contains(this.currentDefinition)) {
            if (!this.typeBranches.containsKey(this.currentDefinition)) {
                
                this.typeBranches.put(this.currentDefinition,this.currentBranches);
                
                this.declareType(this.currentDefinition,this.currentBranches);

                for(int i=0; i<this.currentBranches.size(); i++) {
                	if (!this.simpleBranches.contains(this.currentBranches.get(i))) {
                		HandelC.getInstance().addToolKeyword(this.currentBranches.get(i));
                	}
                }
            }
        }
        this.isDatatype = false;
    }
    
    public void caseACspSimpleBranch(ACspSimpleBranch node) {
    	String branch = node.getTagName().toString().trim();
    	this.simpleBranches.add(branch);
        this.currentBranches.add(branch);
    }
    
    public void caseACspComplexBranch(ACspComplexBranch node) {
        this.currentTag = node.getTagName().toString().trim();
        this.currentTagBranches = new Vector();
        this.currentTagBranches.add(this.currentTag);
        this.currentTagValues = new Vector();
        this.currentTagTypes = new Vector();
        
        this.tagParents.put(this.currentTag,this.currentDefinition);
        
        node.getCspType().apply(this);

        for (int i=0; i<this.currentTagBranches.size(); i++) {
        	String branch = (String)this.currentTagBranches.get(i);
            this.currentBranches.add(branch);
        }
        this.tagValues.put(this.currentTag,this.currentTagValues);
        this.tagTypes.put(this.currentTag,this.currentTagTypes);
    }

    public void caseACspType(ACspType node) {
    	if (node.getName().getText().equals(CspUtil.getBoolType())) {
            if (this.isDatatype) {
                // Creates a list of Strings (one for each boolean)
                // and updates the list of types of the current tag
                List booleanStrings = new Vector();
                booleanStrings.add(HandelC.LOGICAL_FALSE);
                booleanStrings.add(HandelC.LOGICAL_TRUE);
                this.currentTagValues.add(booleanStrings);
                this.currentTagTypes.add(HandelC.BOOLEAN);

                // Updating the tagBranches
                List newTagBranches = new Vector();
                for (int i=0; i < this.currentTagBranches.size(); i++) {
                    String branch = this.currentTagBranches.get(i) + HandelC.UNDERSCORE + HandelC.LOGICAL_TRUE;
                    newTagBranches.add(branch);
                    branch = this.currentTagBranches.get(i) + HandelC.UNDERSCORE + HandelC.LOGICAL_FALSE;
                    newTagBranches.add(branch);
                }
                this.currentTagBranches = newTagBranches;
            }
    	} else if(node.getName().getText().equals(CspUtil.getIntType())) {
            if (this.isDatatype) {
                // Creates a list of Strings (one for each integer)
                // and updates the list of types of the current tag
                List integerStrings = new Vector();
                List range = MathUtil.getSignedIntRangeOrderedBySignedRepresentation(this.directives.getIntBits());
                
                for (int i=0; i < range.size(); i++) {
                    integerStrings.add(((Integer)range.get(i)).intValue()+"");
                }
                this.currentTagValues.add(integerStrings);
                this.currentTagTypes.add(HandelC.INTEGERL);

                // Updating the tagBranches
                List newTagBranches = new Vector();
                for (int i=0; i < this.currentTagBranches.size(); i++) {
                    for (int j=0; j < range.size(); j++) {
                        String branch = this.currentTagBranches.get(i) 
                                        + HandelC.UNDERSCORE 
                                        + MathUtil.getIntName(((Integer)range.get(j)).intValue());
                        newTagBranches.add(branch);
                    }
                }
                this.currentTagBranches = newTagBranches;
            }
    	} else {
            String ref = node.getName().toString().trim();
            if (this.isDatatype) {
                if (this.typeBranches.containsKey(ref)){
                // This is a valid datatype
                    List branches = ListUtil.reverse((List)this.typeBranches.get(ref));
                    this.updateTagBranchesAndTypes(branches);
                    
                    this.currentTagTypes.add(ref);
                } else if (this.constants.containsConstant(ref)) {
                // This a constant definition
                    throw new UnsupportedConstructorException(Messenger.DATATYPE_TAG_WITH_CONSTANT,this.currentDefinition);
                } else {
                // This is a reference to a type which was not yet found
                    if (!this.skippedTypes.contains(this.currentDefinition)) {
                    // I am going to skip the current definition
                        this.skippedTypes.add(this.currentDefinition);
                    }
                    if (this.skippedTypes.contains(ref)){
                    // If I am referencing to some type which has already been skipped
                    // I found a mutually recursive datatype definition
                        this.skippedTypes.add(ref);
                        throw new UnsupportedConstructorException(Messenger.MUTUALLY_RECURSIVE_DATATYPES,this.skippedTypes.toString());
                    }
                } 
            }
    	}
    	
    	
    }

    public void caseACspSetType(ACspSetType node) {
        if (this.isDatatype) {
            throw new UnsupportedConstructorException(Messenger.SET_TYPE,this.currentDefinition);
        }
    }
    
    public void caseACspTupleType(ACspTupleType node) {
        if (this.isDatatype) {
            throw new UnsupportedConstructorException(Messenger.TUPLE_TYPE,this.currentDefinition);
        }
    }
    
    // Updates the list of current tag branches using the new list of branches by 
    // appending the new branches names to every name of the old one
    private void updateTagBranchesAndTypes(List branches){
        this.currentTagValues.add(branches);
        List newTagBranches = new Vector();
        for (int i=0; i < this.currentTagBranches.size(); i++) {
            for (int j=0; j < branches.size(); j++) {
                String branch = this.currentTagBranches.get(i) + HandelC.UNDERSCORE + branches.get(j);
                newTagBranches.add(branch);
            }
        }
        this.currentTagBranches = newTagBranches;
    }
    
    
    private String translateBranch(String typeName, PBranch branch) {
        String answer = branch.toString().trim();
        return answer;
    }

    // Declares a given datatype using the branches given
    public void declareType(String name, List branchNames) {
        String comment = Messenger.DATATYPE_DECLARATION + HandelC.SPACE + name;
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(comment);
        
        this.declareTypeMembers(name,branchNames);
        this.declareTypeCardinality(name,branchNames);
        this.declareTypeSet(name);
        this.declareSingletonSets(name,branchNames);
        this.declareSingletonSetsLookupTable(name,branchNames);
        this.declareLookupTables(name);
    }
    
    // Declares the type
    private void declareTypeMembers(String name, List branchNames) {
        int size = branchNames.size();
        int numberOfBits = MathUtil.getNumberOfBits(size);
        
        // #define name unsigned int numBerOfBits
        String definition = name + HandelC.SPACE + HandelC.UNSIGNED + HandelC.SPACE
                            + HandelC.INT + HandelC.SPACE + numberOfBits;
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(definition);
        
        // #define member integer_representation
        List reverseBranches = ListUtil.reverse(branchNames);

        for (int i = 0; i < size; i++) {
            definition = reverseBranches.get(i) + HandelC.SPACE + i;
            this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(definition);
        }
    }

    // Declares all the singleton sets of a given datatype
    private void declareSingletonSets(String typeName, List branchNames) {
        
        int size = branchNames.size();

        // Declaring sets for each branch
        for (int i = 0; i < size; i++) {
            String line = this.singletonSetDefinition(HandelC.getInstance().getSetVarOf((String)branchNames.get(i)),i,size);
            this.hcCode = this.hcCode + line;
        }
        // Declaring empty set
        String line = this.singletonSetDefinition(HandelC.getInstance().getEmptySetVarOf(typeName),size,size);
        this.hcCode = this.hcCode + line;
    }

    // Declares all the lookup table for the singleton sets of a given datatype
    private void declareSingletonSetsLookupTable(String type, List branches) {
        String tableName = HandelC.getInstance().getSingletonSetsLookupTableName(type);
        String dimension = HandelC.ARRAY_BEGIN + HandelC.getInstance().getCardVarOf(type) + HandelC.ARRAY_END;

        branches = ListUtil.reverse(branches);
        
        String array = HandelC.BLOCK_BEGIN;
        for (int i=0; i < branches.size(); i++) {
            array = HandelC.getInstance().getSetOf(array + branches.get(i));
            if (i < branches.size()-1) {
                array = array + HandelC.SEP_ARGUMENTS;
            }
        }
        array = array + HandelC.BLOCK_END;
        
        String tableDecl = HandelC.STATIC + HandelC.SPACE + HandelC.getInstance().getSetVarOf(type) + HandelC.SPACE + tableName + dimension + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + array + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + tableDecl;
    }
    
    // Declares the cardinality of a given datatype
    private void declareTypeCardinality(String name, List branchNames) {
        String definition = HandelC.getInstance().getCardVarOf(name) + HandelC.SPACE + branchNames.size();
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(definition);
    }

    // Declares the type that declares a set of elements of a given datatype
    private void declareTypeSet(String name) {
        String setsDef = HandelC.getInstance().getSetVarOf(name)
                                + HandelC.SPACE + HandelC.UNSIGNED
                                + HandelC.SPACE + HandelC.INT
                                + HandelC.SPACE + HandelC.getInstance().getCardVarOf(name);

        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(setsDef);
    }
    
    // Declares a singleton set: a set that contains a particular(branchIndex) element of a given datatype
    private String singletonSetDefinition(String setName, int branchIndex, int numberOfBranches) {
        String definition = setName + HandelC.SPACE + HandelC.BINARY_PREFIX;

        for (int j=0; j < numberOfBranches; j++) {
            if (branchIndex == j) {
                definition = definition + "1";
            } else {
                definition = definition + "0";
            }
        }
        return HandelC.getInstance().generateDefine(definition);
    }
    
    // Declares the booleans
    private void declareBool(){
        String comment = Messenger.DATATYPE_DECLARATION + HandelC.SPACE + HandelC.BOOLEAN;
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(comment);

        this.currentDefinition = HandelC.BOOLEAN;
        this.currentBranches = new Vector();
        this.currentBranches.add(HandelC.LOGICAL_TRUE);
        this.currentBranches.add(HandelC.LOGICAL_FALSE);

        this.declareType(this.currentDefinition,this.currentBranches);
    }
    
    // Declares the integers
    private void declareInt(){
        String comment = Messenger.DATATYPE_DECLARATION + HandelC.SPACE + HandelC.INTEGERL;
        this.hcCode = this.hcCode + HandelC.getInstance().generateComment(comment);

        // #define integer int 3
        String declaration = HandelC.INTEGERL + HandelC.SPACE + HandelC.INT 
                            + HandelC.SPACE + this.directives.getIntBits();
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(declaration);

        // #define integer_card 8
        declaration = HandelC.getInstance().getCardVarOf(HandelC.INTEGERL) 
                            + HandelC.SPACE + this.getIntCard();
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(declaration);

        // #define integer_offset integer_card / 2
        declaration = HandelC.INTEGERL + HandelC.OFFSET 
                            + HandelC.SPACE 
                            + HandelC.getInstance().getCardVarOf(HandelC.INTEGERL) 
                            + HandelC.SPACE + HandelC.MATH_DIV
                            + HandelC.SPACE + HandelC.LOGICAL_VALUES;
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(declaration);

        // #define integer_set unsigned int integer_card
        declaration = HandelC.getInstance().getSetVarOf(HandelC.INTEGERL) + HandelC.SPACE
                            + HandelC.UNSIGNED + HandelC.SPACE
                            + HandelC.INT + HandelC.SPACE 
                            + HandelC.getInstance().getCardVarOf(HandelC.INTEGERL);
        this.hcCode = this.hcCode + HandelC.getInstance().generateDefine(declaration);

        List intRange = MathUtil.getSignedIntRangeOrderedBySignedRepresentation(this.directives.getIntBits());

        this.currentDefinition = HandelC.INTEGERL;
        this.currentBranches = new Vector();
        
        for (int i=0; i < intRange.size(); i++) {
            this.currentBranches.add(this.getIntName(((Integer)intRange.get(i)).intValue()));
        }

        this.declareSingletonSets(this.currentDefinition,ListUtil.reverse(this.currentBranches));
        this.declareSingletonSetsLookupTable(this.currentDefinition,ListUtil.reverse(this.currentBranches));

        // AUXILIARY MACROS
        /*
        macro expr unsigned_MACRO(integer_local_) = ((unsigned)(integer_local_ + integer_card + integer_offset) % integer_card);
        */
        /*
        this.hcCode = this.hcCode + HandelC.generateComment(Messenger.INT_MACRO);
        String localInt = HandelC.INTEGERL + HandelC.LOCAL_COPY;
        String unsignedMacro = HandelC.MACRO_EXP + HandelC.SPACE 
                        + HandelC.UNSIGNED_MACRO + HandelC.PAR_BEGIN
                        + localInt + HandelC.PAR_END 
                        + HandelC.SPACE + HandelC.ASSIGNMENT + HandelC.SPACE 
                        + HandelC.PAR_BEGIN + HandelC.PAR_BEGIN                        
                        + HandelC.UNSIGNED + HandelC.PAR_END                        
                        + HandelC.PAR_BEGIN
                        + localInt + HandelC.SPACE 
                        + HandelC.MATH_ADD + HandelC.SPACE
                        + HandelC.INTEGERL + HandelC.TYPE_CARDINALITY + HandelC.SPACE
                        + HandelC.MATH_ADD + HandelC.SPACE
                        + HandelC.INTEGERL + HandelC.OFFSET + HandelC.PAR_END 
                        + HandelC.SPACE + HandelC.MATH_REMAINDER + HandelC.SPACE 
                        + HandelC.INTEGERL + HandelC.TYPE_CARDINALITY 
                        + HandelC.PAR_END + HandelC.END_COMMAND;
        this.hcCode = this.hcCode + unsignedMacro;
         */
    }
    
    // Returns a string representaion for integers
    public String getIntName(int i) {
        // integer_n (for n >= 0)
        // integer_neg_n (for n < 0)
        return HandelC.INTEGERL + HandelC.UNDERSCORE + MathUtil.getIntName(i);
    }

    // Returns the cardinality of the integers
    private int getIntCard() {
        return MathUtil.getCardinality(directives.getIntBits());
    }
    
    // Returns the maximal integer
    private int getIntMax() {
        return this.getIntCard() - 1;
    }

    // Declares the macros used for each tag in a given datatype
    private void declareLookupTables(String type) {
        List tags = this.getTags(type);
        for (int i=0; i<tags.size(); i++) {
            this.declareTagLookupTable(type, (String)tags.get(i));
        }
    }
    
    // Returns the list of tags of a given datatype
    private List getTags(String type){
        List tags = new Vector();
        Enumeration keys = this.tagParents.keys();
        while(keys.hasMoreElements()) {
            String key = (String)keys.nextElement();
            if (this.tagParents.get(key).equals(type)) {
                tags.add(key);
            }
        }
        
        return tags;
    }
    
    // Declares the macro of a given tag used within a given datatype
    private void declareTagLookupTable(String type, String tag) {
        //String comment = Messenger.TAG_MACRO_1 + tag + Messenger.TAG_MACRO_2 + type;
        //this.hcCode = this.hcCode + HandelC.generateComment(comment);
        
        if (this.tagTypes.containsKey(tag)) {
            
            List types = (List)this.tagTypes.get(tag);
            List values = (List)this.tagValues.get(tag);

            // Name of the table is type_tag_LUT (Lookup Table)
            String tableName = HandelC.getInstance().getLookupTableName(type,tag);
            
            // Arguments are type_tag_local_<index>
            String tableDimensions = "";
            for (int i=0; i < types.size(); i++) {
                tableDimensions = tableDimensions + HandelC.ARRAY_BEGIN;
                tableDimensions = tableDimensions + HandelC.getInstance().getCardVarOf(types.get(i).toString());
                tableDimensions = tableDimensions + HandelC.ARRAY_END;
            }

            List options = ListUtil.cartesianProduct(values);
            String array = this.getArrayInstantiation(tag, values,options);
            
            String tableDecl = HandelC.STATIC + HandelC.SPACE + type + HandelC.SPACE + tableName + tableDimensions + HandelC.SPACE + HandelC.ASSIGNMENT 
                            + array + HandelC.END_COMMAND;
            this.hcCode = this.hcCode + tableDecl;
            
        } else {
            throw new InvalidDatatypeTagException(type,tag);
        }
    }
    
    private String getArrayInstantiation(String tag, List values, List options) {
        String array = HandelC.BLOCK_BEGIN;
        
        int cardinality = ((List)values.get(0)).size();

        for (int i=0; i<cardinality; i++) {
            if (values.size() == 1) {
                array = array + this.generateValue(tag, (List)options.remove(0));
            } else {
                array = array + this.getArrayInstantiation(tag,values.subList(1,values.size()),options);
            }
            if (i != cardinality - 1) {
                array = array + HandelC.SPACE + HandelC.SEP_ARGUMENTS;
            }
        }
        array = array + HandelC.BLOCK_END;
        return array;
    }
    
    // Returns a value of a select block of a given and list of values)
    private String generateValue(String tag, List values) {
        String value = tag;
        for (int i=0; i < values.size(); i++) {
            String v = (String)values.get(i);
            try {
                Integer n = new Integer(v);
                v = MathUtil.getIntName(n.intValue());
            } catch (NumberFormatException e) {
                // It is not a number; do nothing 
            }
            value = value + HandelC.UNDERSCORE + v;
        }
        return value;
    }

    // Declares the macro that implements a lookup table for every
    // possible value in the system and returns each possible corresponding
    // singleton set
    /*
    private void declareSingletonSetsMacro(){
        this.hcCode = this.hcCode + HandelC.generateComment(Messenger.SINGLETON_SETS_MACRO_DECLARATION);
        
        List integers = MathUtil.getSignedIntRangeOrderedBySignedRepresentation(this.directives.getIntBits());
        
        List booleans = new Vector();
        booleans.add(HandelC.LOGICAL_TRUE);
        booleans.add(HandelC.LOGICAL_FALSE);
        
        List datatypesBranches = new Vector();
        Enumeration keys = this.typeBranches.keys();
        while(keys.hasMoreElements()) {
            List branches = (List)this.typeBranches.get((String)keys.nextElement());
            datatypesBranches.addAll(branches);
        }
        
        // The argument name is singleton_sets_local_0
        String arg = this.getMacroArgName(HandelC.SINGLETON, HandelC.SETS, 0);
        
        List conditions = new Vector();
        List values = new Vector();
        
        for(int i = 0; i < integers.size(); i++) {
            conditions.add(this.getCondition(arg,integers.get(i).toString()));
            values.add(this.getIntName(((Integer)integers.get(i)).intValue()) + HandelC.DATATYPE_MEMBER_SET);
        }
        for(int i = 0; i < booleans.size(); i++) {
            conditions.add(this.getCondition(arg,booleans.get(i).toString()));
            values.add(booleans.get(i) + HandelC.DATATYPE_MEMBER_SET);
        }
        for(int i = 0; i < datatypesBranches.size(); i++) {
            conditions.add(this.getCondition(arg,datatypesBranches.get(i).toString()));
            values.add(datatypesBranches.get(i) + HandelC.DATATYPE_MEMBER_SET);
        }
        
        // Name of the macro is singleton_sets_MACRO
        String macroName = this.getMacroName(HandelC.SINGLETON, HandelC.SETS);

        // Arguments are type_tag_local_<index>
        String macroArgs = HandelC.PAR_BEGIN + arg + HandelC.PAR_END;
        String macroBody = this.getConditionalBlock(conditions,values,HandelC.MACRO.length());

        String macroDecl = HandelC.MACRO_EXP + HandelC.SPACE 
                        + macroName + macroArgs + HandelC.SPACE + HandelC.ASSIGNMENT 
                        + HandelC.COMMENT_LINE_BREAK + macroBody + HandelC.END_COMMAND;
        
        this.hcCode = this.hcCode + macroDecl;
        
    }
    */
}
