package csp2hc.visitors;

import java.util.Enumeration;
import java.util.HashSet;
import java.util.Hashtable;
import java.util.Iterator;
import java.util.LinkedList;
import java.util.ListIterator;
import java.util.Set;
import java.util.Stack;

import lmf.formula.csp.analysis.ExtendedDepthFirstAdapter;
import lmf.formula.csp.node.ACspAlphabetisedParallelProcess;
import lmf.formula.csp.node.ACspAssert;
import lmf.formula.csp.node.ACspEvent;
import lmf.formula.csp.node.ACspInputComunication;
import lmf.formula.csp.node.ACspInterleaveProcess;
import lmf.formula.csp.node.ACspOutputComunication;
import lmf.formula.csp.node.ACspPrefixProcess;
import lmf.formula.csp.node.ACspProcessCallExpr;
import lmf.formula.csp.node.ACspProcessDefinition;
import lmf.formula.csp.node.ACspSequentialProcess;
import lmf.formula.csp.node.ACspSharingParallelProcess;
import lmf.formula.csp.node.Node;
import lmf.formula.csp.node.PEvent;
import lmf.formula.csp.node.PExpression;
import lmf.formula.csp.node.PHelp;
import lmf.formula.csp.node.Start;
import mathCollection.HashMathSet;
import mathCollection.MathSet;
import mathCollection.SetOfSets;
import csp2hc.exception.UndefinedChannelInOutException;
import csp2hc.messages.Directives;
import csp2hc.messages.HandelC;
import csp2hc.util.Branch2BranchSetLUT;
import csp2hc.util.BranchActionsOnChannel;
import csp2hc.util.ForcedInterleavingOnChannel;
import csp2hc.util.CspUtil;
import csp2hc.util.MathUtil;
import csp2hc.util.SetUtil;

public class ParallelismAnalyserVisitor extends ExtendedDepthFirstAdapter {
    
    // Visitor that has the information about the processes
    private ProcessesInfoVisitor procsInfo;
    
    // Visitor that has the information about the directives
    private DirectivesVisitor directives;

    // Visitor that retrieves information about the alphabets
    private AlphabetsVisitor alphabets;
    
    // Stores the name of the current process being analysed
    private String currentAnalysedProcess;

    // Stores the name of the current process definition being analysed
    private String currentProcessDefinition;

    // Set of processes names that already been analysed while 
    // checking a concurrent process
    private Set<String> checkedProcesses;

    // Indicates to the analyser that a branch has happened to the left
    // of a parallel composition
    private boolean existsLeftProcess;
    
    // Maximum identification used for parallel branches starting from main at 0
    private int nextBranchId;

    // Maps each node of the tree to its branch
    private Hashtable<Integer, Integer> nodeBranches;

    // Stack of maximum ids for each parallel operator
    private Stack<Integer> maximumParallelIds;

    // Stores the information about the forced interleaving events
    private ForcedInterleavingOnChannel forcedInterleavingLUT;
    
    // Set of events on which synchronisation is actually happening 
    private Stack<Set<String>> syncChannels;

    // Stores the information about the synchronisation if 
    // there were no forced interleaving events
    // CHANNEL NAME -> BranchMultiSyncLUT
    private ForcedInterleavingOnChannel fakeSyncLUT;

    // Stores the information about the real synchronisation of events 
    // CHANNEL NAME -> Set BRANCH_ID
    private Hashtable<String,Set<Set<Integer>>> realSyncLUT;

    // For each channel, stores the id of the branches that mention that channel  
    // CHANNEL_NAME -> Set BRANCH_ID
    private Hashtable<String,Set<Integer>> branchAlphabets;

    public int getNumberOfBranches() { return this.nextBranchId + 1; }
    public ForcedInterleavingOnChannel getForcedInterleavingLUT() { return this.forcedInterleavingLUT; }
    
    // Stores the information about the actions that are taken by each branch for a given channel
    private BranchActionsOnChannel branchActionsOnChannel;
    
    public Hashtable<String,Set<Set<Integer>>> getRealSyncLUT() { return this.realSyncLUT; }
    public BranchActionsOnChannel getBranchActionsOnChannel() { return this.branchActionsOnChannel; }
    
    private boolean getExistsLeftProcess() {
    	return this.existsLeftProcess;
    }
    
    private void setExistsLeftProcess(boolean newv) {
    	this.existsLeftProcess = newv;
    }
    
    public boolean isMultiSync() {
    	boolean retorno = false;
    	
    	Enumeration<String> channels = this.realSyncLUT.keys();
    	
    	while(channels.hasMoreElements() && !retorno) {
    		String channel = channels.nextElement();
    		Set<Set<Integer>> syncs = this.realSyncLUT.get(channel);
    		
    		for (Iterator<Set<Integer>> iter = syncs.iterator(); iter.hasNext() && !retorno;) {
				Set<Integer> sync = (Set<Integer>)iter.next();
				if (sync.size() > 2) {
					retorno = true;
				}
			}
    	}
    	
    	return retorno;
    	
    }

    public Set<String> getMultiSyncChannels() {
    	Set<String> retorno = new HashSet<String>();
    	Enumeration<String> channels = this.realSyncLUT.keys();
    	
    	while(channels.hasMoreElements() && !retorno.equals("")) {
    		String channel = channels.nextElement();
    		Set<Set<Integer>> syncs = this.realSyncLUT.get(channel);
    		
    		for (Iterator<Set<Integer>> iter = syncs.iterator(); iter.hasNext() && retorno!=null;) {
				Set<Integer> sync = (Set<Integer>)iter.next();
				if (sync.size() > 2) {
					retorno.add(HandelC.getInstance().getChannel(channel));
				}
			}
    	}
    	
    	return retorno;
    	
    }
    
    public ParallelismAnalyserVisitor(ProcessesInfoVisitor procsInfo, DirectivesVisitor directives) {
        this.procsInfo = procsInfo;
        this.directives = directives;
        this.alphabets = new AlphabetsVisitor(this.procsInfo, this.directives);
        this.initialise();
    }
    
    private void initialise() {
    	this.currentAnalysedProcess = "";
    	this.currentProcessDefinition = "";
        this.checkedProcesses = new HashSet<String>();
        this.nodeBranches = new Hashtable<Integer, Integer>();
        this.initialiseForEachProcess();
        this.forcedInterleavingLUT = new ForcedInterleavingOnChannel();
        this.syncChannels = new Stack<Set<String>>();
        this.fakeSyncLUT = new ForcedInterleavingOnChannel();
        this.realSyncLUT = new Hashtable<String,Set<Set<Integer>>>();
        this.branchAlphabets = new Hashtable<String,Set<Integer>>();
        this.branchActionsOnChannel = new BranchActionsOnChannel();
    }

    private void setCurrentAnalysedProcess(String s) {
    	this.currentAnalysedProcess = s;
    	this.alphabets.setCurrentAnalysedProcess(s);
    }
    private void setCurrentProcessDefinition(String s) {
    	this.setCurrentAnalysedProcess(s);
    	this.currentProcessDefinition = s;
    }
    
    private void initialiseForEachProcess() {
        this.nextBranchId = 0;
        this.maximumParallelIds = new Stack<Integer>();
        this.maximumParallelIds.push(new Integer(0));
    }
    
    public void caseStart(Start node){
    	node.getPSpecification().apply(this);


    	this.filterLUT(this.forcedInterleavingLUT);
    	this.filterLUT(this.fakeSyncLUT);

    	this.buildRealSyncLUT();

    	System.out.println("ParallelismAnalyserVisitor :: caseStart -> Branches alphabets: " + this.branchAlphabets);
    	System.out.println("ParallelismAnalyserVisitor :: caseStart -> ForcedLUT: " + this.forcedInterleavingLUT);
    	System.out.println("ParallelismAnalyserVisitor :: caseStart -> Real SyncLUT: " + this.realSyncLUT);
    	System.out.println("ParallelismAnalyserVisitor :: caseStart -> Multi sync: " + this.isMultiSync());
    	System.out.println("ParallelismAnalyserVisitor :: caseStart -> Actions: " + this.branchActionsOnChannel);
    	
    }
    
    public void caseACspProcessDefinition(ACspProcessDefinition node){
    	this.setCurrentProcessDefinition(node.getName().toString().trim());
        this.initialiseForEachProcess();
        PExpression expression = node.getExpression();
        expression.apply(this);
    }

    public void caseACspProcessCallExpr(ACspProcessCallExpr node) {
    	String pname = node.getName().toString().trim();

        if (!pname.equals(CspUtil.getSkip()) 
        		&& !pname.equals(CspUtil.getStop())
        		&& !this.checkedProcesses.contains(pname)) {
    		this.checkedProcesses.add(pname);

        	this.setCurrentAnalysedProcess(pname);
    		
        	PExpression proc = (PExpression)this.procsInfo.getProcDefinition(pname);
        	proc.apply(this);
    		this.checkedProcesses.remove(pname);

    		this.includeNodeBranchId(node);
    	}
    }
    
    public void caseACspSequentialProcess(ACspSequentialProcess node){
        node.getLeftProcess().apply(this);
        this.setExistsLeftProcess(true);
        node.getRightProcess().apply(this);
        this.setExistsLeftProcess(false);
    }

    public void caseACspPrefixProcess(ACspPrefixProcess node){
        PEvent event = node.getEvent();
        LinkedList comm = ((ACspEvent)node.getEvent()).getComunication();
        PHelp proc = node.getProcess();

        String channel = HandelC.getInstance().getChannelName(((ACspEvent)event).getChannelName().toString().trim());
    	Integer branch = this.maximumParallelIds.peek();
        String action = this.getAction(channel, comm);
        
    	this.includeNodeBranchId(event);
    	
    	if (this.currentProcessDefinition.equals(HandelC.MAINP)) {
        	this.includeBranchInChannel(channel, branch);
        	this.branchActionsOnChannel.update(channel, branch, action);
    	}
    	
    	this.setExistsLeftProcess(true);
        proc.apply(this);
        this.setExistsLeftProcess(false);
    }

    private String getAction(String channel_name, LinkedList comms) {
    	String channel = HandelC.getInstance().getChannel(channel_name);
    	String action = "";
        ListIterator commIter = comms.listIterator(0);
        String process = this.currentAnalysedProcess;
        
        if (!commIter.hasNext()) {
            if (this.directives.containsProcChannelInOut(process)) {
                Hashtable channelInOut = this.directives.getProcChannelInOut(process);
                if (channelInOut.containsKey(channel)){
                    action = (String)channelInOut.get(channel);
                } 
            }
       } else {
            Object next = commIter.next();
            if (next instanceof ACspInputComunication) {
            	action = Directives.CHANNEL_IN;
            } else if (next instanceof ACspOutputComunication) {
            	action = Directives.CHANNEL_OUT;
            }
        }
        return action;
    }
    
    public void caseACspSharingParallelProcess(ACspSharingParallelProcess node) {
        this.analyseConcurrency(node);
    }
    
    public void caseACspAlphabetisedParallelProcess(ACspAlphabetisedParallelProcess node) {
        this.analyseConcurrency(node);
    }
    
    public void caseACspInterleaveProcess(ACspInterleaveProcess node) {
        this.analyseConcurrency(node);
    }

    // Indicates that there are events in which we will need to force the interleaving
    public boolean needsForcedInterleaving(){
        return !this.forcedInterleavingLUT.isEmpty();
    }

    private void analyseConcurrency(Node node) {
        Set forcedInterleaving = null;

        // Checking the branches
        Node left = null;
        Node right = null;
        Set<String> syncChannels = null;

        if (node instanceof ACspSharingParallelProcess) {
            left = ((ACspSharingParallelProcess)node).getLeftProcess();
            right = ((ACspSharingParallelProcess)node).getRightProcess();
            forcedInterleaving = this.getForcedInterleavingChannels((ACspSharingParallelProcess)node);
            syncChannels = this.getSyncChannels((ACspSharingParallelProcess)node);
    	} else if (node instanceof ACspAlphabetisedParallelProcess) {
            left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
            right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
            forcedInterleaving = this.getForcedInterleavingChannels((ACspAlphabetisedParallelProcess)node);
            syncChannels = this.getSyncChannels((ACspAlphabetisedParallelProcess)node);
    	} else if (node instanceof ACspInterleaveProcess) {
            left = ((ACspInterleaveProcess)node).getLeftProcess();
            right = ((ACspInterleaveProcess)node).getRightProcess();
            forcedInterleaving = this.getForcedInterleavingChannels((ACspInterleaveProcess)node);
            syncChannels = this.getSyncChannels((ACspInterleaveProcess)node);
    	}

		if (this.getExistsLeftProcess()) {
    		nextBranchId++;
            this.setExistsLeftProcess(false);
    	}
		this.maximumParallelIds.push(new Integer(nextBranchId));
        
		int initialLeftId = nextBranchId;
        left.apply(this);
		int finalLeftId = nextBranchId;
        
		// Before going to the right, increment the current ID for this parallel composition
        // This is done via the stack
		this.maximumParallelIds.pop();
		nextBranchId++;
		this.maximumParallelIds.push(new Integer(nextBranchId));
		
		int initialRightId = nextBranchId;
        right.apply(this);
		int finalRightId = nextBranchId;

		// Before leaving this parallel composition, remove its current branch id from the stack
        // This is done via the stack
		this.maximumParallelIds.pop();

		// Updates the LUT that stores the mutual exclusion information
        Set leftIds = MathUtil.getInterval(initialLeftId,finalLeftId);
        Set rightIds = MathUtil.getInterval(initialRightId, finalRightId);

        this.updateLUT(this.forcedInterleavingLUT, leftIds, rightIds, forcedInterleaving);
        this.updateLUT(this.fakeSyncLUT, leftIds, rightIds, syncChannels);
    }

    private void updateLUT(ForcedInterleavingOnChannel lut, Set leftIds, Set rightIds, Set channels) {
    	// The lookup table only needs to be updated for the process main
    	if (this.currentProcessDefinition.equals(HandelC.MAINP)) { 
    		lut.update(leftIds,rightIds,channels);
    	}
    }

    private void filterLUT(ForcedInterleavingOnChannel channelLUT) {

    	Set<String> lutKeys = channelLUT.keySet(); 
    	for (Iterator iter = lutKeys.iterator(); iter.hasNext();) {
			String channel = (String)iter.next();
			
			Set<Integer> branchAlphabet = this.branchAlphabets.get(channel);

			Branch2BranchSetLUT branchLUT = channelLUT.get(channel);
	    	Set<Integer> bKeys = branchLUT.keySet(); 
	    	for (Iterator iterB = bKeys.iterator(); iterB.hasNext();) {
	    		Integer branch = (Integer) iterB.next();
	    		Set<Integer> branches = branchLUT.get(branch);
	    		if (branchAlphabet.contains(branch)) {
	    			// Okay, go ahead and remove the branches that do not mention
	    			// this channels
					branches.retainAll(branchAlphabet);
	    		} else {
	    			// This branch does not mention the channel
	    			// Map it to an empty list
					branches.retainAll(new HashSet<Integer>());
	    		}
	    	}
			
    	}
    }
    
    
    private void includeNodeBranchId(Node node) {
    	if (this.checkedProcesses.isEmpty()) {
	    	Integer nodeId = new Integer(node.hashCode());
	    	Integer branchId = this.maximumParallelIds.peek();
	    	this.nodeBranches.put(nodeId, branchId);
    	}
    }
    private void includeBranchInChannel(String name, Integer branchId) {
    	if(this.branchAlphabets.containsKey(name)) {
    		Set<Integer> current = this.branchAlphabets.get(name);
    		current.add(branchId);
    		this.branchAlphabets.put(name, current);
    	} else {
    		Set<Integer> current = new HashSet<Integer>();
    		current.add(branchId);
    		this.branchAlphabets.put(name, current);
    	}
    }
    
            
    public int getNodeBranchId(Node node) {
    	int id = -1;
    	if (this.nodeBranches.containsKey(node.hashCode())) {
    		id = this.nodeBranches.get(node.hashCode()).intValue();
    	}
    	return id;
    }


    public Set<String> getForcedInterleavingChannels(ACspSharingParallelProcess node, String pname) {
    	this.setCurrentAnalysedProcess(pname);
        return this.getForcedInterleavingChannels(node);
    }
    public Set<String> getForcedInterleavingChannels(ACspSharingParallelProcess node) {
        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();
        return this.getForcedInterleavingChannels(left, right, this.getSyncSet(node));
    }
    public Set<String> getForcedInterleavingChannels(ACspAlphabetisedParallelProcess node, String pname) {
    	this.setCurrentAnalysedProcess(pname);
        return this.getForcedInterleavingChannels(node);
    }
    public Set<String> getForcedInterleavingChannels(ACspAlphabetisedParallelProcess node) {
        Node left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
        Node right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
        return this.getForcedInterleavingChannels(left, right, this.getSyncSet(node));
    }
    public Set<String> getForcedInterleavingChannels(ACspInterleaveProcess node, String pname) {
    	this.setCurrentAnalysedProcess(pname);
        return this.getForcedInterleavingChannels(node);
    }
    public Set<String> getForcedInterleavingChannels(ACspInterleaveProcess node) {
        Node left = ((ACspInterleaveProcess)node).getLeftProcess();
        Node right = ((ACspInterleaveProcess)node).getRightProcess();
        return this.getForcedInterleavingChannels(left, right, this.getSyncSet(node));
    }

    private Set<String> getForcedInterleavingChannels(Node left, Node right, Set sync){
        left.apply(alphabets);
        Set leftChannels = alphabets.getVisibleChannels();
        this.alphabets.resetChannels();

        right.apply(alphabets);
        Set rightChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        rightChannels.retainAll(leftChannels); // rightUsedChannels intersection leftUsedChannels
        rightChannels.removeAll(sync); // rightUsedChannels minus sync

        return rightChannels;
    }
    
    public Set<String> getSyncChannels(ACspSharingParallelProcess node) {
        Node left = ((ACspSharingParallelProcess)node).getLeftProcess();
        Node right = ((ACspSharingParallelProcess)node).getRightProcess();
        return this.getSyncChannels(left, right, this.getSyncSet(node));
    }
    public Set<String> getSyncChannels(ACspAlphabetisedParallelProcess node) {
        Node left = ((ACspAlphabetisedParallelProcess)node).getLeftProcess();
        Node right = ((ACspAlphabetisedParallelProcess)node).getRightProcess();
        return this.getSyncChannels(left, right, this.getSyncSet(node));
    }
    public Set<String> getSyncChannels(ACspInterleaveProcess node) {
        Node left = ((ACspInterleaveProcess)node).getLeftProcess();
        Node right = ((ACspInterleaveProcess)node).getRightProcess();
        return this.getSyncChannels(left, right, this.getSyncSet(node));
    }
    
    public Set<String> getSyncSet(ACspSharingParallelProcess node) {
        PExpression syncExp = ((ACspSharingParallelProcess)node).getExpression();
        
        AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procsInfo,this.directives);
        alphabets.setCurrentAnalysedProcess(this.currentAnalysedProcess);
        
        Set<String> channels = alphabets.analyseChannelSet(syncExp);
        
        return channels;
    }
    public Set<String> getSyncSet(ACspAlphabetisedParallelProcess node) {
    	AlphabetsVisitor alphabets = new AlphabetsVisitor(this.procsInfo,this.directives);
        alphabets.setCurrentAnalysedProcess(this.currentAnalysedProcess);

        // Checking the synchronisation channel set
        PExpression syncExpLeft = ((ACspAlphabetisedParallelProcess)node).getLeftExpr();
        PExpression syncExpRight = ((ACspAlphabetisedParallelProcess)node).getRightExpr();
        Set<String> channelsLeft = alphabets.analyseChannelSet(syncExpLeft);
        Set<String> channelsRight = alphabets.analyseChannelSet(syncExpRight);
        
        // Intersection
        channelsLeft.retainAll(channelsRight);
        
        return channelsLeft;
    }
    public Set<String> getSyncSet(ACspInterleaveProcess node) {
        return new HashSet();
    }

    private Set<String> getSyncChannels(Node left, Node right, Set sync){
        left.apply(alphabets);
        Set leftChannels = alphabets.getVisibleChannels();
        this.alphabets.resetChannels();

        right.apply(alphabets);
        Set rightChannels = alphabets.getVisibleChannels();
        alphabets.resetChannels();

        rightChannels.retainAll(leftChannels); // rightUsedChannels intersection leftUsedChannels
        rightChannels.retainAll(sync); // rightUsedChannels minus sync

        return rightChannels;
    }

    private void buildRealSyncLUT() {
    	Enumeration<String> channels = this.fakeSyncLUT.keys();
    	while(channels.hasMoreElements()){
    		String channel = channels.nextElement();

    		Set<Set<Integer>> sync = new HashSet<Set<Integer>>();

    		Branch2BranchSetLUT fakeSync = this.fakeSyncLUT.get(channel);

    		Enumeration<Integer> branchesDomain = fakeSync.keys();
        	while(branchesDomain.hasMoreElements()){
        		Integer branch = branchesDomain.nextElement();
        		Set<Integer> branchesSyncBranch = fakeSync.get(branch);
        		
        		if (!branchesSyncBranch.isEmpty()) {
            		MathSet maximalSet = new HashMathSet();
            		maximalSet.add(branch);
            		maximalSet.addAll(branchesSyncBranch);

    	       		SetOfSets powerset = maximalSet.powerSet();
            		
    	       		// Verify which set of branches synchronisation on channel are valid
    	       		// regarding forcedSyncLUT
    	       		Set<Set<Integer>> validSync = new HashSet<Set<Integer>>();
    				Branch2BranchSetLUT forcedIterleaving = this.forcedInterleavingLUT.get(channel);
            		for (Iterator iterSyncSet = powerset.iterator(); iterSyncSet.hasNext();) {
    					Set<Integer> syncSet = (Set<Integer>) iterSyncSet.next();

    					boolean isValid = true;
    	        		for (Iterator iterBranch = syncSet.iterator(); iterBranch.hasNext();) {
    						Integer b = (Integer) iterBranch.next();

    						// Forced interleaving on channel C with branch b
    						Set<Integer> forced = forcedIterleaving.get(b);

    		        		Set<Integer> syncSetCopy = new HashSet<Integer>();
    						syncSetCopy.addAll(syncSet);
    						syncSetCopy.remove(b);
    						syncSetCopy.retainAll(forced);
    						if(!syncSetCopy.isEmpty()) {
    							isValid = false;
    						}
    					}

    	        		if(isValid) {
    	        			validSync.add(syncSet);
    	        		}
    				}
            		Set<Set<Integer>> maximalValidSync = SetUtil.maximalSets(validSync);
            		sync.addAll(maximalValidSync);
        		}
        			
        	}
        	this.realSyncLUT.put(channel, sync);
    	}
    }
    
    public void caseACspAssert(ACspAssert node) {
    }
    
}

